#include "Graphics.c"
#include "Sound.c"
#include <direct.h>
#include <sys/stat.h>
#include <time.h>
#include <intrinsics.h>

#define MAXP 4 // Maximum players
#define SOUNDCHANNELS 8
#define STEREO 4 // How stereo it is
#define HILEN 20
#define MAXNAMELEN 50

#define every_enemy(n) (n=0;n<am;n++)
#define every_bullet(n) (n=0;n<bm;n++)
#define every_item(n) (n=0;n<im;n++)
#define every_player(n) (n=0;n<pm;n++)
#define dust_particles(n) (n=0;n<fm;n++)
#define new_dust(n) (n=fm;n<nfm;n++)
#define flame_particles(n) (n=0;n<ffm;n++)
#define new_flame(n) (n=ffm;n<nffm;n++)

#define CXP (char *)malloc(MAXP)
#define IXP (int *)malloc(MAXP*sizeof(int))
#define FXP (float *)malloc(MAXP*sizeof(float))

enum sg_retcode {Fail,Pass,Quit,Error,TBA};

struct stat version_attributes;

char seeninfo[128];
double gamestart,gametime,nextsave;
char gamestack[50]; int gamedepth; // VERY VERY IMPORTANT
enum sg_retcode subgameN(int); int nlevel;
enum sg_retcode subgameC(int); int clevel;
enum sg_retcode subgameG(int); int glevel;
enum sg_retcode subgameI(int); int ilevel;

LPDIRECTSOUNDBUFFER soundchannel[SOUNDCHANNELS];
char sound_working;

unsigned char upkey[MAXP]={VK_UP,VK_W,VK_NUMPAD8,VK_U},
				downkey[MAXP]={VK_DOWN,VK_S,VK_NUMPAD2,VK_J},
				leftkey[MAXP]={VK_LEFT,VK_A,VK_NUMPAD4,VK_H},
				rightkey[MAXP]={VK_RIGHT,VK_D,VK_NUMPAD6,VK_K},
				firekey[MAXP]={VK_RETURN,VK_TAB,VK_NUMPAD0,VK_LBUTTON},
				selectkey[MAXP]={VK_CONTROL,VK_SHIFT,VK_DECIMAL,VK_RBUTTON};
int lives[MAXP],level;
long score[MAXP+1],money[MAXP];

const int mswait=12; // Frame delay
const int itypes=14;
int numblox; const int numbg=7,numast=9;
float gamma,autosave;
char technical=0,*openfile;
int pm=1;

unsigned char *num[2][MAXP][10];

void ucswap(unsigned char *x,unsigned char *y)
{
	unsigned char t=*x; *x=*y; *y=t;
}

void readani(unsigned char **anim,const int frames,const char *filename,const long offset)
{
	int f,n,x,y;
	FILE *in=fopen(filename,"rb");
	fseek(in,offset,SEEK_SET);
	for (f=0;f<frames;f++)
	{
		x=fgetc(in); y=fgetc(in);
		anim[f]=(unsigned char *)malloc(x*y+2);
		anim[f][0]=x; anim[f][1]=y;
		for (n=0;n<x*y;n++)
			anim[f][n+2]=fgetc(in);
	}
	fclose(in);
}

void plussign(const unsigned x,const unsigned y,const unsigned char c)
{
	putpix(x,y,c);
	putpix(x+1,y,c);
	putpix(x-1,y,c);
	putpix(x,y+1,c);
	putpix(x,y-1,c);
}

void wrapbitc(int x,int y,const unsigned char *bitmap,const unsigned char col) // Coloured mask
{
	unsigned xs=bitmap[0],ys=bitmap[1],px=(x<0)?(x+=xres):0,py=(y<0)?(y+=yres):0;
	unsigned char *c=bitmap+2,*p=shad+y*xres+x,*l=shad+yres*xres-xs;
	for (py=0;py<ys;py++)
	{
		for (px=0;px<xs;px++,c++,p++) if (*c!=255) *p=col;
		p+=xres-xs;
		if (p>=l) p-=xres*(yres-1);
	}
}

void wrapbitsub(int x,int y,const unsigned char *bitmap,const unsigned char colmap[256])
{ // Colour substitution
	unsigned xs=bitmap[0],ys=bitmap[1],px=(x<0)?(x+=xres):0,py=(y<0)?(y+=yres):0;
	unsigned char *c=bitmap+2,*p=shad+y*xres+x,*l=shad+yres*xres-xs,rc;
	for (py=0;py<ys;py++)
	{
		for (px=0;px<xs;px++,c++,p++) if ((rc=colmap[*c])!=255) *p=rc;
		p+=xres-xs;
		if (p>=l) p-=xres*(yres-1);
	}
}

void writepal(const float gamma)
{
	int n; float r[256],g[256],b[256];
	for (n=0;n<16;n++)
	{
		r[n]=((n&4)>0)?0.5+0.5*((n&8)>0):0;
		g[n]=((n&2)>0)?0.5+0.5*((n&8)>0):0;
		b[n]=((n&1)>0)?0.5+0.5*((n&8)>0):0;
	}
	r[8]=g[8]=b[8]=0.25;
	for (n=16;n<32;n++)
		r[n]=g[n]=b[n]=(float)(n-16)/15;
	for (n=32;n<128;n++)
	{
		float a=1,i,c=0,x=(n-32)/24;
		if (x==3) a=0.5;
		else if (x==1) c=0.5;
		else if (x==2) a=0.25;
		x=(n-32)%8; if (x>4) x=8-x; x/=4;
		i=a*x+c*(1.0-x);
		x=(n-32)/4%6;
		if (x==0) {r[n]=i; g[n]=c; b[n]=a;}
		else if (x==1) {r[n]=a; g[n]=c; b[n]=i;}
		else if (x==2) {r[n]=a; g[n]=i; b[n]=c;}
		else if (x==3) {r[n]=i; g[n]=a; b[n]=c;}
		else if (x==4) {r[n]=c; g[n]=a; b[n]=i;}
		else {r[n]=c; g[n]=i; b[n]=a;}
	}
	for (n=128;n<253;n++)
	{
		r[n]=0.25*((n-128)%5);
		g[n]=0.25*((n-128)/5%5);
		b[n]=0.25*((n-128)/25);
	}
	for (n=253;n<256;n++) r[n]=g[n]=b[n]=0;
	{
		PALETTEENTRY colour[256];
		for (n=0;n<256;n++)
		{
			colour[n].peRed=pow(r[n],1.0/gamma)*255.4;
			colour[n].peGreen=pow(g[n],1.0/gamma)*255.4;
			colour[n].peBlue=pow(b[n],1.0/gamma)*255.4;
			colour[n].peFlags=PC_NOCOLLAPSE;
		}
		dde(IDirectDrawPalette_SetEntries(ddpal,0,0,256,colour));
	}
}

void titlepal(const float gamma)
{ /**/
	int n; float r[256],g[256],b[256];
	for (n=0;n<16;n++)
	{
		r[n]=((n&4)>0)?0.5+0.5*((n&8)>0):0;
		g[n]=((n&2)>0)?0.5+0.5*((n&8)>0):0;
		b[n]=((n&1)>0)?0.5+0.5*((n&8)>0):0;
	}
	for (n=16;n<32;n++)
		r[n]=g[n]=b[n]=(float)(n-16)/15;
	for (n=32;n<128;n++)
	{
		float a=1,i,c=0,x=(n-32)/24;
		if (x==3) a=0.5;
		else if (x==1) c=0.5;
		else if (x==2) a=0.25;
		x=(n-32)%8; if (x>4) x=8-x; x/=4;
		i=a*x+c*(1.0-x);
		x=(n-32)/4%6;
		if (x==0) {r[n]=i; g[n]=c; b[n]=a;}
		else if (x==1) {r[n]=a; g[n]=c; b[n]=i;}
		else if (x==2) {r[n]=a; g[n]=i; b[n]=c;}
		else if (x==3) {r[n]=i; g[n]=a; b[n]=c;}
		else if (x==4) {r[n]=c; g[n]=a; b[n]=i;}
		else {r[n]=c; g[n]=i; b[n]=a;}
	}
	for (n=128;n<256;n++)
	{
		float ff=(float)n/128-1,th=M_PI*tanh(0.1*n),rp=-cos(th),gp=-sin(th),bp=cos(th);
		r[n]=pow(ff,pow(2,rp)),g[n]=pow(ff,pow(2,gp)),b[n]=pow(ff,pow(2,bp));
	}
	{
		PALETTEENTRY colour[256];
		for (n=0;n<256;n++)
		{
			colour[n].peRed=pow(r[n],1.0/gamma)*255.4;
			colour[n].peGreen=pow(g[n],1.0/gamma)*255.4;
			colour[n].peBlue=pow(b[n],1.0/gamma)*255.4;
			colour[n].peFlags=PC_NOCOLLAPSE;
		}
		dde(IDirectDrawPalette_SetEntries(ddpal,0,0,256,colour));
	}
}

void palcoeffs(float r[256],float g[256],float b[256])
{
	int n;
	for (n=0;n<16;n++)
	{
		r[n]=((n&4)>0)?0.5+0.5*((n&8)>0):0;
		g[n]=((n&2)>0)?0.5+0.5*((n&8)>0):0;
		b[n]=((n&1)>0)?0.5+0.5*((n&8)>0):0;
	}
	for (n=16;n<32;n++)
		r[n]=g[n]=b[n]=(float)(n-16)/15;
	for (n=32;n<128;n++)
	{
		float a=1,i,c=0,x=(n-32)/24;
		if (x==3) a=0.5;
		else if (x==1) c=0.5;
		else if (x==2) a=0.25;
		x=(n-32)%8; if (x>4) x=8-x; x/=4;
		i=a*x+c*(1.0-x);
		x=(n-32)/4%6;
		if (x==0) {r[n]=i; g[n]=c; b[n]=a;}
		else if (x==1) {r[n]=a; g[n]=c; b[n]=i;}
		else if (x==2) {r[n]=a; g[n]=i; b[n]=c;}
		else if (x==3) {r[n]=i; g[n]=a; b[n]=c;}
		else if (x==4) {r[n]=c; g[n]=a; b[n]=i;}
		else {r[n]=c; g[n]=i; b[n]=a;}
	}
	for (n=128;n<253;n++)
	{
		r[n]=0.25*((n-128)%5);
		g[n]=0.25*((n-128)/5%5);
		b[n]=0.25*((n-128)/25);
	}
	for (n=253;n<256;n++) r[n]=g[n]=b[n]=0;
	for (n=0;n<256;n++)
	{
		r[n]=pow(r[n],1.0/gamma);
		g[n]=pow(g[n],1.0/gamma);
		b[n]=pow(b[n],1.0/gamma);
	}
}

unsigned char firecolour()
{
	if (rnd(2)) return 129+rnd(4);
		else return 137+5*rnd(4);
}

unsigned char plasmacolour()
{
	if (!rnd(10)) return 11;
		else return 228+rnd(10);
}

void gethiscores(char **names,long *scores)
{
	int n;
	FILE *in=fopen(".\\hiscores.dat","rb");
	for (n=0;n<HILEN;n++)
	{
		fread(names[n],MAXNAMELEN,1,in);
		fread(&scores[n],4,1,in);
	}
	fclose(in);
}

void sethiscores(const char **names,const long *scores)
{
	int n;
	FILE *out=fopen(".\\hiscores.dat","wb");
	for (n=0;n<HILEN;n++)
	{
		fwrite(names[n],MAXNAMELEN,1,out);
		fwrite(&scores[n],4,1,out);
	}
	fclose(out);
}

void resethiscores()
{
	int n;
	char *names[20]; long scores[20];
	for (n=0;n<20;n++)
	{
		names[n]=(char *)malloc(50);
		names[n]="Nobody";
		scores[n]=(int)(pow(0.1,n*0.1)*10000)*5;
	}
	sethiscores(names,scores);
	for (n=0;n<20;n++) free(names[n]);
}

char *ordinal(const int n)
{
	char *ordinals[4]={"st","nd","rd","th"};
	if (n%10==1 && n%100!=11) return ordinals[0];
	else if (n%10==2 && n%100!=12) return ordinals[1];
	else if (n%10==3 && n%100!=13) return ordinals[2];
	else return ordinals[3];
}

unsigned char scrambfunc(const int x)
{
	return (x*x)%255;
}

void createcredits(const char *filename)
{
	FILE *in=fopen(filename,"rb"),*out=fopen(".\\c2.dat","wb");
	int n;
	for (n=0;!feof(in);n++)
		fputc(fgetc(in)^scrambfunc(n),out);
	fclose(in); fclose(out);
}

void maxpix(const int x,const int y,const unsigned char c)
{
	unsigned char *a=shad+x+y*xres;
	if (*a<c) *a=c;
}

void wupix(const float x,const float y,float c)
{
	int ix=(int)x,iy=(int)y; float rx=x-ix,ry=y-iy,sx=1.0-rx,sy=1.0-ry;
	if (ix<0 || iy<0 || ix>=xres-1 || iy>=yres-1) return;
	c-=16;
	maxpix(ix,iy,16+c*sqrt(sx*sy));
	maxpix(ix+1,iy,16+c*sqrt(rx*sy));
	maxpix(ix,iy+1,16+c*sqrt(sx*ry));
	maxpix(ix+1,iy+1,16+c*sqrt(rx*ry));
}

const int legalsize[21]={0,1,2,3,4,5,6,7,8,10,12,14,16,20,24,28,32,48,64,128,256};

int islegalsize(const int mass)
{
	int n,ret=0;
	for (n=0;n<21;n++)
		if (legalsize[n]==mass) ret=n;
	return ret;
}

void makeasteroidset(unsigned char **ast,unsigned char *asz,const int set)
{
	int n,i,x,y;
	asz[0]=0; // Asteroid bitmap creation
	for (n=1;n<=20;n++)
	{
		int d,raster=1; float hd,ds;
		d=(int)(10.0*sqrt(legalsize[n])); // Could be changed by set
		if (set==1) d=(int)(15.0*sqrt(legalsize[n]));
		if (set==7) raster=0;
		if (set==8)	d=(int)(70.0/sqrt(legalsize[n]));
		hd=0.5*(d-1); ds=hd*hd;
		asz[n]=d;
		ast[n-1]=(unsigned char *)malloc(d*d+2);
		ast[n-1][0]=ast[n-1][1]=d;
		if (raster) for (x=0;x<d;x++) for (y=0;y<d;y++)
		{
			if (set==1) // Coloured bullseyes
			{
				float dd=sq((float)x-hd)+sq((float)y-hd);
				dd=(int)(sqrt(dd)/hd*5);
				ast[n-1][x+y*d+2]=(dd==0 || dd==2 || dd==4)?n%7+9:255;
			}
			else if (set==2) // Groovy technicolour asteroids
			{
				float efx=1-(sq((float)x-hd)+sq((float)y-hd))/ds;
				if (efx>0)
				{
					float r,g,b;
					if (n>=1 && n<=5) {r=5-n; g=0; b=4;}
					else if (n<=9) {r=0; g=n-5; b=4;}
					else if (n<=13) {r=0; g=4; b=13-n;}
					else if (n<=17) {r=n-13; g=4; b=0;}
					else {r=4; g=21-n; b=0;}
					r*=efx; g*=efx; b*=efx;
					ast[n-1][x+y*d+2]=128+(int)(r+frnd(.999))+5*(int)(g+frnd(.999))+25*(int)(b+frnd(.999));
				}
				else ast[n-1][x+y*d+2]=255;
			}
			else if (set==3) // Astoroids [a.k.a. Cheerios]
			{
				float dd=sqrt(sq((float)x-hd)+sq((float)y-hd))/hd;
				unsigned char c;
				if (dd*3<1 || dd*3>3) c=255;
				else if (dd*3<2)
				{
					dd=(dd*3-1)*2+frnd(1);
					c=128+(int)(dd+2)+(int)(dd+1)*5+(int)(dd)*25;
				}
				else
				{
					dd=(3.0-dd*3)*2+frnd(1);
					c=128+(int)(dd+2)+(int)(dd+1)*5+(int)(dd)*25;
				}
				ast[n-1][x+y*d+2]=c;
			}
			else if (set==4) // Bubbles
			{
				float dd=floor(sqrt(sq((float)x-hd*4/3)+sq((float)y-hd*2/3)));
				unsigned char c=255;
				if (dd==0) c=15;
				else if (dd==1) c=11;
				else if (dd==2) c=7;
				dd=floor(hd-sqrt(sq((float)x-hd)+sq((float)y-hd)));
				if (dd==0) c=15;
				else if (dd==1) c=11;
				else if (dd==2) c=7;
				ast[n-1][x+y*d+2]=c;
			}
			else if (set==5) // Black discs
			{
				float dd=sq((float)x-hd)+sq((float)y-hd);
				ast[n-1][x+y*d+2]=(dd<hd*hd)?0:255;
			}
			else if (set==6) // Radioactive
			{
				float dd=sq((float)x-hd)+sq((float)y-hd);
				int yc=128+6*(int)((1.0-dd/sq(hd))*3+1+frnd(1));
				if (dd>hd*hd) ast[n-1][x+y*d+2]=255;
				else if (dd>sq(hd*5/6)) ast[n-1][x+y*d+2]=yc;
				else if (dd<sq(hd/6)) ast[n-1][x+y*d+2]=0;
				else if (dd<sq(hd/3)) ast[n-1][x+y*d+2]=yc;
				else
				{
					int i=(y>hd);
					i+=(fabs(x-hd)<fabs(y-hd)*0.5);
					ast[n-1][x+y*d+2]=i%2*yc;
				}
			}
			else // Asteroids (sets 0 and 8)
			{
				float efx=ds-sq((float)x-hd)-sq((float)y-hd);
				if (efx>0) ast[n-1][x+y*d+2]=20+8.0*efx/ds+frnd(1);
					else ast[n-1][x+y*d+2]=255;
			}
		}
		else if (set==7) // Classic
		{
			float px[25],py[25];
			int j=n+5;
			for (i=0;i<j;i++)
			{
				px[i]=d/2+d*0.4*sin(i*M_PI*2/j)+frnd(d*0.2)-d*0.1;
				py[i]=d/2+d*0.4*cos(i*M_PI*2/j)+frnd(d*0.2)-d*0.1;
			}
			rect(0,0,d,d,255);
			for (i=0;i<j;i++)
			{
				line(px[i]+1,py[i],px[(i+1)%j]+1,py[(i+1)%j],2);
				line(px[i]-1,py[i],px[(i+1)%j]-1,py[(i+1)%j],2);
				line(px[i],py[i]+1,px[(i+1)%j],py[(i+1)%j]+1,2);
				line(px[i],py[i]-1,px[(i+1)%j],py[(i+1)%j]-1,2);
			}
			for (i=0;i<j;i++)
				line(px[i],py[i],px[(i+1)%j],py[(i+1)%j],10);
			for (x=0;x<d;x++)
				for (y=0;y<d;y++)
					ast[n-1][x+y*d+2]=getpix(x,y);
		}
	}
}

float xfrom(float x)
{
	while (x>xres/2) x-=xres;
	while (x<-xres/2) x+=xres;
	return x;
}

float yfrom(float y)
{
	while (y>yres/2) y-=yres;
	while (y<-yres/2) y+=yres;
	return y;
}

float ddfrom(float x1,float y1,const float x2,const float y2)
{
	x1=xfrom(x1-x2);
	y1=yfrom(y1-y2);
	return x1*x1+y1*y1;
}

unsigned char bcalc(int x,int y,const int set,const int sub,float t)
{
	int n;
	if (y<40 || y>=yres-40) return 0;
	if (set==1) // Mantle
	{
		unsigned char c=202,col[10]={128,129,130,131,132,137,142,147,152,177};
		x+=t*3; y+=t*3;
		for (n=0;n<10 && c==202;n++)
		{
			x=(x+xres)%xres;
			y=(y+yres)%yres;
			if ((x*x+y*y)/50000%2) c=col[n];
			x*=2; y*=2;
			x-=t*3; y-=t*3;
		}
		return c;
	}
	else if (set==2) // Rainbow rings
	{
		float dd=sq(x-xres/2)+sq(y-yres/2);
		return 32+sub+(int)(sqrt(dd)/(sin(t/10)*10+20)+frnd(1))%24;
	}
	else if (set==3) // Perspex
	{
		int phase=(int)t/10%3;
		float px=x-xres/2,py=fabs(y-yres/2)+1e-6,grey;
		grey=24+4*(py/yres*4-1)+frnd(2)-1;
		px=px/py+t/10; py=log(py);
		if (phase==0) return ((int)px%2 && (int)py%2)?230:grey;
		else if (phase==1) return ((int)px%2 && y%2)?230:grey;
		else if (phase==2) return ((int)px%2 && (int)(py*5)%2)?230:grey;
	}
	else if (set==4) // Dark paravoid
	{
		float px=x-xres/2,py=y-yres/2,dd=(px*px+py*py+1)/5000;
		px/=dd; py/=dd;
		px+=t; py+=t/3;
		x=(int)fabs(px); y=(int)fabs(py);
		x%=256; y%=256;
		return 16+32/(5+(x&y));
	}
	else if (set==5) // Spinning multicoloured chequerboard
	{
		float ct=cos(t/25),st=sin(t/25),px=0.02*(x-xres/2),py=0.02*(y-yres/2),opx=px;
		px=px*ct+py*st+100;
		py=py*ct-opx*st+100;
		return (((int)px%2+(int)py%2)==1)?104+(int)(t/20+px/4+py/7)%24:0;
	}
	else if (set==6) // Neptune
	{
		float py=5000.0/y,px=4.0*(x-xres/2)/y;
		float p;
		px+=0.3*t;
		if (((int)px&(int)py)==0)
		{
			px=1000*sin(px*0.02);
			py=800*sin(py*0.03);
			return 16+15/(((int)px&(int)py)+1);
		}
		px-=0.3*t;
		px*=4.7; py*=4.7;
		px+=0.3*t;
		if (((int)px&(int)py)==4)
		{
			px=400*sin(px*0.05);
			py=300*sin(py*0.04);
			return 16+15/(((int)px&(int)py)+1);
		}
		p=0.002*x+0.01*y-2.5+frnd(1);
		if (p>0) return 128+25*(int)p;
	}
	return sub;
}

float mass(const int t,const int s)
{
	if (t==1 || t==12) return legalsize[s];
	else if (t==10) return 1e38;
	else return 1;
}

float radius(const int t,const int s,const unsigned char *asz)
{
	if (t==1 && s>=1 && s<=20) return 0.5*(asz[s]-1);
	else if (t==2) return 10;
	else if (t==3) return 5;
	else if (t==4) return sqrt(9*9+5*5);
	else if (t==5) return 1;
	else if (t==6) return 6;
	else if (t==7) return 3*s;
	else if (t==8) return 5;
	else if (t==9) return sqrt(12*12+9*9);
	else if (t==10 && s>=0 && s<=4) return 4*pot[s];
	else if (t==11) return 6;
	else if (t==12 && s>=1 && s<=20) return 0.5*(asz[s]-1);
	else if (t==13) return 5;
	else if (t==14) return sqrt(10*10+7*7);
	else if (t==15) return 3;
	else if (t==16) return 7;
	else if (t==17) return 10;
	else if (t==18) return 14;
	else if (t==19) return 20;
	else if (t==20) return 6;
	else if (t==21) return 9.9;
	else if (t==22) return 5;
	return 0;
}

int findfree(const char priority)
{
	int ret=(priority?rnd(SOUNDCHANNELS-1):SOUNDCHANNELS-1),n; char stop=0;
	for (n=0;n<SOUNDCHANNELS && !stop;n++)
	{
		DWORD status;
		IDirectSoundBuffer_GetStatus(soundchannel[n],&status);
		if ((status&DSBSTATUS_PLAYING)==0) {ret=n; stop=1;}
	}
	if (n==SOUNDCHANNELS-1 && priority) return rnd(SOUNDCHANNELS-1);
	return ret;
}

void gtdisp(const double gametime,const char *openfile)
{
	char str[20],*str2;
	sprintf(str,"#%d:%02d:%02d",(int)(gametime/3600),
										(int)(gametime/60)%60,
										(int)(gametime)%60);
	rect(xres/2-1,0,41,6,0);
	write(xres/2,0,xres,str,10,8);
	// Current open game display
	str2=strrchr(openfile,'\\')+1;
	if (str2!=(char *)1)
	{
		int n;
		strcpy(str,str2);
		*strrchr(str,'.')=0;
		n=estlength(str);
		rect(xres/2-11-n,0,n+10,6,0);
		write(xres/2-10-n,0,xres,str,2,0);
	}
}

void mouseoff()
{
	while ((mb&3 || kbhit) && !dealwithmessages() && !KEY(VK_F12));
}

int pause()
{
	int x,y;
	gamestart-=0.001*GetTickCount();
	nextsave-=0.001*GetTickCount();
	for (x=0;x<xres;x+=10)
		for (y=(x/10)%2*10;y<yres;y+=20)
		{
			rect(x+1,y+1,8,8,20);
			rect(x+2,y+2,6,6,19);
		}
	write(xres/2-10,yres/2-2,xres,"Paused",15,7);
	cclb();
	while (KEY(VK_P)) if (dealwithmessages()) return 1;
	while (!KEY(VK_P)) if (dealwithmessages()) return 1;
	while (KEY(VK_P)) if (dealwithmessages()) return 1;
	gamestart+=0.001*GetTickCount();
	nextsave+=0.001*GetTickCount();
	return 0;
}

void replotgrid(const unsigned char grid[32][24],const unsigned char **blox)
{
	int x,y,i;
	for (x=0;x<32;x++) for (y=0;y<24;y++) // Plot it
	{
		i=grid[x][y]; if (i==0xFA || i==0xFB) i=0;
		if (i>=0x50 && i<=0x53) // Notched WKobo blocks
		{
			unsigned char t[402];
			int xx,yy,xi=(x+1)%32,xd=(x+31)%32,yi=(y+1)%24,yd=(y+23)%24;
			memcpy(t,blox[i],402);
			if (!grid[xi][y])
				for (xx=0;xx<6;xx++) for (yy=0;yy<6;yy++) t[156+xx+yy*20]=255;
			if (!grid[xd][y])
				for (xx=0;xx<6;xx++) for (yy=0;yy<6;yy++) t[142+xx+yy*20]=255;
			if (!grid[x][yi])
				for (xx=0;xx<6;xx++) for (yy=0;yy<6;yy++) t[289+xx+yy*20]=255;
			if (!grid[x][yd])
				for (xx=0;xx<6;xx++) for (yy=0;yy<6;yy++) t[9+xx+yy*20]=255;
			wrapbitb(x*20,y*20,t);
		}
		else if (i>0 && grid[(x+1)%32][y]!=0xFA && grid[x][(y+1)%24]!=0xFB) wrapbitb(x*20,y*20,blox[i]);
		else if (i>0) // Scaled blocks
		{
			int xx,yy,xs=1,ys=1; char ok;
			while (grid[(x+xs)%32][y]==0xFA && x+xs<32) xs++;
			ok=1;
			while (ok)
			{
				for (xx=x;xx<x+xs;xx++) if (grid[xx%32][(y+ys)%24]!=0xFB) ok=0;
				if (y+ys>=24) ok=0;
				if (ok) ys++;
			}
			for (xx=0;xx<20;xx++) for (yy=0;yy<20;yy++)
				rectb(x*20+xx*xs,y*20+yy*ys,xs,ys,blox[i][2+xx+yy*20]);
		}
	}
}

int numberletters(const char *s)
{
	int n,ret=0;
	for (n=0;s[n];n++) if (lc(s[n])>='a' && lc(s[n])<='z') ret++;
	return ret;
}

int numericindex(const char *s,const int k)
{
	int n,i; char t[50];
	for (n=i=0;i<k;n++) if (lc(s[n])>='a' && lc(s[n])<='z') i++;
	strcpy(t,s+n);
	for (n=0;t[n]>='0' && t[n]<='9';n++);
	t[n]=0;
	return atoi(t);
}

char letterindex(const char *s,const int k)
{
	int n,i;
	for (n=i=0;i<=k;n++) if (lc(s[n])>='a' && lc(s[n])<='z') i++;
	return s[n-1];
}

void blackborder(const int x,const int y,const unsigned char *bit)
{
	wrapbitc(x-1,y,bit,0);
	wrapbitc(x,y-1,bit,0);
	wrapbitc(x,y+1,bit,0);
	wrapbitc(x+1,y,bit,0);
}

void generatedigits()
{
	int size,p,n,i,j,x,y;
	for (size=0;size<2;size++) for (p=0;p<MAXP;p++)
	{
		if (!size) for (n=0;n<10;n++) // Nice Numbers...
		{
			num[size][p][n]=(unsigned char *)malloc(2+10*15);
			readbit(num[size][p][n],".\\nums.dat",n*152);
		}
		else for (n=0;n<10;n++) // (x2 scale)
		{
			unsigned char *cur=num[size][p][n]=(unsigned char *)malloc(2+20*30);
			readbit(cur,".\\nums.dat",n*152);
			cur[0]=20; cur[1]=30;
			for (y=29;y>=0;y--)
				for (x=19;x>=0;x--)
					cur[2+x+y*20]=cur[2+x+y*10];
			for (x=19;x>=0;x--)
				for (y=29;y>=0;y--)
					cur[2+x+y*20]=cur[2+x/2+y/2*20];
		}
		if (p) for (n=0;n<10;n++) // Other coloured versions of the above
		{
			unsigned char *cur=num[size][p][n];
			for (i=2;i<2+cur[0]*cur[1];i++)
				if (cur[i]>=128 && cur[i]<255)
				{
					int r,g,b;
					j=cur[i]-128;
					r=j%5,g=j/5%5,b=j/25;
					if (p==1) cur[i]=128+g+b*5+r*25;
					else if (p==2) cur[i]=128+g+r*5+r*25;
					else if (p==3) cur[i]=128+r+r*5+b*25;
				}
				else cur[i]=255;
		}
	}
}

void deltmp(const int n)
{
	char *str=(char *)malloc(20),*str2=(char *)malloc(10);
	str=".\\";
	strcat(str,itoa(n,str2,10));
	strcat(str,".tmp");
	remove(str); // Delete the temporary file
}

void blackbox(const int x,const int y,const int xs,const int ys)
{
	if (xs<4 || ys<4) return;
	rect(x,y,xs,ys,0);
	box(x,y,xs-1,ys-1,20);
	box(x+1,y+1,xs-3,ys-3,18);
}

void strbound(char *s,const char c)
{
	int n;
	for (n=0;s[n];n++)
		if (s[n]>c) s[n]=c;
}

void numplot(const int x,const int y,const int s,const int p,int tl)
{
	int n;
	for (n=0;n<9 && tl || !n;n++)
	{
		blackborder(x-n*8*(s+1)-n/3*2*(s+1),y,num[s][p][(int)(tl%10)]);
		wrapbit(x-n*8*(s+1)-n/3*2*(s+1),y,num[s][p][(int)(tl%10)]);
		tl/=10;
	}
}

void dispscores(const int them,const long sc[4])
{	// Score display
	int p,tot=0,pos=0;
	for (p=0;p<MAXP;p++) tot+=them>>p&1;
	for (p=0;p<MAXP;p++) if (them>>p&1)
	{
		numplot(xres-21,pos*((tot<=2)?40:20)+1,(tot<=2),p,sc[p]);
		pos++;
	}
}

int startsubgame(const char c,const int playing)
{
	if (c=='N') return subgameN(playing);
	else if (c=='C') return subgameC(playing);
	else if (c=='G') return subgameG(playing);
	else if (c=='I') return subgameI(playing);
	return 1; // No effect if unknown letter
}

void titlescreen_fractal()
{
	int rx,ry;
	char str[100];
	FILE *in=NULL,*out;
	char floppy=(_getdrive()<=2); // Floopy drives A:1 and B:2
	in=fopen(".\\tsfract.cgf","rb");
	if (!in)
	{
		for (ry=40;ry<440;ry++)
		{
			bwrite(100,100,xres,"Generating title-screen fractal...",200);
			sprintf(str,"%d%%",(ry-40)/4);
			bwrite(100,120,xres,str,160);
			ccl(0);
			for (rx=0;rx<xres;rx++)
			{
				double oxr,cr=0.3,ci=0.017,xr=(double)rx/xres,xi=(double)ry/xres;
				float p=(float)rx/xres,q=1.0-p,r;
				int i;
				for (i=0;i<127 && xr*xr+xi*xi<1e6;i++)
				{
					oxr=xr;
					xr=xr*xr-xi*xi+cr;
					xi=oxr*xi*2+ci;
				}
				r=i+128;
				r-=log(log(xr*xr+xi*xi))/log(2);
				if (r>255) r=255;
				putpixb(rx,ry,p*r+q*160+frnd(1));
			}
		}
		if (!floppy)
		{
			out=fopen(".\\tsfract.cgf","wb");
			fwrite(back+40*xres,400*xres,1,out);
			fclose(out);
		}
	}
	else
	{
		fread(back+40*xres,400*xres,1,in);
		fclose(in);
	}
}

void loadprefs()
{
	int p;
	FILE *in=fopen(".\\prefs.dat","rb");
	fread(&gamma,4,1,in); fread(&autosave,4,1,in);
	for (p=0;p<MAXP;p++)
	{
		leftkey[p]=fgetc(in);
		rightkey[p]=fgetc(in);
		upkey[p]=fgetc(in);
		downkey[p]=fgetc(in);
		firekey[p]=fgetc(in);
		selectkey[p]=fgetc(in);
	}
	fclose(in);
	writepal(gamma);
}

void saveprefs()
{
	int p;
	FILE *out=fopen(".\\prefs.dat","wb");
	fwrite(&gamma,4,1,out); fwrite(&autosave,4,1,out);
	for (p=0;p<MAXP;p++)
	{
		fputc(leftkey[p],out);
		fputc(rightkey[p],out);
		fputc(upkey[p],out);
		fputc(downkey[p],out);
		fputc(firekey[p],out);
		fputc(selectkey[p],out);
	}
	fclose(out);
}

void adddot(unsigned char *bitmap,const int x,const int y)
{
	int xs=bitmap[0],i=2+x+y*xs;
	bitmap[i++]=6; bitmap[i++]=14; bitmap[i++]=14; bitmap[i]=6; i+=xs-3;
	bitmap[i++]=14; bitmap[i++]=15; bitmap[i++]=15; bitmap[i]=14; i+=xs-3;
	bitmap[i++]=14; bitmap[i++]=15; bitmap[i++]=15; bitmap[i]=14; i+=xs-3;
	bitmap[i++]=6; bitmap[i++]=14; bitmap[i++]=14; bitmap[i]=6;
}

void loadblockset(unsigned char **blox,const int level)
{
	int n,i,x,y;
	FILE *in;
	for (n=0;n<0x100;n++)
	{
		blox[n][0]=blox[n][1]=20;
		for (i=2;i<402;i++) blox[n][i]=255;
	}
	readbit(blox[0x40],".\\ncgi.dat",6423); // Shop symbol
	readbit(blox[0x41],".\\ncgi.dat",6825); // Upward magnet
	for (n=0x42;n<=0x44;n++) // The others (right, down, left)
		for (x=0;x<20;x++) for (y=0;y<20;y++)
			blox[n][2+x+y*20]=blox[n-1][2+y+(19-x)*20];
	readbit(blox[0x45],".\\ncgi.dat",7227); // RHS turret
	for (n=0x46;n<=0x48;n++) // The others (down, left, up)
		for (x=0;x<20;x++) for (y=0;y<20;y++)
			blox[n][2+x+y*20]=blox[n-1][2+y+(19-x)*20];
	readbit(blox[0x49],".\\ncgi.dat",8739); // Upwards MG turret
	for (n=0x4A;n<=0x4C;n++) // The others (right, down, left)
		for (x=0;x<20;x++) for (y=0;y<20;y++)
			blox[n][2+x+y*20]=blox[n-1][2+y+(19-x)*20];
	for (n=0x4D;n<=0x4F;n++) // Fragile walls
		readbit(blox[n],".\\ncgi.dat",9543);
	for (i=2;i<402;i++)
	{
		blox[0x4D][i]=(blox[0x4D][i]-128)/30*26+128;
		blox[0x4F][i]=(blox[0x4F][i]-128)/30*31+128;
	}
	for (n=0x50;n<=0x52;n++) // WKobo-esque things
		readbit(blox[n],".\\ncgi.dat",11334);
	for (x=0;x<20;x++) for (y=0;y<20;y++)
	{
		if (x%5==1 || x%5==3 || y%5==1 || y%5==3) blox[0x53][2+x+y*20]=20-rnd(2);
		if (x%5==2 || y%5==2) blox[0x53][2+x+y*20]=24-rnd(4);
	}
	readbit(blox[0x54],".\\ncgi.dat",11882); // Upwards spark turret
	for (n=0x55;n<=0x57;n++) // The others (right, down, left)
		for (x=0;x<20;x++) for (y=0;y<20;y++)
			blox[n][2+x+y*20]=blox[n-1][2+y+(19-x)*20];
	for (x=0;x<20;x++) for (y=0;y<20;y++) // A blank pacman space
	{
		unsigned char c=0;
		if (x==0 || x==19 || y==0 || y==19) c=9;
		else if (x==1 || x==18 || y==1 || y==18) c=1;
		for (n=0xD0;n<=0xD8;n++) blox[n][2+x+y*20]=c;
	}
	adddot(blox[0xD1],8,8); // Add some dots
	adddot(blox[0xD2],3,3); adddot(blox[0xD2],13,13);
	adddot(blox[0xD3],3,3); adddot(blox[0xD3],8,8); adddot(blox[0xD3],13,13);
	adddot(blox[0xD4],3,3); adddot(blox[0xD4],3,13);
	adddot(blox[0xD4],13,3); adddot(blox[0xD4],13,13);
	adddot(blox[0xD5],3,3); adddot(blox[0xD5],3,13); adddot(blox[0xD5],8,8);
	adddot(blox[0xD5],13,3); adddot(blox[0xD5],13,13);
	adddot(blox[0xD6],3,3); adddot(blox[0xD6],8,3); adddot(blox[0xD6],13,3);
	adddot(blox[0xD6],3,13); adddot(blox[0xD6],8,13); adddot(blox[0xD6],13,13);
	blox[0xD7][2+9+9*20]=15;
	{ // Add a chilli pepper
		unsigned char chilli[146];
		readbit(chilli,".\\ncgi.dat",11736);
		for (x=0;x<12;x++) for (y=0;y<12;y++)
		{
			i=chilli[2+x+y*12];
			if (i!=255) blox[0xD8][2+(x+4)+(y+4)*20]=i;
		}
	}
	for (x=0;x<20;x++) for (y=0;y<20;y++) // Ghost source and walls and scaling block
	{
		i=2+x+y*20;
		blox[0xDE][i]=24+rnd(4);
		blox[0xDF][i]=233+rnd(3)*5+rnd(3);
		blox[0xFA][i]=blox[0xFB][i]=0;
	}
	in=fopen(".\\levs.dat","rb");
	fseek(in,(level-1)*1024+768,SEEK_SET);
	for (n=1;n<=0x3F;n++) // Block set
	{
		i=fgetc(in)%(numblox+1);
		if (i) readbit(blox[n],".\\blox.dat",(i-1)*402);
	}
	for (n=0xC0;n<=0xCF;n++)
	{
		i=fgetc(in)%(numblox+1);
		if (i) readbit(blox[n],".\\blox.dat",(i-1)*402);
	}
	for (n=0xFC;n<=0xFF;n++)
	{
		i=fgetc(in)%(numblox+1);
		if (i) readbit(blox[n],".\\blox.dat",(i-1)*402);
	}
	fclose(in);
}

void calcmagfield(const unsigned char grid[32][24],unsigned char **mf)
{
	int n,x,y,u,v;
	unsigned char st[256]; // 0x0C to 0x0F and 0x2C to 0x2F stop magnets
	for (n=0;n<256;n++) st[n]=(n%16>=0xC && n%16<=0xF && sq(n/16-1)==1);
	for (x=0;x<32;x++)
		for (y=0;y<24;y++)
		{
			int xf,yf; xf=yf=0;
			for (u=x;u>=0 && !st[grid[u][y]];u--) // Rightward-pointing magnets
				if (grid[u][y]==0x42) xf--;
			for (u=x;u<32 && !st[grid[u][y]];u++) // Leftward-pointing magnets
				if (grid[u][y]==0x44) xf++;
			for (v=y;v>=0 && !st[grid[x][v]];v--) // Downward-pointing magnets
				if (grid[x][v]==0x43) yf--;
			for (v=y;v<32 && !st[grid[x][v]];v++) // Upward-pointing magnets
				if (grid[x][v]==0x41) yf++;
			if (xf<-7) xf=-7;
			else if (xf>7) xf=7;
			if (yf<-7) yf=-7;
			else if (yf>7) yf=7;
			n=xf+7+(yf+7)*16;
			mf[x][y]=n;
		}
}

double bglag=0,bgstart;
int cludgetastic[32][32];

void plotbgsq(const int x,const int y,const int bg,const int bgsub,const int bplot)
{
	int xx,yy;
	for (xx=x*20;xx<x*20+20;xx++)
		for (yy=y*20;yy<y*20+20;yy++)
			putpixb(xx,yy,bcalc(xx,yy,bg,bgsub,0.001*GetTickCount()-
				bglag*(float)((1024+bplot-cludgetastic[xx%32][yy%32])%1024)/1024.0));
}

void coins_initialisation(unsigned char ****addrcoin)
{
	int n,i;
	unsigned char ***coin=*addrcoin=(unsigned char ***)malloc(9*sizeof(unsigned char **));
	for (n=0;n<3;n++) // Smallest coins
	{
		coin[n]=(unsigned char **)malloc(4*sizeof(unsigned char *));
		readani(coin[n],4,".\\ncgi.dat",5913);
	}
	for (i=0;i<4;i++)
		for (n=2;n<2+5*5;n++)
		{
			if (coin[1][i][n]==136) coin[1][i][n]=24;
			else if (coin[1][i][n]==142) coin[1][i][n]=28;
			else if (coin[1][i][n]==14) coin[1][i][n]=30;
		}
	for (i=0;i<4;i++)
		for (n=2;n<2+5*5;n++)
		{
			if (coin[2][i][n]==136) coin[2][i][n]=140;
			else if (coin[2][i][n]==142) coin[2][i][n]=146;
			else if (coin[2][i][n]==14) coin[2][i][n]=202;
		}
	for (n=3;n<6;n++) // Middlest coins
	{
		coin[n]=(unsigned char **)malloc(4*sizeof(unsigned char *));
		readani(coin[n],3,".\\ncgi.dat",8433);
		coin[n][3]=(unsigned char *)malloc(2+10*10);
		coin[n][3][0]=coin[n][3][1]=10;
		for (i=0;i<100;i++)
			coin[n][3][2+i]=coin[n][1][2+i/10*10+9-i%10];
	}
	for (i=0;i<4;i++)
		for (n=2;n<2+10*10;n++)
		{
			if (coin[4][i][n]==130) coin[4][i][n]=24;
			else if (coin[4][i][n]==136) coin[4][i][n]=28;
			else if (coin[4][i][n]==142) coin[4][i][n]=30;
		}
	for (i=0;i<4;i++)
		for (n=2;n<2+10*10;n++)
		{
			if (coin[5][i][n]==130) coin[5][i][n]=140;
			else if (coin[5][i][n]==136) coin[5][i][n]=146;
			else if (coin[5][i][n]==142) coin[5][i][n]=202;
		}
	for (n=6;n<9;n++) // Largest coins
	{
		coin[n]=(unsigned char **)malloc(6*sizeof(unsigned char *));
		readani(coin[n],6,".\\ncgi.dat",1750);
	}
	for (i=0;i<6;i++)
		for (n=2;n<2+15*15;n++)
		{
			if (coin[7][i][n]==130) coin[7][i][n]=24;
			else if (coin[7][i][n]==136) coin[7][i][n]=28;
			else if (coin[7][i][n]==142) coin[7][i][n]=30;
		}
	for (i=0;i<6;i++)
		for (n=2;n<2+15*15;n++)
		{
			if (coin[8][i][n]==130) coin[8][i][n]=140;
			else if (coin[8][i][n]==136) coin[8][i][n]=146;
			else if (coin[8][i][n]==142) coin[8][i][n]=202;
		}
}

char *ncgitime(double t,char *s)
{
	t=t/86400-10773;
	sprintf(s,"%d %c%c%c",(int)floor(t),
		(t-=floor(t),t*=26,(char)t+'A'),
		(t-=floor(t),t*=26,(char)t+'A'),
		(t-=floor(t),t*=26,(char)t+'A'));
	return s;
}

void ncgisound(const int x,const unsigned char *sound,
	const int length,const int frequency,const char priority)
{
	int i;
	if (sound_working)
	{
		i=findfree(priority);
		stopsound(soundchannel[i]); killbuffer(soundchannel[i]);
		soundchannel[i]=makebuffer(length,frequency);
		writesound(sound,soundchannel[i],length);
		IDirectSoundBuffer_SetPan(soundchannel[i],(x-xres/2)*STEREO);
		playsound(soundchannel[i]);
	}
}

void loading(const char *s)
{
	bwrite(xres/2-estblength(s)/2,yres/2-4,999,s,1);
	ccl(0);
}

////////////////////////////// UTILITIES //////////////////////////////

void leveled()
{
	int n,i,x,y;
	char str[100];
	FILE *lev;
	int omb=1,olevel,level=1,lchange=2;
	int copyfrom=100;
	unsigned char grid[32][24];
	unsigned char **blox=(unsigned char **)malloc(0x100*sizeof(unsigned char *));
	int bg,bgsub,aset;
	int bxtable[0x100];
	unsigned char *ast[20],asz[21];
	int bplot=0,bplotsize=32;
	unsigned char *bplotx,*bploty;
	char leave=0;
	int ok,k=0,cb=0x01;
	unsigned char *save=(unsigned char *)malloc(400*xres);
	memcpy(save,back+40*xres,400*xres);
	// Background plotting order
	bplotx=(unsigned char *)malloc(bplotsize*bplotsize);
	bploty=(unsigned char *)malloc(bplotsize*bplotsize);
	for (n=0;n<bplotsize*bplotsize;n++)
	{
		char c=1;
		while (c)
		{
			bplotx[n]=rnd(bplotsize); bploty[n]=rnd(bplotsize);
			c=0;
			for (i=0;i<n;i++)
				if (bplotx[i]==bplotx[n])
					if (bploty[i]==bploty[n]) c=1;
		}
	}
	for (n=0;n<20;n++) ast[n]=NULL;
	for (n=0;n<0x100;n++) blox[n]=(unsigned char *)malloc(402);
	writepal(gamma);
	while (!leave && !dealwithmessages())
	{
		unsigned long framestart=GetTickCount();
		olevel=level;
		if (KEY(VK_ESCAPE) || KEY(VK_F12)) leave=1;
		else if (KEY(VK_Z)) {level=(level+98)%100+1; lchange=1;}
		else if (KEY(VK_X)) {level=level%100+1; lchange=1;}
		else if (KEY(VK_C)) copyfrom=level;
		else if (KEY(VK_V))
		{
			unsigned char *levdata=(unsigned char *)malloc(1024);
			lev=fopen(".\\levs.dat","r+b");
			fseek(lev,(copyfrom-1)*1024,SEEK_SET);
			fread(levdata,1024,1,lev);
			fseek(lev,(level-1)*1024,SEEK_SET);
			fwrite(levdata,1024,1,lev);
			fclose(lev); free(levdata);
			lchange=2;
		}
		else if (KEY(VK_LEFT))
			for (y=0;y<24;y++)
			{
				i=grid[0][y];
				for (x=0;x<31;x++)
				{
					grid[x][y]=grid[x+1][y];
					wrapbitb(x*20,y*20,blox[grid[x][y]]);
				}
				grid[x][y]=i;
				wrapbitb(x*20,y*20,blox[grid[x][y]]);
			}
		else if (KEY(VK_RIGHT))
			for (y=0;y<24;y++)
			{
				i=grid[31][y];
				for (x=31;x>0;x--)
				{
					grid[x][y]=grid[x-1][y];
					wrapbitb(x*20,y*20,blox[grid[x][y]]);
				}
				grid[x][y]=i;
				wrapbitb(x*20,y*20,blox[grid[x][y]]);
			}
		else if (KEY(VK_UP))
			for (x=0;x<32;x++)
			{
				i=grid[x][0];
				for (y=0;y<23;y++)
				{
					grid[x][y]=grid[x][y+1];
					wrapbitb(x*20,y*20,blox[grid[x][y]]);
				}
				grid[x][y]=i;
				wrapbitb(x*20,y*20,blox[grid[x][y]]);
			}
		else if (KEY(VK_DOWN))
			for (x=0;x<32;x++)
			{
				i=grid[x][23];
				for (y=23;y>0;y--)
				{
					grid[x][y]=grid[x][y-1];
					wrapbitb(x*20,y*20,blox[grid[x][y]]);
				}
				grid[x][y]=i;
				wrapbitb(x*20,y*20,blox[grid[x][y]]);
			}
		else if (KEY(VK_CONTROL))
		{
			mouseoff();
			while (!kbhit && !dealwithmessages()) x=getch-'0';
			if (x<0) x=0; if (x>9) x=9;
			mouseoff();
			while (!kbhit && !dealwithmessages()) y=getch-'0';
			if (y<0) y=0; if (y>9) y=9;
			mouseoff();
			level=x*10+y;
			if (level==0) level=100;
			lchange=1;
		}
		if (lchange)
		{
			if (lchange==1) // Save old level
			{
				lev=fopen(".\\levs.dat","r+b");
				fseek(lev,(olevel-1)*1024,SEEK_SET);
				for (x=0;x<32;x++) // Save map
					for (y=0;y<24;y++)
						fputc(grid[x][y],lev);
				for (n=1;n<=0x3F;n++) fputc(bxtable[n],lev);
				for (n=0xC0;n<=0xCF;n++) fputc(bxtable[n],lev);
				for (n=0xFC;n<=0xFF;n++) fputc(bxtable[n],lev);
				fputc(bg%numbg,lev); fputc(bgsub,lev);
				fputc(aset%numast,lev);
				fclose(lev);
			}
			lev=fopen(".\\levs.dat","rb"); // Load new level
			fseek(lev,(level-1)*1024,SEEK_SET);
			for (x=0;x<32;x++) // Load map
				for (y=0;y<24;y++)
					grid[x][y]=fgetc(lev);
			for (n=1;n<=0x3F;n++) bxtable[n]=fgetc(lev)%(numblox+1); // Block numbers
			for (n=0xC0;n<=0xCF;n++) bxtable[n]=fgetc(lev)%(numblox+1);
			for (n=0xFC;n<=0xFF;n++) bxtable[n]=fgetc(lev)%(numblox+1);
			bg=fgetc(lev)%numbg; bgsub=fgetc(lev);
			aset=fgetc(lev)%numast;
			fclose(lev);
			loadblockset(blox,level);
			for (x=0;x<32;x++) // Plot the grid
				for (y=0;y<24;y++)
					wrapbitb(x*20,y*20,blox[grid[x][y]]);
			for (n=0;n<20;n++) if (ast[n]) free(ast[n]);
			makeasteroidset(ast,asz,aset);
			cclb(); // Plot it.
			lchange=0;
		}
		for (x=0;x<32;x++)
			for (y=0;y<24;y++)
				if ((n=grid[x][y])>=0x85) if (n<=0x8D)
				{
					n=islegalsize(1<<(n-0x85));
					wrapbit(x*20+10-asz[n]/2,y*20+10-asz[n]/2,ast[n-1]);
				}
		if (level>=10000) wrapbitc(32,0,num[0][3][level/10000%10],15);
		if (level>=1000) wrapbitc(40,0,num[0][3][level/1000%10],15);
		if (level>=100) wrapbitc(48,0,num[0][3][level/100%10],15);
		if (level>=10) wrapbitc(56,0,num[0][3][level/10%10],15);
		wrapbitc(64,0,num[0][3][level%10],15);
		for (x=0;x<32;x++)
			for (y=0;y<24;y++)
			{
				unsigned char col=15;
				i=grid[x][y];
				if (i/64==2) col=12;
				else if (i>=0xFC) col=9;
				else if (i/64==1) col=13;
				else if (i/64==3) col=10;
				else if (i>>4==2) col=8;
				if (col!=15) box(x*20,y*20,19,19,col);
			}
		x=mx/20; y=my/20; i=grid[x][y];
		{
			unsigned char col=15;
			if (i/64==2) col=12;
			else if (i>=0xFC) col=9;
			else if (i/64==1) col=13;
			else if (i/64==3) col=10;
			box(x*20,y*20,19,19,col);
			putpix(mx+1,my,15); // Dot for mouse
			putpix(mx-1,my,15);
			putpix(mx,my+1,15);
			putpix(mx,my-1,15);
		}
		sprintf(str,"%02x",grid[x][y]);
		write(x*20+2,y*20+2,xres,str,15,7);
		cclb();
		ok=k; k=1;
		if (KEY(VK_2)) {cb=grid[x][y]=(grid[x][y]+0xF0*ok)%256; k=0;}
		if (KEY(VK_4)) {cb=grid[x][y]=(grid[x][y]+0xFF*ok)%256; k=0;}
		if (KEY(VK_6)) {cb=grid[x][y]=(grid[x][y]+0x01*ok)%256; k=0;}
		if (KEY(VK_8)) {cb=grid[x][y]=(grid[x][y]+0x10*ok)%256; k=0;}
		if (KEY(VK_RETURN)) grid[x][y]=cb;
		if (KEY(VK_SHIFT)) cb=grid[x][y];
		if (KEY(VK_SPACE)) grid[x][y]=0;
		wrapbitb(x*20,y*20,blox[grid[x][y]]);
		if (mb && !omb) // Mouse click
		{
			if (i>=1 && i<=0x3F || i/16==0xC || i>=0xFC) // Change block
			{
				if (mb&1) bxtable[i]+=numblox+2;
				if (mb&2) bxtable[i]+=numblox;
				bxtable[i]%=numblox+1;
				lchange=1;
			}
			else if (i==0) // Change background
			{
				if (mb&1) bg++;
				if (mb&2) bg--;
				bg=(bg+numbg)%numbg;
				lchange=1;
			}
			else if (i>=0x85 && i<=0x8D) // Change asteroids
			{
				if (mb&1) aset++;
				if (mb&2) aset--;
				aset=(aset+numast)%numast;
				lchange=1;
			}
		}
		omb=mb;
		for (n=0;n<10 || GetTickCount()-framestart<mswait;n++) // Synchronize end of frame
		{
			for (x=bplotx[bplot];x<xres;x+=bplotsize)
				for (y=bploty[bplot];y<yres;y+=bplotsize)
					if (blox[grid[x/20][y/20]][2+x%20+y%20*20]==255)
						putpixb(x,y,bcalc(x,y,bg,bgsub,0.001*GetTickCount()));
			bplot=(bplot+1)%(bplotsize*bplotsize);
		}
	}
	// Save level when leaving
	lev=fopen(".\\levs.dat","r+b");
	fseek(lev,(olevel-1)*1024,SEEK_SET);
	for (x=0;x<32;x++) // Save map
		for (y=0;y<24;y++)
			fputc(grid[x][y],lev);
	for (n=1;n<=0x3F;n++) fputc(bxtable[n],lev);
	for (n=0xC0;n<=0xCF;n++) fputc(bxtable[n],lev);
	for (n=0xFC;n<=0xFF;n++) fputc(bxtable[n],lev);
	fputc(bg%numbg,lev); fputc(bgsub,lev);
	fputc(aset%numast,lev);
	fclose(lev);
	// Uninitialise (finalise?)
	titlepal(gamma);
	memset(back,0,40*xres); memset(back+440*xres,0,40*xres);
	memcpy(back+40*xres,save,400*xres);
	free(save);
	for (n=0;n<0x100;n++) free(blox[n]); free(blox);
	mouseoff(); KEYUP(VK_ESCAPE);
}

////////////////////////////// TITLE SUBSCREENS //////////////////////////////

void instructions()
{
	int n;
	char str[100],leave=0;
	unsigned char **boni=(unsigned char **)malloc(itypes*sizeof(unsigned char *));
	char **desc=(char **)malloc(itypes*sizeof(unsigned char *));
	unsigned char *savedbg=(unsigned char *)malloc(xres*400);
	unsigned char *pointer=(unsigned char *)malloc(146);
	memcpy(savedbg,back+xres*40,xres*400);
	memcpy(back,shad,xres*yres);
	readbit(pointer,".\\ncgi.dat",1400);
	for (n=0;n<itypes;n++)
	{
		boni[n]=(unsigned char *)malloc(102);
		readbit(boni[n],".\\boni.dat",n*102);
		desc[n]=(char *)malloc(50);
	}
	desc[0]="Flame thrower";
	desc[1]="+3 plasma balls";
	desc[2]="White-hot fire";
	desc[3]="Spread-fire";
	desc[4]="Attract all other boni";
	desc[5]="+150 points";
	desc[6]="Faster bullets";
	desc[7]="+1 extra life";
	desc[8]="Stream-fire";
	desc[9]="Retro-thrusters";
	desc[10]="Machine-gun and bullets";
	desc[11]="Surround-fire";
	desc[12]="Invincibility";
	desc[13]="Longer-range bullets";
	mouseoff();
	while (!leave)
	{
		line(150,217,320,217,0);
		line(320,70,320,410,0);
		line(320,70,330,70,0);
		line(320,410,330,410,0);
		write(350,80,xres,"Objective",14,7);
		write(360,90,590,"To score as many points as possible and to progress through \
levels.  In the main game, you will normally need to shoot almost everything in \
sight to do this.  Other parts of the game are more subtle, but the same six keys \
are always used.",31,24);
		write(350,200,xres,"Display",14,7);
		write(360,210,590,"Players' scores are displayed in the top-right corner of \
the screen and the level number is in the top-left corner.  Next to the level number \
is a display of the number of lives you have remaining.",31,24);
		write(350,240,xres,"Key to types of boni",14,7);
		for (n=0;n<itypes;n++)
		{
			wrapbit(360+(n/10)*100,249+(n%10)*12,boni[n]);
			sprintf(str,"%s",desc[n]);
			write(375+(n/10)*100,251+(n%10)*12,xres-10,str,31,24);
		}
		write(350,400,xres,"Press Escape or click mouse to continue",15,7);
		mskbit(mx,my,pointer); // The mouse
		cclb();
		leave|=mb|KEY(VK_ESCAPE)|KEY(VK_F12)|dealwithmessages();
	}
	mouseoff();
	memset(back,0,40*xres);
	memcpy(back+40*xres,savedbg,400*xres); free(savedbg);
	memset(back+440*xres,0,40*xres);
	for (n=0;n<itypes;n++)
	{
		free(desc[n]);
		free(boni[n]);
	}
	free(desc); free(boni);
	free(pointer);
}

void credits()
{
	int n,i; FILE *in;
	float *x,*y,*z,*b; int snum=1000;
	char *text,**line; int maxtext=50000,maxlines=5000,numlines;
	float ypos,dt=1;
	char leave=0;
	x=(float *)malloc(snum*sizeof(float)); y=(float *)malloc(snum*sizeof(float));
	z=(float *)malloc(snum*sizeof(float)); b=(float *)malloc(snum*sizeof(float));
	if (!b) exit(0);
	line=(char **)malloc(maxlines*sizeof(char *)); text=(char *)malloc(maxtext);
	in=fopen(".\\c2.dat","rb");
	line[0]=&text[0]; i=1;
	for (n=0;!feof(in);n++)
	{
		text[n]=fgetc(in)^scrambfunc(n);
		if (text[n]==13) {text[n]=0; line[i]=&text[n]+1; i++;}
	}
	numlines=i;
	fclose(in);
	for (n=0;n<snum;n++) z[n]=-1;
	mouseoff();
	for (ypos=-200;!leave;ypos+=0.25*dt)
	{
		long framestart=GetTickCount();
		for (n=0;n<snum;n++)
		{
			int c; float px,py;
			z[n]-=0.15;
			if (z[n]<0 || fabs(x[n]/z[n])>0.5 || fabs(y[n]/z[n])>0.5*yres/xres)
			{
				b[n]=pow(100,frnd(1));
				z[n]=b[n]*10;
				x[n]=frnd(z[n])-z[n]/2;
				y[n]=frnd(z[n])-z[n]/2;
			}
			c=31.5-1.5*z[n]/b[n];
			px=x[n]/z[n]*xres+xres/2; py=y[n]/z[n]*xres+yres/2;
			if (c>31) c=31; if (c<16) c=16;
			wupix(px,py,c);
		}
		for (n=0;n<numlines-1;n++)
		{
			int y=n*20-ypos;
			if (y>=0 && y<yres-8)
			{
				float k=1.0-(1.0-sin(M_PI*(y+2)/yres))*1.5;
				if (k>0) bwrite(xres/2-estblength(line[n])*0.5,y,xres,line[n],16.0+k*15.9);
			}
		}
		ccl(0);
		do leave|=dealwithmessages()|kbhit|mb;
		while (GetTickCount()<framestart+mswait); // Synchronize frame
		dt=(GetTickCount()-framestart)/mswait;
	}
	free(x); free(y); free(z); free(b);
	while (!kbhit && mb && !dealwithmessages());
	mouseoff();
}

void options()
{
	int n,cp=0;
	unsigned char *pointer=(unsigned char *)malloc(146);
	char **keydesc=(char **)malloc(256*sizeof(char *));
	unsigned char *ship[MAXP];
	char numb[40],*str=(char *)malloc(100),leave=0;
	memcpy(back,shad,xres*yres);
	mouseoff();
	readbit(pointer,".\\ncgi.dat",1400);
	for (n=0;n<MAXP;n++) ship[n]=(unsigned char *)malloc(258);
	readbit(ship[0],".\\ncgi.dat",0);
	readbit(ship[1],".\\ncgi.dat",4914);
	readbit(ship[2],".\\ncgi.dat",5172);
	readbit(ship[3],".\\ncgi.dat",5430);
	for (n=0;n<256;n++)
	{
		keydesc[n]=(char *)malloc(25);
		sprintf(str,"Unknown key %d",n);
		strcpy(keydesc[n],str);
	}
	keydesc[VK_A]="A"; keydesc[VK_B]="B";
	keydesc[VK_C]="C"; keydesc[VK_D]="D";
	keydesc[VK_E]="E"; keydesc[VK_F]="F";
	keydesc[VK_G]="G"; keydesc[VK_H]="H";
	keydesc[VK_I]="I"; keydesc[VK_J]="J";
	keydesc[VK_K]="K"; keydesc[VK_L]="L";
	keydesc[VK_M]="M"; keydesc[VK_N]="N";
	keydesc[VK_O]="O"; keydesc[VK_P]="P";
	keydesc[VK_Q]="Q"; keydesc[VK_R]="R";
	keydesc[VK_S]="S"; keydesc[VK_T]="T";
	keydesc[VK_U]="U"; keydesc[VK_V]="V";
	keydesc[VK_W]="W"; keydesc[VK_X]="X";
	keydesc[VK_Y]="Y"; keydesc[VK_Z]="Z";
	keydesc[VK_UP]="Up arrow";
	keydesc[VK_DOWN]="Down arrow";
	keydesc[VK_LEFT]="Left arrow";
	keydesc[VK_RIGHT]="Right arrow";
	keydesc[VK_RETURN]="Enter/Return";
	keydesc[VK_SHIFT]="Shift";
	keydesc[VK_CONTROL]="Control";
	keydesc[VK_TAB]="Tab";
	for (n=0;n<10;n++)
	{
		sprintf(keydesc[VK_NUMPAD0+n],"Keypad %d",n);
		sprintf(keydesc[VK_0+n],"%d",n);
	}
	keydesc[VK_DECIMAL]="Keypad decimal point";
	keydesc[VK_LBUTTON]="Left mouse button";
	keydesc[VK_RBUTTON]="Right mouse button";
	keydesc[VK_MBUTTON]="Middle mouse button";
	keydesc[VK_SPACE]="Space barr";
	keydesc[VK_MENU]="Alt";
	while (!leave)
	{
		long framestart=GetTickCount();
		char on=0,over;
		float rx=mx,ry=my; int rb=mb|KEY(VK_RETURN)|KEY(VK_SPACE);
		line(105,187,200,187,0);
		line(200,70,200,410,0);
		line(200,70,210,70,0);
		line(200,410,210,410,0);
		on=(rx>=240 && ry>=75 && ry<=92);
		bwrite(240,80,xres,"Gamma Correction:",(on)?0:15);
		sprintf(numb,"%.2f",gamma);
		bwrite(440,80,xres,numb,15);
		on=(rx>=420 && rx<428 && ry>=75 && ry<=92);
		bwrite(420,80,xres,"<",(on)?0:15);
		if (on && rb && gamma>0.1) // Gamma -
		{
			gamma/=1.01; if (gamma<0.1) gamma=0.1;
			titlepal(gamma);
		}
		on=(rx>=490 && rx<498 && ry>=75 && ry<=92);
		bwrite(490,80,xres,">",(on)?0:15);
		if (on && rb && gamma<10) // Gamma +
		{
			gamma*=1.01; gamma+=0.001; if (gamma>10) gamma=10;
			titlepal(gamma);
		}
		on=(rx>=240 && ry>=115 && ry<=132);
		sprintf(numb,"Players: %d",pm);
		bwrite(240,120,xres,numb,(on)?0:15);
		if (kbhit && getch>='1' && getch<='4') pm=getch-'0';
		if (on) // Players
		{
			if (rb&1) pm++;
			if (rb&2) pm--;
			if (rb&3) mouseoff();
		}
		if (pm<1) pm=MAXP;
		if (pm>MAXP) pm=1;
		on=(rx>=240 && ry>=155 && ry<=172);
		sprintf(numb,"Autosave period: %d",(int)((autosave<60)?autosave:autosave/60));
		if (autosave<60) strcat(numb," sec");
			else strcat(numb," min");
		bwrite(240,160,xres,numb,(on)?0:15);
		if (on) // Autosave
		{
			if (rb&1)
			{
				if (autosave==1) autosave=3;
				else if (autosave==3) autosave=10;
				else if (autosave==10) autosave=30;
				else if (autosave==30) autosave=60;
				else if (autosave==60) autosave=180;
				else if (autosave==180) autosave=600;
				else autosave=1;
				mouseoff();
			}
			if (rb&2)
			{
				if (autosave==3) autosave=1;
				else if (autosave==10) autosave=3;
				else if (autosave==30) autosave=10;
				else if (autosave==60) autosave=30;
				else if (autosave==180) autosave=60;
				else if (autosave==600) autosave=180;
				else autosave=600;
				mouseoff();
			}
		}
		sprintf(str,"Keys for player %d",cp+1);
		on=(rx>=600 && ry>=300 && rx<616 && ry<316);
		bwrite(240,300,xres,str,(on)?0:15);
		strcpy(str,"[click spaceship to see other players]");
		write(240,310,xres,str,15,7);
		if (on)
		{
			wrapbitc(601,300,ship[cp],15);
			wrapbitc(599,300,ship[cp],15);
			wrapbitc(600,301,ship[cp],15);
			wrapbitc(600,299,ship[cp],15);
			wrapbitc(600,300,ship[cp],0);
		}
		else wrapbitc(600,300,ship[cp],15);

		sprintf(str,"Left:  %s",keydesc[leftkey[cp]]);
		over=(rx>=240 && (int)ry/10==33);
		bwrite(240,330,590,str,(over)?0:31);
		if (over) for (n=0;n<256;n++) if (KEY(n)) leftkey[cp]=n;
		sprintf(str,"Right:  %s",keydesc[rightkey[cp]]);
		over=(rx>=240 && (int)ry/10==34);
		bwrite(240,340,590,str,(over)?0:31);
		if (over) for (n=0;n<256;n++) if (KEY(n)) rightkey[cp]=n;
		sprintf(str,"Up:  %s",keydesc[upkey[cp]]);
		over=(rx>=240 && (int)ry/10==35);
		bwrite(240,350,590,str,(over)?0:31);
		if (over) for (n=0;n<256;n++) if (KEY(n)) upkey[cp]=n;
		sprintf(str,"Down:  %s",keydesc[downkey[cp]]);
		over=(rx>=240 && (int)ry/10==36);
		bwrite(240,360,590,str,(over)?0:31);
		if (over) for (n=0;n<256;n++) if (KEY(n)) downkey[cp]=n;
		sprintf(str,"Fire:  %s",keydesc[firekey[cp]]);
		over=(rx>=240 && (int)ry/10==37);
		bwrite(240,370,590,str,(over)?0:31);
		if (over) for (n=0;n<256;n++) if (KEY(n)) firekey[cp]=n;
		sprintf(str,"Select:  %s",keydesc[selectkey[cp]]);
		over=(rx>=240 && (int)ry/10==38);
		bwrite(240,380,590,str,(over)?0:31);
		if (over) for (n=0;n<256;n++) if (KEY(n)) selectkey[cp]=n;
		if (rx>=600 && ry>=300 && rx<616 && ry<316)
		{
			if (rb&2) cp--;
			if (cp<0) cp=MAXP-1;
			if (rb&1) cp++;
			if (cp>=MAXP) cp=0;
			mouseoff();
		}
		on=(rx>=240 && rx<=276 && ry>=395 && ry<=412 || rx<200);
		bwrite(240,400,xres,"Back",(on)?0:15);
		if (on && rb) leave=1;
		while (KEY(VK_ESCAPE) || KEY(VK_F12)) leave=1;
		mskbit(rx,ry,pointer); // The mouse
		cclb();
		do
		{
			leave|=dealwithmessages();
			leave|=KEY(VK_ESCAPE)|KEY(VK_F12);
		}
		while (GetTickCount()<framestart+mswait); // Synchronize frame
	}
	free(pointer);
	saveprefs();
	mouseoff();
	free(str);
	for (n=0;n<MAXP;n++) free(ship[n]);
	for (n=0;n<256;n++) free(keydesc[n]); free(keydesc);
}

void hiscores()
{
	char leave=0;
	char *str=(char *)malloc(MAXNAMELEN);
	float t;
	int n;
	char **names=(char **)malloc(HILEN*sizeof(char *)); long *scores=(long *)malloc(HILEN*sizeof(long));
	memcpy(back,shad,xres*yres);
	for (n=0;n<HILEN;n++) names[n]=(char *)malloc(MAXNAMELEN);
	gethiscores(names,scores);
	for (n=0;n<HILEN;n++) names[n][32]=0;
	while (kbhit || mb) dealwithmessages();
	for (t=0;!leave;t+=0.1)
	{
		long framestart=GetTickCount();
		line(137,157,200,157,0);
		line(200,70,200,410,0);
		line(200,70,210,70,0);
		line(200,410,210,410,0);
		for (n=0;n<HILEN;n++)
		{
			int d=sin(n+t)*5;
			sprintf(str,"%d.",n+1);
			bwrite(230+d,80+n*15,xres,str,15);
			sprintf(str,"%s",names[n]);
			bwrite(255+d,80+n*15,xres,str,53-n);
			sprintf(str,"%ld",scores[n]);
			bwrite(550+d,80+n*15,xres,str,31-12*n/HILEN);
		}
		write(350,400,xres,"Press any key or click mouse to continue",15,7);
		if (mb || kbhit) leave=1;
		cclb(); if (dealwithmessages()) gracefulexit();
		while (GetTickCount()<framestart+mswait); // Synchronize frame
	}
	mouseoff();
}

////////////////////////////// THE SHOP //////////////////////////////

void theshop(const int p,int *upgrade)
{
	int n,i,j,x,y,fr; char str[100];
	int things=27,maxmoney=money[p];
	float red[256],green[256],blue[256];
	unsigned char grey[256];
	float pos=-100,kx=0,ky=0,kxs=0,kys=0;
	char leave=0,**desc=(char **)malloc(things*sizeof(char *)),**pics=(char **)malloc(things*sizeof(char *));
	int *price=(int *)malloc(things*sizeof(int));
	gamestart-=0.001*GetTickCount();
	nextsave-=0.001*GetTickCount();
	for (n=0;n<things;n++)
	{
		desc[n]=(char *)malloc(50);
		pics[n]=(char *)malloc(402);
		pics[n][0]=pics[n][1]=20;
		if (n<itypes)
		{
			unsigned char *item=(unsigned char *)malloc(102);
			readbit(item,".\\boni.dat",n*102);
			for (x=0;x<20;x++)
				for (y=0;y<20;y++)
					pics[n][2+x+y*20]=item[2+x/2+y/2*10];
			free(item);
		}
		else if (n>=14 && n<=19) // Shield-related icons
		{
			readbit(pics[n],".\\ncgi.dat",6021);
			if (n==15) // Battery
			{
				for (y=10;y<=17;y++)
					for (x=11+2*(y==10);x<=17-2*(y==10);x++)
						pics[n][2+x+y*20]=28-2*abs(x-14);
			}
			else if (n==16) // Energy
			{
				pics[n][2+12+16*20]=pics[n][2+13+15*20]=pics[n][2+14+14*20]=
				pics[n][2+15+13*20]=pics[n][2+15+14*20]=pics[n][2+15+15*20]=
				pics[n][2+15+16*20]=pics[n][2+15+17*20]=pics[n][2+16+16*20]=
				pics[n][2+17+15*20]=pics[n][2+18+14*20]=11;
				pics[n][2+12+15*20]=pics[n][2+12+17*20]=pics[n][2+13+14*20]=
				pics[n][2+13+16*20]=pics[n][2+14+13*20]=pics[n][2+14+15*20]=
				pics[n][2+15+12*20]=pics[n][2+15+18*20]=pics[n][2+16+17*20]=
				pics[n][2+16+17*20]=pics[n][2+17+14*20]=pics[n][2+17+16*20]=
				pics[n][2+18+13*20]=pics[n][2+18+15*20]=9;
			}
			else if (n==17) // Auto-recharge unit
			{
				pics[n][2+11+13*20]=pics[n][2+11+14*20]=pics[n][2+11+15*20]=
				pics[n][2+12+13*20]=pics[n][2+12+15*20]=pics[n][2+13+11*20]=
				pics[n][2+13+12*20]=pics[n][2+13+13*20]=pics[n][2+13+14*20]=
				pics[n][2+13+15*20]=pics[n][2+13+16*20]=pics[n][2+13+17*20]=
				pics[n][2+14+11*20]=pics[n][2+14+17*20]=230;
				pics[n][2+15+13*20]=pics[n][2+15+14*20]=pics[n][2+15+15*20]=
				pics[n][2+16+14*20]=pics[n][2+17+14*20]=13;
			}
			else if (n==18) // Refraction shield
			{
				for (i=2;i<2+20*20;i++)
					if (pics[n][i]==15) pics[n][i]=11;
			}
			else if (n==19) // Destructor shield
				for (i=2;i<2+20*20;i++)
					if (pics[n][i]==15) pics[n][i]=12;
		}
		else if (n==20) readbit(pics[n],".\\ncgi.dat",7629); // Light sabre icon
		else if (n==21) // Anti-SRM pulse
		{
			unsigned char mine[146];
			readbit(mine,".\\ncgi.dat",667);
			for (x=0;x<20;x++)
				for (y=0;y<20;y++)
				{
					i=2+x+y*20;
					pics[n][i]=255;
					if (x>=4 && x<16 && y>=4 && y<16) pics[n][i]=mine[2+x-4+(y-4)*12];
					if (pics[n][i]==255 && (x+y)%2==0) pics[n][i]=12;
				}
		}
		else if (n==22) readbit(pics[n],".\\ncgi.dat",8031); // Laser icon
		else if (n==23) readbit(pics[n],".\\ncgi.dat",9141); // Prodder
		else if (n==24) // Sell life
		{
			unsigned char *item=(unsigned char *)malloc(102);
			readbit(item,".\\ncgi.dat",9945);
			for (x=0;x<20;x++)
				for (y=0;y<20;y++)
					pics[n][2+x+y*20]=item[2+x/2+y/2*10];
			free(item);
		}
		else if (n==25) // Inverse prodder
		{
			unsigned char *spr=(unsigned char *)malloc(402);
			readbit(spr,".\\ncgi.dat",9141);
			pics[n][0]=pics[n][1]=20;
			for (i=2;i<402;i++)
			{
				pics[n][i]=spr[403-i];
				if (pics[n][i]==3 || pics[n][i]==11) pics[n][i]+=2;
			}
			free(spr);
		}
		else if (n==26) // Flame mine
		{
			unsigned char fmine[258];
			readbit(fmine,".\\ncgi.dat",11076);
			for (x=0;x<20;x++)
				for (y=0;y<20;y++)
				{
					pics[n][i=2+x+y*20]=0;
					if (x>=2 && x<18 && y>=2 && y<18) pics[n][i]=fmine[2+x-2+(y-2)*16];
					if (pics[n][i]==255) pics[n][i]=0;
					if (x==0 || x==19 || y==0 || y==19) pics[n][i]=7;
				}
		}
	}
	strcpy(desc[0],"Flame thrower"); price[0]=3;
	strcpy(desc[1],"+3 plasma balls"); price[1]=3;
	strcpy(desc[2],"White-hot fire"); price[2]=5;
	strcpy(desc[3],"Spread-fire");
	strcpy(desc[4],"Attract all other boni"); price[4]=10;
	strcpy(desc[5],"+150 points"); price[5]=1;
	strcpy(desc[6],"Faster bullets"); price[6]=1;
	strcpy(desc[7],"+1 extra life"); price[7]=100;
	strcpy(desc[8],"Stream-fire");
	strcpy(desc[9],"Retro-thrusters"); price[9]=2;
	strcpy(desc[10],"Machine-gun and bullets"); price[10]=3;
	strcpy(desc[11],"Surround-fire");
	strcpy(desc[12],"Invincibility"); price[12]=5;
	strcpy(desc[13],"Longer-range bullets"); price[13]=2;
	strcpy(desc[14],"Shield emitter"); price[14]=20;
	strcpy(desc[15],"Shield capacitor"); price[15]=10;
	strcpy(desc[16],"Shield energy"); price[16]=1;
	strcpy(desc[17],"Shield auto-recharge unit"); price[17]=35;
	strcpy(desc[18],"Upgrade to refraction shields"); price[18]=25;
	strcpy(desc[19],"'Szurko' brand DESTRUCTOR (tm)\nshields"); price[19]=120;
	strcpy(desc[20],"Light sabre"); price[20]=80;
	strcpy(desc[21],"Anti-S.R.M. pulse"); price[21]=20;
	strcpy(desc[22],"Laser upgrade"); price[22]=5;
	strcpy(desc[23],"Prodder weapon"); price[23]=15;
	strcpy(desc[24],"Sell life"); price[24]=-Min(level,100);
	strcpy(desc[25],"Inverse Prodder"); price[25]=30;
	strcpy(desc[26],"+3 flame mines"); price[26]=10;
	palcoeffs(red,green,blue);
	for (n=0;n<256;n++)
	{
		float brightness=red[n]/3+green[n]/2+blue[n]/6;
		grey[n]=16+(int)(brightness*15.9);
	}
	for (x=0;x<xres;x++)
		for (y=0;y<yres;y++)
			putpixb(x,y,grey[getpix(x,y)]);
	cclb();
	for (fr=0;!leave;fr++)
	{
		double framestart=GetTickCount();
		int shown=0;
		float omx=mx,omy=my,rx=mx+kx,ry=my+ky,rb=mb|KEY(firekey[p]);
		price[8]=upgrade[3]+upgrade[4]; // Stream fire
		price[3]=floor(1.5*price[8]); // Spread fire
		price[11]=4+2*upgrade[8]; // Surround fire
		if (KEY(leftkey[p])) kxs--;
		if (KEY(rightkey[p])) kxs++;
		if (KEY(upkey[p])) kys--;
		if (KEY(downkey[p])) kys++;
		kx+=kxs; ky+=kys;
		kxs*=0.8; kys*=0.8;
		if (mx+kx<0) {kx=-mx; kxs=0;}
		if (my+ky<0) {ky=-my; kys=0;}
		if (mx+kx>xres) {kx=xres-mx; kxs=0;}
		if (my+ky>yres) {ky=yres-my; kys=0;}
		sprintf(str,"%d.%02d",money[p]/100,money[p]%100);
		bwrite(500,15,xres,str,14);
		sprintf(str,"the shop");
		strbound(str,Min(fr+96,127));
		bwrite(51,15,xres,str,0);
		bwrite(49,15,xres,str,0);
		bwrite(50,16,xres,str,0);
		bwrite(50,14,xres,str,0);
		bwrite(50,15,xres,str,15);
		n=(rx>=500 && ry>=450);
		bwrite(500,450,xres,"Exit",2+n*8);
		if (n && rb) leave=1;
		if (KEY(VK_ESCAPE) || KEY(VK_F12)) leave=1;
		for (n=j=0;n<things;n++) if (price[n]<=maxmoney)
		{
			int oy=j-pos;
			j+=30; shown++;
			if (oy>=50 && oy<450)
			{
				int avail=1;
				if (n==2 && upgrade[2]>=400) avail=0; // Limited white-fire
				else if (n==6 && upgrade[12]>=20000) avail=0; // Limited bullet speed
				else if (n==10 && upgrade[7]>=2500) avail=0; // Limited machine-gun bullets
				else if (n==12 && upgrade[9]>=1000) avail=0; // Limited invincibility
				else if (n==13 && upgrade[14]>=10000000) avail=0; // Limited bullet range
				else if (n==14 && upgrade[15]) avail=0; // Only one shield generator
				else if (n>=15 && n<=19 && !upgrade[15]) avail=0; // No upgrades to non-existant shields
				else if (n==16 && upgrade[17]>=upgrade[16]) avail=0; // Shield capacity
				else if (n==18 && upgrade[15]==11) avail=0; // Can't upgrade shields to themselves
				else if (n==19 && upgrade[15]==12) avail=0; // Same again
				else if (n==20 && upgrade[19]) avail=0; // Only one light sabre
				else if (n==22 && upgrade[21]>=8) avail=0; // Maximum white lasers
				else if (n==23 && upgrade[22]) avail=0; // One prodder
				else if (n==24 && lives[p]<=1) avail=0; // Can't sell last life
				else if (n==25 && upgrade[23]) avail=0; // One inverse prodder
				wrapbit(50,oy,pics[n]);
				sprintf(str,"%s",desc[n]);
				i=(rx<500 && ry>=oy && ry<=oy+20);
				bwrite(100,oy,380,str,2+i*8);
				if (price[n]>=0) sprintf(str,"%d.%02d",price[n]/100,price[n]%100);
					else sprintf(str,"-%d.%02d",-price[n]/100,-price[n]%100);
				bwrite(400,oy,480,str,9+i*(2+(price[n]>money[p])));
				if (i && price[n]<=money[p])
				{
					if (avail) box(48,oy-2,23,23,2+8*(fr/10%2));
					else for (x=0;x<20;x++) for (y=x%2;y<20;y+=2)
						putpix(50+x,oy+y,0);
				}
				if (i && price[n]<=money[p] && rb && avail) // Buy it
				{
					money[p]-=price[n];
					maxmoney=Max(maxmoney,money[p]);
					if (n==0) upgrade[0]++;
					else if (n==1) upgrade[1]+=3;
					else if (n==2) upgrade[2]=400;
					else if (n==3) upgrade[3]++;
					else if (n==4) upgrade[5]++;
					else if (n==5) score[p]+=150;
					else if (n==6)
					{
						upgrade[12]+=0.4*upgrade[12]+1;
						if (upgrade[12]>20000) upgrade[12]=20000;
					}
					else if (n==7) lives[p]++;
					else if (n==8) upgrade[4]++;
					else if (n==9) upgrade[6]++;
					else if (n==10)
					{
						upgrade[7]+=300;
						if (upgrade[7]>2500) upgrade[7]=2500;
					}
					else if (n==11) upgrade[8]++;
					else if (n==12) upgrade[9]=1000;
					else if (n==13)
					{
						upgrade[14]+=0.4*upgrade[14]+1;
						if (upgrade[14]>10000000) upgrade[14]=10000000;
					}
					else if (n==14) upgrade[15]=15;
					else if (n==15) upgrade[16]+=5;
					else if (n==16) upgrade[17]++;
					else if (n==17) upgrade[18]++;
					else if (n==18) upgrade[15]=11;
					else if (n==19) upgrade[15]=12;
					else if (n==20) upgrade[19]=1;
					else if (n==21) upgrade[20]++;
					else if (n==22) upgrade[21]++;
					else if (n==23) upgrade[22]=1;
					else if (n==24) lives[p]--;
					else if (n==25) upgrade[23]=1;
					else if (n==26) upgrade[24]+=3;
					else money[p]+=price[n]; // Refund if feature not implemented!
					mouseoff();
				}
			}
		}
		if (rx<500)
		{
			if (ry<50) pos-=0.02*sq(50-ry);
			if (ry>430) pos+=0.02*sq(ry-430);
			if (pos>-400+shown*30) pos=-400+shown*30;
			if (pos<-100) pos=-100;
		}
		for (n=0;n<8;n++)
		{
			float th=0.03*fr+M_PI*n/4;
			line(rx+10*cos(th),ry+10*sin(th),
					rx+15*cos(th),ry+15*sin(th),5);
			th*=-1;
			putpix(rx+12*cos(th),ry+12*sin(th),13);
			th*=2;
			putpix(rx+9*cos(th),ry+9*sin(th),15);
		}
		cclb();
		do leave|=dealwithmessages();
		while (GetTickCount()-framestart<mswait && !leave);
		if (omx!=mx || omy!=my) kx=ky=kxs=kys=0;
	}
	for (n=0;n<itypes;n++) {free(desc[n]); free(pics[n]);} free(desc); free(pics);
	mouseoff();
	gamestart+=0.001*GetTickCount();
	nextsave+=0.001*GetTickCount();
}

////////////////////////////// MAIN ROUTINE //////////////////////////////

int WINAPI WinMain(HINSTANCE hinstance,HINSTANCE previous,LPSTR cmdline,int showcmd)
{
	double CSTART=_rdtsc(),TSTART=0.001*GetTickCount();
	int p,n,i,j,k,x,y; float efx,efy;
	FILE *in,*out;
	unsigned char **ship[MAXP];
	char *menu[6]={"Start","High Scores","Options","Instructions","Credits","Exit"};
	unsigned char *pointer,*loggo;
	char leave=0;
	int background,bgsub;
	int macpixmax=5; unsigned char macpixcol[5]={40,42,44,46,48};
	unsigned char **macpix;
	unsigned char **ast,*asz;
	unsigned char *saucer,*timership,*mine,*glowmine,*bomber,*baiter,*pulsar,
		*rubberduck,*duckegg,*ghost,*flamemine,*nball[10];
	unsigned char bullet[MAXP][11]={{3,3,2,10,2,10,14,10,2,10,2},
		{3,3,54,13,54,13,15,13,54,13,54},
		{3,3,7,11,7,11,15,11,7,11,7},
		{3,3,6,14,6,14,15,14,6,14,6}};
	unsigned char redbull[11]={3,3,4,12,4,12,13,12,4,12,4},
		mbullet[6]={2,2,27,29,25,27};
	float *bx,*by,*bxs,*bys; int *bs,*bt; char *bo; int bmax=1000,bm;
	unsigned char *bhit;
	float *fx,*fy,*fxs,*fys; unsigned char *fc; int fmax=10000,fm=0; // Dust: 17 bytes each
	int *ffx,*ffy; unsigned char *ffc; int ffmax=10000,ffm;
	float *ex,*ey,*exs,*eys,*ef,*efs,*enf; unsigned char *er;
	unsigned char ***ea; int emax=2500,em;
	float *ax,*ay,*axs,*ays,*ar,*ars; int *at,*as; int amax=1000,am;
	unsigned char *ahit,*afire,*aplas,*awhit;
	float *ix,*iy,*ixs,*iys; int *iage,*it; int imax=1000,im;
	float *wx,*wy; int *ws; char *wt; int wmax=10,wm;
	unsigned char **items;
	unsigned char ***coin;
	int *cola,*colb; int cmax=1000,cm;
	float *vr=FXP;
	char *okf=CXP,*okbs=CXP;
	int *dead=IXP,blank,*btype=IXP,*mgrech=IXP,*autodestruct=IXP;
	int **upgrade=(int **)malloc(MAXP*sizeof(int *)),upgrades=26,
		*basemodel=(int *)malloc(upgrades*sizeof(int));
	long *vsc=IXP;
	float *px=FXP,*py=FXP,*xs=FXP,*ys=FXP,*rot=FXP,*rs=FXP;
	float *rhx=FXP,*rhy=FXP,*destx=FXP,*desty=FXP;
	char *subgame=CXP,*pacman=CXP;
	char *str,*gamefile;
	char **names; long *scores;
	int place;
	int bplot,bplotsize=32,levtrans_start,levtrans_complete;
	unsigned char *bplotx,*bploty;
	float dt; int idt;
	unsigned char grid[32][24],**magneticfield; int heat[32][24];
	char *battrib,*sattrib,*eattrib,*iattrib;
	unsigned char **blox;
	unsigned char /*cj*/*n_shoot[3]/*/cj*/,*n_explode,*n_dead,*n_thrust,*n_dot,*n_fmine,*n_spark;
	unsigned char **e_fireball,**e_firesmall,**e_corner,**e_sparkle,**e_square;
	int *thstatus=IXP,*moneydisp=IXP,*shopped=IXP,*meterdisp=IXP;
	char *levname[100];
	char theverystart,f12ed; int gameover;
	double framesdone;
	//***********************************
	stat(".\\ncgi.exe",&version_attributes);
	in=fopen(".\\blox.dat","rb"); fseek(in,0,SEEK_END);
	numblox=ftell(in)/402; fclose(in);
/*
	resethiscores();
	createcredits(".\\credits.txt"); // Creates the encrypted credits from file every time
*/
	winhandle=makethewindow(hinstance);
	sound_working=1-directsoundsetup(winhandle);
	directdrawsetup(winhandle); loadprefs();
	makebanks(); cl(0);
	loadfonts(); loading("Generating font"); generatedigits();
	str=(char *)malloc(100);
	openfile=(char *)malloc(20); gamefile=(char *)malloc(20);
	// Misc background initialisation
	loading("Generating plotting cycle");
	bplotx=(unsigned char *)malloc(bplotsize*bplotsize);
	bploty=(unsigned char *)malloc(bplotsize*bplotsize);
	for (n=0;n<bplotsize*bplotsize;n++)
	{
		char c=1;
		while (c)
		{
			bplotx[n]=rnd(bplotsize); bploty[n]=rnd(bplotsize);
			c=0;
			for (i=0;i<n;i++)
				if (bplotx[i]==bplotx[n])
					if (bploty[i]==bploty[n]) c=1;
		}
		cludgetastic[bplotx[n]][bploty[n]]=n;
	}
	loading("Generating sound effects");
	if (sound_working) // Sound effect generation
	{
		float *ta=(float *)malloc(5000*sizeof(float));
		for (n=0;n<SOUNDCHANNELS;n++)
			soundchannel[n]=makebuffer(100,10000);
		n_shoot[0]=(unsigned char *)malloc(5000);
		n_shoot[0][0]=128;
		for (n=1;n<5000;n++)
			n_shoot[0][n]=(int)128.0 + 10.0*sin(1e6/(n*2)) + 5*sin(1e6/(n*5));
		n_explode=(unsigned char *)malloc(1000);
		for (n=0;n<1000;n++)
			n_explode[n]=rnd(exp(-0.002*n)*256)+128-exp(-0.002*n)*128;
		n_dead=(unsigned char *)malloc(20000);
		for (n=0;n<20000;n++)
			n_dead[n]=((frnd(2)-1)*exp(-0.0002*n)+sin(4e7/n)+sin(2.5e7/n))*(64-n*n/6e6)+128;
		n_thrust=(unsigned char *)malloc(5000);
		ta[0]=128;
		for (n=0;n<5000;n++)
		{
			ta[n]=frnd(10)+246.0/256*ta[n-1];
			if (n>4900) ta[n]=(1.0-0.01*(n-4900))*ta[n]+0.01*(n-4900)*ta[0];
			n_thrust[n]=ta[n];
		}
		for (p=0;p<pm;p++) thstatus[p]=-1;
		n_dot=(unsigned char *)malloc(1000);
		for (n=0;n<1000;n++)
			n_dot[n]=0.05*(n%5-2)*(1000-n)+128;
		n_fmine=(unsigned char *)malloc(25000);	// blatant copy of 'completed subgame N' sound for mine
		n_fmine[0]=128;
		for (n=1;n<25000;n++)
		{
			float p=exp(-0.0002*n),q=1.0-p,r=exp(-0.0005*(25000-n));
			n_fmine[n]=(1.0-r)*(q*n_fmine[n-1]+p*rnd(256))+r*128;
		}
		n_spark=(unsigned char *)malloc(256);
		for (n=0;n<256;n++) n_spark[n]=(n+128)%256;
/*cj*/
		n_shoot[1]=(unsigned char *)malloc(5000);
		for (n=0;n<5000;n++)
			n_shoot[1][n]=128;
		for (n=0;n<5000;n++)
			ta[n]=exp(-0.002*n)*(frnd(100)-50)
				 +(exp(-0.001*n)*sin(n/20+frnd(1.1))*10)
				 +(exp(-0.0007*n)*sin(n/10+frnd(1.1))*5)
				 +(exp(-0.0008*n)*sin(n/15+frnd(1.1))*3)
				 +(exp(-0.001*n)*sin(n/40+frnd(1.1))*10)
				 +(exp(-0.0007*n)*sin(n/27+frnd(1.1))*9)
				 +(exp(-0.0008*n)*sin(n/18+frnd(1.1))*8)
				 +128;

		for (n=1;n<5000;n++)
		{
			ta[n]=(ta[n]+ta[n-1])/2;
			ta[n]+=exp(-0.01*n)*(frnd(100)-50);
			n_shoot[1][n]=ta[n];
		}
		n_shoot[2]=(unsigned char *)malloc(5000);
		for (n=0;n<5000;n++)
		{
			ta[n]=128;
			n_shoot[2][n]=128;
		}
		for (n=1;n<5000;n++)
			ta[n]=frnd(10)+246.0/256*ta[n-1]; // Engine noise recycled
		for (n=0;n<5000;n++)
			ta[n]=(ta[n]-128)*(.5+((n<2000)*((float)n/2000)+(n>2000)*((float)5000-n)/3000)/2)+128;
		for (n=3;n<5000;n++)
		{
			float fil=1-((n<2000)*((float)n/2000)
					 +(n>2000)*(((float)5000-n)/3000));
			ta[n]=(ta[n]+(fil*ta[n-1])+(fil*ta[n-2])+(fil*ta[n-3]))/(1+fil*3);	// Dodgy filter
			n_shoot[2][n]=ta[n];
		}
/*/cj*/

	}
	// Ship initialisation
	loading("Pre-rendering ships");
	for (p=0;p<MAXP;p++)
	{
		ship[p]=(unsigned char **)malloc(63*sizeof(unsigned char *));
		for (n=0;n<63;n++)
		{
			ship[p][n]=(unsigned char *)malloc(258);
			i=0;
			if (p==1) i=4914;
			else if (p==2) i=5172;
			else if (p==3) i=5430;
			if (n==0) readbit(ship[p][n],".\\ncgi.dat",i);
			else
			{
				float theta=0.1*n;
				ship[p][n][0]=ship[p][n][1]=16;
				for (x=0;x<16;x++)
					for (y=0;y<16;y++)
					{
						efx=((float)x-7.5)*cos(theta)+((float)y-7.5)*sin(theta)+7.5;
						efy=((float)y-7.5)*cos(theta)-((float)x-7.5)*sin(theta)+7.5;
						efx=(int)(efx+0.5); efy=(int)(efy+0.5);
						if (efx>=0 && efy>=0 && efx<16 && efy<16)
							ship[p][n][x+y*16+2]=ship[p][0][(int)(efx+efy*16)+2];
						else ship[p][n][x+y*16+2]=255;
					}
			}
		}
	}
	loading("Loading bitmaps");
	saucer=(unsigned char *)malloc(15*7+2); // Loading bitmaps
	readbit(saucer,".\\ncgi.dat",258);
	timership=(unsigned char *)malloc(20*15+2);
	readbit(timership,".\\ncgi.dat",365);
	mine=(unsigned char *)malloc(12*12+2);
	readbit(mine,".\\ncgi.dat",667);
	glowmine=(unsigned char *)malloc(12*12+2);
	readbit(glowmine,".\\ncgi.dat",813);
	flamemine=(unsigned char *)malloc(16*16+2);
	readbit(flamemine,".\\ncgi.dat",11076);
	loading("Generating graphics");
	macpix=(unsigned char **)malloc(macpixmax*sizeof(unsigned char *)); // Macro-pixel initialisation
	for (n=0;n<macpixmax;n++)
	{
		int size=4*pot[n]-1;
		macpix[n]=(unsigned char *)malloc(size*size+2);
		macpix[n][0]=macpix[n][1]=size;
		for (i=0;i<size*size;i++)
			macpix[n][2+i]=(i%size==0 || i%size==size-1 ||
								i/size==0 || i/size==size-1)*72+macpixcol[n];
	}
	ast=(unsigned char **)malloc(20*sizeof(unsigned char *)),asz=(unsigned char *)malloc(21);
	loading("Loading more bitmaps");
	pulsar=(unsigned char *)malloc(102);
	readbit(pulsar,".\\ncgi.dat",959);
	rubberduck=(unsigned char *)malloc(302);
	readbit(rubberduck,".\\ncgi.dat",1061);
	duckegg=(unsigned char *)malloc(37);
	readbit(duckegg,".\\ncgi.dat",1363);
	bomber=(unsigned char *)malloc(123);
	readbit(bomber,".\\ncgi.dat",5688);
	baiter=(unsigned char *)malloc(102);
	readbit(baiter,".\\ncgi.dat",5811);
	ghost=(unsigned char *)malloc(402);
	readbit(ghost,".\\ncgi.dat",10047);
	loading("Generating balls");
	for (n=0;n<10;n++) // Make N ball sprites
	{
		char c=40+n*2;
		nball[n]=(unsigned char *)malloc(102);
		nball[n][0]=nball[n][1]=10;
		for (x=0;x<10;x++) for (y=0;y<10;y++)
		{
			i=fabs(4.5-x)+fabs(4.5-y);
			if (i<7) nball[n][2+x+y*10]=c;
			else if (i>7) nball[n][2+x+y*10]=255;
			else nball[n][2+x+y*10]=c+72;
		}
		nball[n][5]=nball[n][8]=nball[n][95]=nball[n][98]=c+72;
		nball[n][32]=nball[n][41]=nball[n][62]=nball[n][71]=c+72;
	}
	// Animation creation
	loading("Creating animations");
	e_fireball=(unsigned char **)malloc(10*sizeof(unsigned char *));
	for (n=0;n<10;n++)
	{
		unsigned char c[12]={255,130,131,132,137,142,147,152,177,202,227,252};
		e_fireball[n]=(unsigned char *)malloc(402);
		e_fireball[n][0]=e_fireball[n][1]=20;
		for (x=0;x<20;x++) for (y=0;y<20;y++)
			e_fireball[n][2+x+y*20]=
				c[(int)Max(0,Min(11,2*(9-n)-1.8*sqrt(sq(x-10)+sq(y-10))+frnd(1)))];
	}
	e_firesmall=(unsigned char **)malloc(10*sizeof(unsigned char *));
	for (n=0;n<10;n++)
	{
		unsigned char c[12]={255,130,131,132,137,142,147,152,177,202,227,252};
		e_firesmall[n]=(unsigned char *)malloc(66);
		e_firesmall[n][0]=e_firesmall[n][1]=8;
		for (x=0;x<8;x++) for (y=0;y<8;y++)
			e_firesmall[n][2+x+y*8]=
				c[(int)Max(0,Min(11,2*(9-n)-4.5*sqrt(sq(x-3.5)+sq(y-3.5))+frnd(1)))];
	}
	e_corner=(unsigned char **)malloc(10*sizeof(unsigned char *));
	for (n=0;n<10;n++)
	{
		float th=M_TWOPI*n/10,inc=M_PI*2/3;
		rect(0,0,6,6,255);
		line(3+2.5*cos(th),3+2.5*sin(th),3+2.5*cos(th+inc),3+2.5*sin(th+inc),55);
		th+=inc;
		line(3+2.5*cos(th),3+2.5*sin(th),3+2.5*cos(th+inc),3+2.5*sin(th+inc),55);
		e_corner[n]=(unsigned char *)malloc(38);
		e_corner[n][0]=e_corner[n][1]=6;
		for (x=0;x<6;x++)
			for (y=0;y<6;y++)
				e_corner[n][2+x+y*6]=getpix(x,y);
	}
	e_sparkle=(unsigned char **)malloc(2*sizeof(unsigned char *));
	readbit(e_sparkle[0]=(unsigned char *)malloc(102),".\\ncgi.dat",1546);
	readbit(e_sparkle[1]=(unsigned char *)malloc(102),".\\ncgi.dat",1648);
	e_square=(unsigned char **)malloc(30*sizeof(unsigned char *));
	for (n=0;n<30;n++)
	{
		e_square[n]=(unsigned char *)malloc(402);
		e_square[n][0]=e_square[n][1]=20;
		for (i=2;i<402;i++) e_square[n][i]=(rnd(900)<sq(30-n))?15:255;
	}
	loading("Allocating memory");
	// Bullet initialisation // 25.125 bytes each
	bx=(float *)malloc(bmax*sizeof(float)); by=(float *)malloc(bmax*sizeof(float));
	bxs=(float *)malloc(bmax*sizeof(float)); bys=(float *)malloc(bmax*sizeof(float));
	bs=(int *)malloc(bmax*sizeof(int)); bt=(int *)malloc(bmax*sizeof(int));
	bo=(char *)malloc(bmax);
	bhit=(unsigned char *)malloc((bmax+7)/8);
	// Dust initialisation // 17 bytes each
	fx=(float *)malloc(fmax*sizeof(float)); fy=(float *)malloc(fmax*sizeof(float));
	fxs=(float *)malloc(fmax*sizeof(float)); fys=(float *)malloc(fmax*sizeof(float));
	fc=(unsigned char *)malloc(fmax);
	// Flame initialisation // 9 bytes each
	ffx=(int *)malloc(ffmax*sizeof(int)); ffy=(int *)malloc(ffmax*sizeof(int));
	ffc=(unsigned char *)malloc(ffmax);
	//	Bitmapped explosion initialisation // 33 bytes each
	ex=(float *)malloc(emax*sizeof(float)); ey=(float *)malloc(emax*sizeof(float));
	exs=(float *)malloc(emax*sizeof(float)); eys=(float *)malloc(emax*sizeof(float));
	ef=(float *)malloc(emax*sizeof(float)); efs=(float *)malloc(emax*sizeof(float));
	er=(unsigned char *)malloc(emax); enf=(float *)malloc(emax*sizeof(float));
	ea=(unsigned char ***)malloc(emax*sizeof(unsigned char **));
	// Enemy initialisation // 36 bytes each
	ax=(float *)malloc(amax*sizeof(float)); ay=(float *)malloc(amax*sizeof(float));
	axs=(float *)malloc(amax*sizeof(float)); ays=(float *)malloc(amax*sizeof(float));
	ar=(float *)malloc(amax*sizeof(float)); ars=(float *)malloc(amax*sizeof(float));
	at=(int *)malloc(amax*sizeof(int)); as=(int *)malloc(amax*sizeof(int));
	ahit=(unsigned char *)malloc(amax);
	afire=(unsigned char *)malloc(amax);
	aplas=(unsigned char *)malloc(amax);
	awhit=(unsigned char *)malloc(amax);
	// Items initialisation // 24 bytes each
	ix=(float *)malloc(imax*sizeof(float)); iy=(float *)malloc(imax*sizeof(float));
	ixs=(float *)malloc(imax*sizeof(float)); iys=(float *)malloc(imax*sizeof(float));
	iage=(int *)malloc(imax*sizeof(int)); it=(int *)malloc(imax*sizeof(int));
	items=(unsigned char **)malloc(itypes*sizeof(unsigned char *));
	loading("Loading even more bitmaps");
	for (n=0;n<itypes;n++)
	{
		items[n]=(unsigned char *)malloc(10*10+2);
		readbit(items[n],".\\boni.dat",n*102);
	}
	loading("Minting currency");
	coins_initialisation(&coin); // Make coins
	loading("Allocating memory, again");
	// Wormhole initialisation // 13 bytes each
	wx=(float *)malloc(wmax*sizeof(float)); wy=(float *)malloc(wmax*sizeof(float));
	ws=(int *)malloc(wmax*sizeof(int)); wt=(char *)malloc(wmax);
	// Collision buffer initialisation
	cola=(int *)malloc(cmax*sizeof(int)); colb=(int *)malloc(cmax*sizeof(int));
	// Grid initialisation
	loading("Initialising game grid");
	blox=(unsigned char **)malloc(0x100*sizeof(unsigned char *));
	for (n=0;n<0x100;n++) blox[n]=(unsigned char *)malloc(20*20+2);
	battrib=(char *)malloc(0x100); // 0=through 1=bounce 2=absorb
	for (n=0;n<=0xFF;n++)
	{
		if (n>=0x01 && n<=0x13 || n>=0x18 && n<=0x1F || n>=0x32 && n<=0x44 ||
			n>=0x50 && n<=0x53 || n>=0xC0 && n<=0xCF || n==0xDE || n>=0xFA && n<=0xFB) battrib[n]=1;
		else if (n>=0x14 && n<=0x17 || n==0xDF) battrib[n]=2;
		else battrib[n]=0;
	}
	sattrib=(char *)malloc(0x100); // 0=through 1,2,3=bounce:100%,50%,15% 4=stop 5=destroy
	for (n=0;n<=0xFF;n++)
	{
		if (n>=0x10 && n<=0x13 || n==0x40) sattrib[n]=1;
		else if (n>=0x01 && n<=0x0F || n>=0x30 && n<=0x33 || n>=0x36 && n<=0x39 ||
			n>=0x41 && n<=0x44 || n>=0x4D && n<=0x4F || n>=0xC0 && n<=0xCF || n==0xDE ||
			n==0xDF || n>=0xFA && n<=0xFB) sattrib[n]=2;
		else if (n>=0x14 && n<=0x17) sattrib[n]=3;
		else if (n>=0x18 && n<=0x1B) sattrib[n]=4;
		else if (n>=0x1C && n<=0x1F || n>=0x3C && n<=0x3F || n>=0x45 && n<=0x4C ||
			n>=0x50 && n<=0x57) sattrib[n]=5;
		else sattrib[n]=0;
	}
	eattrib=(char *)malloc(0x100); // 0=through 1=bounce
	for (n=0;n<=0xFF;n++)
		eattrib[n]=(n>=0x01 && n<=0x1F || n>=0x30 && n<=0x35 || n>=0x3A && n<=0x44 ||
			n>=0x4D && n<=0x53 || n>=0xC0 && n<=0xCF || n==0xDE || n==0xDF || n>=0xFA && n<=0xFB);
	iattrib=(char *)malloc(0x100); // 0=through 1=bounce
	for (n=0;n<=0xFF;n++)
		iattrib[n]=(n>=0x01 && n<=0x1F || n>=0x30 && n<=0x31 || n>=0x34 && n<=0x37 ||
			n>=0x3C && n<=0x44 || n>=0x4D && n<=0x53 || n>=0xC0 && n<=0xCF || n==0xDE ||
			n==0xDF || n>=0xFA && n<=0xFB);
	magneticfield=(unsigned char **)malloc(32*sizeof(unsigned char *));
	for (x=0;x<32;x++) magneticfield[x]=(unsigned char *)malloc(24);
	// Initialise upgrade system
	loading("Initialising upgrade system");
	for (n=0;n<MAXP;n++) upgrade[n]=(int *)malloc(upgrades*sizeof(int));
	pointer=(unsigned char *)malloc(146);
	loggo=(unsigned char *)malloc(1802);
	// Level names
	loading("Naming levels");
	for (n=0;n<100;n++)
	{
		levname[n]=(char *)malloc(100);
		strcpy(levname[n],"Unnamed Level");
	}
	strcpy(levname[8-1],"'Alpha' by Michael Spencer");
	strcpy(levname[9-1],"'Beta' by Michael Spencer");
	strcpy(levname[10-1],"'Gamma' by Michael Spencer");
	strcpy(levname[11-1],"'Delta' by Michael Spencer");
	strcpy(levname[100-1],"The Missing Link");
	loading("Loading title screen");
title_screen:															/* TITLE SCREEN STARTS HERE */
	leave=0; strcpy(openfile,"");
	readbit(pointer,".\\ncgi.dat",1400);
	readbit(loggo,".\\ncgi.dat",3112);
	titlepal(gamma);
	clb(0); mouseoff();
	titlescreen_fractal();
	if (!strstr(cmdline,".ncgi")) {float kx=0,ky=0,kxs=0,kys=0;
	while (!leave) // Title screen loop
	{
		long framestart=GetTickCount();
		float omx=mx,omy=my,rx=mx+kx,ry=my+ky; int rb=mb|KEY(VK_RETURN)|KEY(VK_SPACE);
		char str2[100];
		if (KEY(VK_LEFT)) kxs--;
		if (KEY(VK_RIGHT)) kxs++;
		if (KEY(VK_UP)) kys--;
		if (KEY(VK_DOWN)) kys++;
		kx+=kxs; ky+=kys;
		kxs*=0.7; kys*=0.7;
		if (mx+kx<0) {kx=-mx; kxs=0;}
		if (my+ky<0) {ky=-my; kys=0;}
		if (mx+kx>=xres-1) {kx=xres-1-mx; kxs=0;}
		if (my+ky>=yres-1) {ky=yres-1-my; kys=0;}
		if (KEY(VK_CONTROL) && KEY(VK_T)) {technical^=1; while (KEY(VK_T) && !dealwithmessages());}
		wrapbit(30,70,loggo);
		write(124,70,xres,"distribution 2a+",15,7); // Title
		sprintf(str,"Developer's Version %s",ncgitime(version_attributes.st_mtime,str2));
		write(30,94,xres,str,0,1);
		for (n=0,i=-1;n<6;n++)
		{
			if (ry>=118+n*30 && ry<=129+n*30 && rx<xres/2)
			{
				i=n;
				bwrite(40,120+n*30,xres,menu[n],15);
			}
			else bwrite(40,120+n*30,xres,menu[n],0);
		}
		if (!rb) i=42;
		if (i==0 || KEY(VK_S)) leave=1;
		else if (i==1 || KEY(VK_H)) hiscores();
		else if (i==2 || KEY(VK_O)) options();
		else if (i==3 || KEY(VK_I)) instructions();
		else if (i==4 || KEY(VK_C)) credits();
		else if (i==5 || KEY(VK_ESCAPE) || KEY(VK_F12)) goto THE_END;
		else if (rb && rx<=4 && ry>=yres-4 && technical) leveled();
		if (technical)
		{
			efx=(_rdtsc()-CSTART)/(0.001*GetTickCount()-TSTART);
			sprintf(str,"%ld",(long)(efx/1e6));
			bwrite(0,0,xres,str,13);
			sprintf(str,"%s",ncgitime(time(NULL),str));
			write(xres*3/4,0,xres,str,15,7);
			line(0,yres-1,2,yres-1,15); // GAMMA
			line(1,yres-1,1,yres-4,15);
			line(1,yres-4,4,yres-4,15);
			putpix(4,yres-3,15);
		}
		mskbit(rx,ry,pointer); // The mouse
		if (!leave) cclb();
		do if (dealwithmessages()) goto THE_END;
		while (GetTickCount()<framestart+mswait); // Synchronize frame
		if (omx!=mx || omy!=my) kx=ky=kxs=kys=0;
	}}
	mouseoff();
	leave=0; str[0]=0;
	if (strstr(cmdline,".ncgi")) // Command-line argument
	{
		char *s;
		if (s=strstr(cmdline,".exe")) strcpy(openfile,s+5);
			else strcpy(openfile,cmdline);
	}
	else
	{
		while (mb) if (dealwithmessages()) goto THE_END;
		while (!leave && !dealwithmessages()) // Load game screen
		{
			char a=lc(getch); n=strlen(str);
			if (mb) leave=1;
			if (KEY(VK_ESCAPE) || KEY(VK_F12)) goto title_screen;
			while (kbhit && getch==a) if (dealwithmessages()) goto THE_END;
			memset(shad,0,xres*40);
			memset(shad+xres*40,160,xres*400);
			memset(shad+xres*440,0,xres*40);
			bwrite(60,60,xres,"Start:",15);
			bwrite(60,80,xres,"Enter filename below to load/create\nor press enter to start new game",200);
			bwrite(60,100,xres,str,0);
			if (a>='a' && a<='z' || a>='0' && a<='9' || a=='-')
			{
				if (n<32) {str[n]=a; str[n+1]=0;}
			}
			else if (a==8 && n) str[n-1]=0;
			else if (a==13) leave=1;
			if (!leave) ccl(0);
		}
		if (*str)
		{
			openfile=".\\";
			strcat(openfile,strlwr(str));
			if (openfile[2]>='a' && openfile[2]<='z') openfile[2]-=32;
			strcat(openfile,".ncgi");
		}
		else openfile="";
	}
	leave=0; str[0]=0;
	if (!*openfile && technical) while (!leave && !dealwithmessages()) // TechJump if no game loading
	{
		char a=lc(getch); n=strlen(str);
		if (KEY(VK_ESCAPE) || KEY(VK_F12)) goto title_screen;
		while (kbhit && getch==a) if (dealwithmessages()) goto THE_END;
		memset(shad,160,xres*40);
		memset(shad+xres*40,0,xres*400);
		memset(shad+xres*440,160,xres*40);
		bwrite(60,60,xres,"Technical Mode:",15);
		bwrite(60,80,xres,"Enter level/subgame code below",200);
		bwrite(60,100,xres,str,160);
		if (a>='a' && a<='z' || a>='0' && a<='9')
		{
			if (n<32) {str[n]=uc(a); str[n+1]=0;}
		}
		else if (a==8 && n) str[n-1]=0;
		else if (a==13) leave=1;
		if (!leave) ccl(0);
	}
	cl(0); writepal(gamma);
	gamestack[0]=0; gamedepth=0; // Start in root game
	for (n=0;n<26;n++) seeninfo[n]=0;
	for every_player(p) lives[p]=3;
	level=background=0;
	nlevel=clevel=glevel=ilevel=1;
	//									Title screen ends, game begins
	am=bm=fm=ffm=em=im=wm=0; // Wipe all objects
	for (x=0;x<32;x++) for (y=0;y<24;y++) // Clear grid
		grid[x][y]=heat[x][y]=0;
	// Upgrades handling
	for (n=0;n<upgrades;n++) basemodel[n]=0;
	basemodel[3]=basemodel[4]=1; basemodel[9]=250;
	basemodel[10]=0.08/0.001;
	basemodel[12]=basemodel[11]=3.0/0.001;
	basemodel[14]=basemodel[13]=350.0/0.001;
	// Players initialisation
	blank=0;
	for every_player(p)
	{
		vr[p]=5;
		okf[p]=okbs[p]=0;
		dead[p]=0,btype[p]=1,mgrech[p]=autodestruct[p]=0;
		for (n=0;n<upgrades;n++) upgrade[p][n]=basemodel[n];
		score[p]=0,vsc[p]=0; money[p]=0;
		px[p]=xres/2,py[p]=yres/2,xs[p]=0,ys[p]=0,rot[p]=0,rs[p]=0;
		rhx[p]=sin(rot[p]),rhy[p]=-cos(rot[p]);
		thstatus[p]=-1; moneydisp[p]=shopped[p]=meterdisp[p]=0;
		subgame[p]=0;
	}
	if (technical) // Set up gamestack if TechJump used
	{
		i=numberletters(str);
		for (n=0;n<i;n++)
		{
			char a=letterindex(str,n);
			gamestack[n]=a;
			if (a=='C') clevel=Max(1,numericindex(str,n+1));
			else if (a=='N') nlevel=Max(1,numericindex(str,n+1));
			if (n==0) for every_player(p) subgame[p]=a;
		}
		gamestack[i]=0;
		level=noneg(numericindex(str,0)-1);
	}
	// Timing / Control stuff
	bplot=0; dt=1;
	theverystart=1; leave=0;
	gamestart=0.001*GetTickCount(); nextsave=gamestart+autosave;
// BEGINNING OF MAIN LOOP
	for (framesdone=0;!leave;framesdone++)
	{
		long framestart=GetTickCount();
		int shit[MAXP]={0,0,0,0};
		int signal_savegame=0,signal_loadgame=0;
		if (dealwithmessages()) goto THE_END;
		if (theverystart) gameover=0;
		gameover++;
		for every_player(p) gameover*=(dead[p]>100 && !lives[p]);
		if (!gameover) gametime=0.001*GetTickCount()-gamestart;
		if (theverystart && *openfile)
		{
			strcpy(gamefile,openfile);
			signal_loadgame=1;
		}
		theverystart=0;
		if (0.001*GetTickCount()>=nextsave && *openfile) // Auto-save
		{
			strcpy(gamefile,openfile); signal_savegame=1;
			nextsave+=autosave;
		}
		if (KEY(VK_F6) && KEY(VK_F11) && technical) for every_player(p) money[p]=100000000;
		idt=(int)floor(dt+0.5);
		for every_player(p) if (!subgame[p] && !dead[p]) // PLAYERS LOOP
		{
			int anyleft=1,qx,qy,ox=(int)px[p]/20,oy=(int)py[p]/20;
			px[p]+=xs[p]*dt; py[p]+=ys[p]*dt; rot[p]+=rs[p]*dt;
			if (px[p]<0) px[p]+=xres;
			else if (px[p]>=xres) px[p]-=xres;
			if (py[p]<0) py[p]+=yres;
			else if (py[p]>=yres) py[p]-=yres;
			qx=(int)px[p]/20; qy=(int)py[p]/20;
			pacman[p]=(grid[qx][qy]>>4==0xD);
			vr[p]=(pacman[p])?9.5:5;
			if (btype[p]==2 && !upgrade[p][0]) anyleft=0;
			else if (btype[p]==3 && !upgrade[p][1]) anyleft=0;
			else if (btype[p]==4 && !upgrade[p][2]) anyleft=0;
			else if (btype[p]==5 && !upgrade[p][7]) anyleft=0;
			else if (btype[p]==6 && !upgrade[p][19]) anyleft=0;
			else if (btype[p]==7 && !upgrade[p][20]) anyleft=0;
			else if (btype[p]==8 && !upgrade[p][21]) anyleft=0;
			else if (btype[p]==9 && !upgrade[p][22]) anyleft=0;
			else if (btype[p]==10 && !upgrade[p][23]) anyleft=0;
			else if (btype[p]==11 && !upgrade[p][24]) anyleft=0;
			if (!anyleft) {btype[p]=1; meterdisp[p]=150;}
			if (upgrade[p][25] && pacman[p]) btype[p]=2; // Auto-switch to chilli breath
			k=(KEY(firekey[p]) && (!pacman[p] || upgrade[p][25]));

			if (btype[p]==6)
			{
				k=0; // No firing with light sabre
				if (bm+18<=bmax)
					for (n=0;n<=51;n+=3)
					{
						bx[bm]=px[p]+sin(rot[p])*n;
						by[bm]=py[p]-cos(rot[p])*n;
						bxs[bm]=bys[bm]=0;
						bt[bm]=6; bo[bm]=p;
						bs[bm]=1;
						bm++;
					}
				for (n=2;n>=0;n--) for (x=-1;x<=1;x++) for (y=-1;y<=1;y++) if (abs(x)+abs(y)==n)
					line(px[p]+x,py[p]+y,px[p]+x+sin(rot[p])*50,py[p]+y-cos(rot[p])*50,248-5*n);
			}
			else if (btype[p]==7) // Anti-SRM pulse
			{
				if (k && !okf[p])
				{
					for every_enemy(n) if (at[n]==11) {as[n]=2; at[n]=1;}
					upgrade[p][20]--;
				}
			}
			else if (btype[p]==8) // Lasers
			{
				if (k && !okf[p] && bm<bmax)
				{
					bx[bm]=px[p]; by[bm]=py[p];
					bxs[bm]=xs[p]+sin(rot[p])*10;
					bys[bm]=ys[p]-cos(rot[p])*10;
					bt[bm]=7; bo[bm]=p;
					bs[bm]=29+30*(upgrade[p][21]-1);
					bm++;
				}
			}
			else if (btype[p]==9 || btype[p]==10) // Prodder of some sort
			{
				if (k && !okf[p] && bm<bmax)
				{
					bx[bm]=px[p]; by[bm]=py[p];
					bxs[bm]=xs[p]+sin(rot[p]);
					bys[bm]=ys[p]-cos(rot[p]);
					bt[bm]=btype[p]-1; bo[bm]=p;
					bs[bm]=1; // Lasts forever
					bm++;
				}
			}
			else if (btype[p]==11) // Flame Mine
			{
				if (k && !okf[p] && bm<bmax)
				{
					bx[bm]=px[p]; by[bm]=py[p];
					bxs[bm]=bys[bm]=0;
					bt[bm]=10; bo[bm]=p;
					bs[bm]=200;
					bm++; upgrade[p][24]--;
				}
			}
			else if (!okf[p] || btype[p]==5 && !mgrech[p]) // btype[p]<=5
				if (k && bm+upgrade[p][3]*upgrade[p][4]+upgrade[p][8]<=bmax) // Shoot
			{
				int btotal;
				for (n=0;n<upgrade[p][3]*upgrade[p][4];n++)
				{
					float dev=0.5+n%upgrade[p][3]-upgrade[p][3]*0.5;
					bx[bm+n]=px[p]+rhx[p]*8; by[bm+n]=py[p]+rhy[p]*8;
					bxs[bm+n]=xs[p]+rhx[p]*0.001*upgrade[p][12]+rhy[p]*dev;
					bys[bm+n]=ys[p]+rhy[p]*0.001*upgrade[p][12]-rhx[p]*dev;
					i=n/upgrade[p][3];
					dev=exp(0.1*i-0.05*upgrade[p][4]);
					bxs[bm+n]*=dev; bys[bm+n]*=dev;
					bs[bm+n]=(float)upgrade[p][14]/upgrade[p][12]; bt[bm+n]=btype[p]; bo[bm+n]=p;
					if (btype[p]==5) {bxs[bm+n]*=2; bys[bm+n]*=2; bs[bm+n]/=2;}
				}
				bm+=upgrade[p][3]*upgrade[p][4];
				for (n=0;n<upgrade[p][8];n++)
				{
					bx[bm+n]=px[p]; by[bm+n]=py[p];
					bxs[bm+n]=xs[p]+0.001*upgrade[p][12]*sin(rot[p]+M_PI*2*(n+1)/(upgrade[p][8]+1));
					bys[bm+n]=ys[p]+0.001*upgrade[p][12]*-cos(rot[p]+M_PI*2*(n+1)/(upgrade[p][8]+1));
					bs[bm+n]=(float)upgrade[p][14]/upgrade[p][12]; bt[bm+n]=btype[p]; bo[bm+n]=p;
					if (btype[p]==5) {bxs[bm+n]*=2; bys[bm+n]*=2; bs[bm+n]/=2;}
				}
				bm+=upgrade[p][8];
				btotal=upgrade[p][3]*upgrade[p][4]+upgrade[p][8];
				if (btype[p]==3) upgrade[p][1]--;
				if (btype[p]==5)
				{
					mgrech[p]=3;
					upgrade[p][7]-=btotal;
					if (upgrade[p][7]<0) upgrade[p][7]=0;
					btotal-=upgrade[p][8]+(upgrade[p][8]>0);
					xs[p]-=0.001*upgrade[p][10]*rhx[p]*0.07*0.001*upgrade[p][12]*btotal;
					ys[p]-=0.001*upgrade[p][10]*rhy[p]*0.07*0.001*upgrade[p][12]*btotal;
				}
				if(btype[p]==4 || btype[p]==2) ncgisound(px[p],n_shoot[2],5000,11025,0);        // if you're flaming, then it flames
				else if (btype[p]==5) ncgisound(px[p],n_shoot[1],5000,11025,0);								// if you're machine-gunning...
				else ncgisound(px[p],n_shoot[0],5000,11025,0);
			}
			okf[p]=k;
			mgrech[p]=noneg(mgrech[p]-idt);
			k=KEY(selectkey[p]);
			if (k) autodestruct[p]+=idt; else autodestruct[p]=0;
			if (!okbs[p] && !pacman[p]) if (k)
			{
				btype[p]++;
				if (btype[p]==2 && !upgrade[p][0]) btype[p]++;
				if (btype[p]==3 && !upgrade[p][1]) btype[p]++;
				if (btype[p]==4 && !upgrade[p][2]) btype[p]++;
				if (btype[p]==5 && !upgrade[p][7]) btype[p]++;
				if (btype[p]==6 && !upgrade[p][19]) btype[p]++;
				if (btype[p]==7 && !upgrade[p][20]) btype[p]++;
				if (btype[p]==8 && !upgrade[p][21]) btype[p]++;
				if (btype[p]==9 && !upgrade[p][22]) btype[p]++;
				if (btype[p]==10 && !upgrade[p][23]) btype[p]++;
				if (btype[p]==11 && !upgrade[p][24]) btype[p]++;
				if (btype[p]>11) btype[p]=1;
				meterdisp[p]=100;
			}
			okbs[p]=k;
			if (upgrade[p][2]) btype[p]=4; // Auto-switch to white-fire
			if (pacman[p]) // Pacman movement
			{
				if (KEY(leftkey[p])) {xs[p]=-1; ys[p]=0; rot[p]=-M_PI/2;}
				if (KEY(rightkey[p])) {xs[p]=1; ys[p]=0; rot[p]=M_PI/2;}
				if (KEY(upkey[p])) {xs[p]=0; ys[p]=-1; rot[p]=0;}
				if (KEY(downkey[p])) {xs[p]=0; ys[p]=1; rot[p]=M_PI;}
				efx=fabs(xs[p]); efy=fabs(ys[p]);
				if (efx<efy) px[p]=qx*20+10; else py[p]=qy*20+10; // Snap to corridor centres
				if (ox!=qx || oy!=qy) // Runs onto new square
				{
					i=grid[qx][qy];
					if (i>=0xD1 && i<=0xD6) // Eats a dot
					{
						grid[qx][qy]--;
						wrapbitb(qx*20,qy*20,blox[i-1]);
						score[p]+=10;
						ncgisound(px[p],n_dot,1000,11025,0);
					}
					else if (i==0xD7) // Power pill
					{
						grid[qx][qy]=0xD0;
						wrapbitb(qx*20,qy*20,blox[0xD0]);
						score[p]+=50;
						ncgisound(px[p],n_dot,1000,1024,0);
						for every_enemy(n) if (at[n]==21) as[n]=1;
					}
					else if (i==0xD8) // Chilli pepper
					{
						grid[qx][qy]=0xD0;
						wrapbitb(qx*20,qy*20,blox[0xD0]);
						score[p]+=20;
						upgrade[p][25]=1000;
					}
				}
				if (sound_working && (i=thstatus[p])>=0) // No thrust noise
				{
					stopsound(soundchannel[i]); killbuffer(soundchannel[i]);
					soundchannel[i]=makebuffer(100,10000);
					thstatus[p]=-1;
				}
			}
			else // Non-pacman movement
			{
				if (KEY(leftkey[p])) rot[p]-=0.1*dt;
				if (KEY(rightkey[p])) rot[p]+=0.1*dt;
				if (KEY(upkey[p])) // Go forward
				{
					int nfm=Min(fm+6,fmax);
					float ca=cos(rot[p]-0.3),cb=cos(rot[p]+0.3),
							sa=sin(rot[p]-0.3),sb=sin(rot[p]+0.3);
					xs[p]+=0.001*upgrade[p][10]*rhx[p]*dt;
					ys[p]+=0.001*upgrade[p][10]*rhy[p]*dt;
					if (fm<fmax) for new_dust(n)
					{
						fx[n]=px[p]; fy[n]=py[p];
						if (p==0) // Normal thrust
						{
							fxs[n]=frnd(1)-0.5-rhx[p]*2;
							fys[n]=frnd(1)-0.5-rhy[p]*2;
							fc[n]=40+rnd(5);
						}
						else if (p==1) // Two tails
						{
							if (n%2)
							{
								fxs[n]=frnd(0.5)-0.25-sa*2;
								fys[n]=frnd(0.5)-0.25+ca*2;
							}
							else
							{
								fxs[n]=frnd(0.5)-0.25-sb*2;
								fys[n]=frnd(0.5)-0.25+cb*2;
							}
							fc[n]=40+rnd(5);
						}
						else if (p==2) // Blue fine jet
						{
							fxs[n]=frnd(0.1)-0.05-rhx[p]*0.2;
							fys[n]=frnd(0.1)-0.05-rhy[p]*0.2;
							fc[n]=52+rnd(5);
						}
						else if (p==3) // Yellow jet
						{
							fx[n]+=rhx[p]*rnd(8);
							fy[n]+=rhy[p]*rnd(8);
							fxs[n]=frnd(0.4)-0.2-rhx[p]*3;
							fys[n]=frnd(0.4)-0.2-rhy[p]*3;
							fc[n]=14;
						}
					}
					fm=nfm;
					if (sound_working && thstatus[p]==-1)
					{
						thstatus[p]=i=findfree(0);
						stopsound(soundchannel[i]); killbuffer(soundchannel[i]);
						soundchannel[i]=makebuffer(5000,10000);
						writesound(n_thrust,soundchannel[i],5000);
						loopsound(soundchannel[i]);
					}
				}
				else if (sound_working && thstatus[p]>=0)
				{
					i=thstatus[p];
					stopsound(soundchannel[i]); killbuffer(soundchannel[i]);
					soundchannel[i]=makebuffer(100,10000);
					thstatus[p]=-1;
				}
				if ((i=thstatus[p])>=0) IDirectSoundBuffer_SetPan(soundchannel[i],
					(px[p]-xres/2)*STEREO); // Continuous panning
				if (KEY(downkey[p]) && upgrade[p][6]) // Go backward
				{
					int nfm=Min(fm+6,fmax);
					xs[p]-=0.001*upgrade[p][10]*rhx[p]*(1.0-pow(0.5,upgrade[p][6]))*dt;
					ys[p]-=0.001*upgrade[p][10]*rhy[p]*(1.0-pow(0.5,upgrade[p][6]))*dt;
					if (fm<fmax) for new_dust(n)
					{
						i=rnd(2)*10-5;
						fx[n]=px[p]+rhy[p]*i; fy[n]=py[p]-rhx[p]*i;
						fxs[n]=frnd(0.3)-0.15+rhx[p]*(3+frnd(1));
						fys[n]=frnd(0.3)-0.15+rhy[p]*(3+frnd(1));
						fc[n]=(upgrade[p][6]<4)?40+upgrade[p][6]:44;
					}
					fm=nfm;
				}
			}
			rhx[p]=sin(rot[p]); rhy[p]=-cos(rot[p]);
			if (rot[p]<0) rot[p]+=6.283;
			if (rot[p]>=6.283) rot[p]-=6.283;
			if (!pacman[p]) {xs[p]*=pow(0.99,dt); ys[p]*=pow(0.99,dt); rs[p]*=pow(0.9,dt);}
			x=(magneticfield[qx][qy]&15)-7;
			y=(magneticfield[qx][qy]>>4)-7;
			xs[p]+=0.07*x*dt; // Magnetic force
			ys[p]+=0.07*y*dt;
			// Ship interaction with grid
			for (efx=0;efx<M_PI*2;efx+=2.0/vr[p]) // Ship boundary loop
			{
				float cx=px[p]+vr[p]*cos(efx),cy=py[p]+vr[p]*sin(efx),rx,ry;
				int gx,gy,gs,gg;
				if (cx<0) cx+=xres;
				else if (cx>=xres) cx-=xres;
				if (cy<0) cy+=yres;
				else if (cy>=yres) cy-=yres;
				gx=(int)cx/20; gy=(int)cy/20;
				gs=grid[gx][gy]; gg=sattrib[gs];
				rx=cx-(gx*20+10); ry=cy-(gy*20+10);
				if (gg>=1 && gg<=5) // Ship bounces
				{
					float mul=1;
					int q=(rx<ry)+(rx<-ry)*2;
					if (pacman[p]) gg=4;
					if (gg==2) mul=0.5;
					else if (gg==3 || gg==5) mul=0.15;
					else if (gg==4) mul=0;
					if (q==0 && xs[p]<0 && !sattrib[grid[(gx+1)%32][gy]]) xs[p]*=-mul;
					else if (q==3 && xs[p]>0 && !sattrib[grid[(gx+31)%32][gy]]) xs[p]*=-mul;
					else if (q==1 && ys[p]<0 && !sattrib[grid[gx][(gy+1)%24]]) ys[p]*=-mul;
					else if (q==2 && ys[p]>0 && !sattrib[grid[gx][(gy+23)%24]]) ys[p]*=-mul;
					if (gg==5) shit[p]=-1; // Fatal block kills ship
				}
				if (gs==0x40 && !shopped[p]) // Enter the shop
				{
					int q,r;
					if (sound_working) for every_player(q) // Stop that annoying noise
						if ((r=thstatus[q])>=0)
						{
							stopsound(soundchannel[r]); killbuffer(soundchannel[r]);
							soundchannel[r]=makebuffer(100,10000);
							thstatus[q]=-1;
						}
					theshop(p,upgrade[p]);
					shopped[p]=100;
					replotgrid(grid,blox);
				}
				if (gs>>4==0xC && !heat[gx][gy]) // Movable block
				{
					int q,vx=-1,vy;
					char ok[0x100];
					for (n=0;n<0x100;n++) ok[n]=(n==0 || n>>4==0xD && !sattrib[n]);
					q=(rx<ry)+(rx<-ry)*2;
					if (q==0 && ok[grid[(gx+31)%32][gy]] && xs[p]>0.2) // Move left
						ucswap(&grid[vx=(gx+31)%32][vy=gy],&grid[gx][gy]);
					else if (q==3 && ok[grid[(gx+1)%32][gy]] && xs[p]<-0.2) // Move right
						ucswap(&grid[vx=(gx+1)%32][vy=gy],&grid[gx][gy]);
					else if (q==1 && ok[grid[gx][(gy+23)%24]] && ys[p]>0.2) // Move up
						ucswap(&grid[vx=gx][vy=(gy+23)%24],&grid[gx][gy]);
					else if (q==2 && ok[grid[gx][(gy+1)%24]] && ys[p]<-0.2) // Move down
						ucswap(&grid[vx=gx][vy=(gy+1)%24],&grid[gx][gy]);
					if (vx>=0)
					{
						if (grid[gx][gy]==0) plotbgsq(gx,gy,background,bgsub,bplot);
							else wrapbitb(gx*20,gy*20,blox[grid[gx][gy]]);
						wrapbitb(vx*20,vy*20,blox[grid[vx][vy]]);
						heat[vx][vy]=-p-1; // Block moved: Check for matches later
					}
				}
			}
		}
//		for (n=0;n<am;n++) if (!rnd(1000)) {at[n]=rnd(5)+1; as[n]=rnd(200)-100;} // Random mode
// Turrets and things
		for (x=0;x<32;x++) for (y=0;y<24;y++)
		{
			i=grid[x][y];
			if (i>=0x45 && i<=0x48) // There is a turret
			{
				double t=framestart*0.001+(x*x+y*y*y)*0.42;
				t/=1.0-heat[x][y]*0.033;
				if (fmod(t,2)<1)
				{
					int c,d;
					if (heat[x][y]<10) {c=2; d=10;}
					else if (heat[x][y]<20) {c=6; d=14;}
					else {c=12; d=14;}
					putpix(x*20+10,y*20+10,d); // The little flashing light
					putpix(x*20+9,y*20+10,c);
					putpix(x*20+11,y*20+10,c);
					putpix(x*20+10,y*20+9,c);
					putpix(x*20+10,y*20+11,c);
				}
				if (!rnd(25) && am<amax) // Turret fires
				{
					int victim=-1;
					float bdd=1e9;
					for every_player(p) if (!dead[p] && !subgame[p])
						if (ddfrom(px[p],py[p],x*20+10,y*20+10)<bdd)
						{
							bdd=ddfrom(px[p],py[p],x*20+10,y*20+10);
							victim=p;
						}
					ax[am]=x*20+10; ay[am]=y*20+10;
					if (victim>=0)
					{
						axs[am]=xfrom(px[victim]-(x*20+10))/sqrt(bdd);
						ays[am]=yfrom(py[victim]-(y*20+10))/sqrt(bdd);
					}
					else axs[am]=ays[am]=0;
					at[am]=5; as[am]=100;
				ar[am]=0; ars[am]=0;
					if (i==0x45 && axs[am]<-0.1) am--;
					else if (i==0x46 && ays[am]<-0.1) am--;
					else if (i==0x47 && axs[am]>0.1) am--;
					else if (i==0x48 && ays[am]>0.1) am--;
					else if (axs[am]==0 && ays[am]==0) am--;
					am++;
				}
			}
			else if (i>=0x49 && i<=0x4C) // MG turret
			{
				if (fmod(framesdone,3)<0.9 && am<amax) // Turret fires
				{
					int victim=-1;
					float bdd=1e9;
					for every_player(p) if (!dead[p] && !subgame[p])
						if (ddfrom(px[p],py[p],x*20+10,y*20+10)<bdd)
						{
							bdd=ddfrom(px[p],py[p],x*20+10,y*20+10);
							victim=p;
						}
					ax[am]=x*20+10; ay[am]=y*20+10;
					if (victim>=0)
					{
						axs[am]=xfrom(px[victim]-(x*20+10))/sqrt(bdd)*5;
						ays[am]=yfrom(py[victim]-(y*20+10))/sqrt(bdd)*5;
					}
					else axs[am]=ays[am]=0;
					if (fabs(axs[am])>fabs(ays[am]*2)) ays[am]=0;
					else if (fabs(ays[am])>fabs(axs[am]*2)) axs[am]=0;
					else if (axs[am]*ays[am]>0) ays[am]=axs[am]=(axs[am]+ays[am])/2;
					else axs[am]=-ays[am];
					at[am]=5; as[am]=40;
					ar[am]=2; ars[am]=0;
					if (i==0x49 && yfrom(py[victim]-y*20-10)>0) am--;
					else if (i==0x4A && xfrom(px[victim]-x*20-10)<0) am--;
					else if (i==0x4B && yfrom(py[victim]-y*20-10)<0) am--;
					else if (i==0x4C && xfrom(px[victim]-x*20-10)>0) am--;
					else if (axs[am]==0 && ays[am]==0) am--;
					am++;
				}
			}
			else if (i>=0x50 && i<=0x53) // WKobo-esque conduit
			{
				int xi=(x+1)%32,xd=(x+31)%32,yi=(y+1)%24,yd=(y+23)%24;
				n=(grid[xi][y]!=0)+(grid[xd][y]!=0)+(grid[x][yi]!=0)+(grid[x][yd]!=0);
				if (n<1+i-0x50 || heat[x][y]>0) // Disintegrating
				{
					if (em<emax)
					{
						ex[em]=x*20+rnd(20)-4; ey[em]=y*20+rnd(20)-4;
						exs[em]=eys[fm]=0;
						ef[em]=0; efs[em]=0.3; enf[em]=10;
						ea[em]=e_firesmall; er[em]=1;
						em++;
					}
					heat[x][y]+=idt;
				}
				if (heat[x][y]>((i==0x53)?rnd(200):10)) // Destroyed
				{
					grid[x][y]=heat[x][y]=0;
					plotbgsq(x,y,background,bgsub,bplot);
				}
			}
			else if (i>=0x54 && i<=0x57) // Spark turret
			{
				int xx,yy;
				heat[x][y]++;
				for (xx=0;xx<32;xx++) if (xfrom((xx-x)*20)<heat[x][y])
					for (yy=0;yy<24;yy++) if (!rnd(2)) if (grid[xx][yy]/4==0x54/4 && (xx!=x || yy!=y))
						if ((efx=ddfrom(xx*20+10,yy*20+10,x*20+10,y*20+10))<sq(heat[x][y]))
						{
							float dx=xfrom((xx-x)*20),dy=yfrom((yy-y)*20),z=0;
							efx=sqrt(efx); dx/=efx; dy/=efx;
							for (n=0;n<efx;n++)
							{
								int spx,spy;
								z+=frnd(2)-1;
								z*=1.0-1.0/(efx-n+1);
								putpix(spx=x*20+10+dx*n+dy*z,spy=y*20+10+dy*n-dx*z,11);
								if (getpix(spx+1,spy)!=11) putpix(spx+1,spy,9);
								if (getpix(spx-1,spy)!=11) putpix(spx-1,spy,9);
								if (getpix(spx,spy+1)!=11) putpix(spx,spy+1,9);
								if (getpix(spx,spy-1)!=11) putpix(spx,spy-1,9);
								if (n%5==0 || n>=efx-1) for every_player(p) if (am<amax)
									if (ddfrom(px[p],py[p],spx,spy)<sq(vr[p]))
									{
										ax[am]=spx; ay[am]=spy;
										at[am]=ar[am]=5; as[am]=1;
										axs[am]=ays[am]=ars[am]=0;
										am++;
									}
							}
							ncgisound(x*10+xx*10+10,n_spark,256,11025,0);
							if (!rnd(3)) heat[x][y]=heat[xx][yy]=0;
						}
			}
		}
		for (x=0;x<32;x++) for (y=0;y<24;y++)
			if (grid[x][y]==0xD7) // Power pill plotting
				circlefill(x*20+10,y*20+10,2.5+2.5*sq(sin(framestart*0.005)),15);
		for every_bullet(n) // Bullets interaction with grid
		{
			int gx=(int)bx[n]/20,gy=(int)by[n]/20,l=0;
			i=grid[gx][gy];
			if (i>=0x45 && i<=0x4C) for (efx=0;efx<=1 && !l;efx+=1.0/sqrt(sq(bxs[n])+sq(bys[n])))
			{
				int rx=(int)(bx[n]+efx*bxs[n])-gx*20,ry=(int)(by[n]+efx*bys[n])-gy*20;
				if (blox[i][2+rx+ry*20]!=255) // Turret takes damage
				{
					wrapbitc(gx*20,gy*20,blox[i],15);
					heat[gx][gy]++;
					if (heat[gx][gy]>=30) // Turret destroyed
					{
						if (grid[gx][gy]<=0x48) score[bo[n]]+=100;
							else score[bo[n]]+=200;
						heat[gx][gy]=grid[gx][gy]=0;
						plotbgsq(gx,gy,background,bgsub,bplot);
					}
					bm--;
					bx[n]=bx[bm]; by[n]=by[bm];
					bxs[n]=bxs[bm]; bys[n]=bys[bm];
					bt[n]=bt[bm]; bs[n]=bs[bm];
					bo[n]=bo[bm];
					n--; l=1;
				}
			}
			else if (i>>4==0xC && bt[n]>=8 && bt[n]<=9) // Prods a block
			{
				float rx=bx[n]-gx*20-10,ry=by[n]-gy*20-10;
				int q,vx=-1,vy=-1;
				q=(rx<ry)+(rx<-ry)*2;
				if (bt[n]==9) q=3-q;
				if (q==0 && grid[(gx+31)%32][gy]==0) // Move left
				{
					grid[vx=(gx+31)%32][vy=gy]=grid[gx][gy];
					grid[gx][gy]=0; plotbgsq(gx,gy,background,bgsub,bplot);
					wrapbitb(vx*20,vy*20,blox[grid[vx][vy]]);
				}
				else if (q==3 && grid[(gx+1)%32][gy]==0) // Move right
				{
					grid[vx=(gx+1)%32][vy=gy]=grid[gx][gy];
					grid[gx][gy]=0; plotbgsq(gx,gy,background,bgsub,bplot);
					wrapbitb(vx*20,vy*20,blox[grid[vx][vy]]);
				}
				else if (q==1 && grid[gx][(gy+23)%24]==0) // Move up
				{
					grid[vx=gx][vy=(gy+23)%24]=grid[gx][gy];
					grid[gx][gy]=0; plotbgsq(gx,gy,background,bgsub,bplot);
					wrapbitb(vx*20,vy*20,blox[grid[vx][vy]]);
				}
				else if (q==2 && grid[gx][(gy+1)%24]==0) // Move down
				{
					grid[vx=gx][vy=(gy+1)%24]=grid[gx][gy];
					grid[gx][gy]=0; plotbgsq(gx,gy,background,bgsub,bplot);
					wrapbitb(vx*20,vy*20,blox[grid[vx][vy]]);
				}
				if (vx>=0) heat[vx][vy]=-bo[n]-1; // Block moved: Check for matches later
				bm--;
				bx[n]=bx[bm]; by[n]=by[bm];
				bxs[n]=bxs[bm]; bys[n]=bys[bm];
				bt[n]=bt[bm]; bs[n]=bs[bm];
				bo[n]=bo[bm];
				n--;
			}
			else if (i>=0x4D && i<=0x4F) // Fragile walls
			{
				int limit=(i==0x4D)?1:(i==0x4E)?5:25;
				heat[gx][gy]++;
				if (heat[gx][gy]>=limit) // Destroyed
				{
					grid[gx][gy]=0; plotbgsq(gx,gy,background,bgsub,bplot);
				}
				bm--;
				bx[n]=bx[bm]; by[n]=by[bm];
				bxs[n]=bxs[bm]; bys[n]=bys[bm];
				bt[n]=bt[bm]; bs[n]=bs[bm];
				bo[n]=bo[bm];
				n--;
			}
		}
		for (x=0;x<32;x++)
			for (y=0;y<24;y++) if (-heat[x][y]>=1 && -heat[x][y]<=MAXP) // Block moved: Check for matches
			{
				int tx,ty;
				unsigned char flood[32][24];
				p=-heat[x][y]-1; heat[x][y]=0;
				k=grid[x][y];
				for (tx=0;tx<32;tx++) for (ty=0;ty<24;ty++) flood[tx][ty]=0;
				flood[x][y]=1; n=i=1;
				while (n>0) // Flood filling
				{
					n=0;
					for (tx=0;tx<32;tx++) for (ty=0;ty<24;ty++) if (grid[tx][ty]==k)
					{
						int ix=(tx+1)%32,iy=(ty+1)%24,dx=(tx+31)%32,dy=(ty+23)%24;
						if (flood[ix][ty] || flood[dx][ty] || flood[tx][iy] || flood[tx][dy])
						{
							n+=(flood[tx][ty]==0);
							flood[tx][ty]=1;
						}
					}
					i+=n;
				}
				if (i>=2) // Match occured: Delete blocks
				{
					for (tx=0;tx<32;tx++) for (ty=0;ty<24;ty++)
						if (flood[tx][ty]) heat[tx][ty]=(p<<16)+i*5;
					score[p]+=(int)(sqrt(i)*i)*5;
				}
			}
// Block puzzle plotting
		for (x=0;x<32;x++) for (y=0;y<24;y++)
			if (grid[x][y]>>4==0xC && heat[x][y]>0) // Disintegrating
			{
				int c=15;
				p=heat[x][y]>>16;
				if (p==0) c=132+5*rnd(5);
				else if (p==1) c=228+rnd(5);
				else if (p==2) c=248+rnd(5);
				else if (p==3) c=140+31*rnd(3);
				wrapbitc(x*20,y*20,blox[grid[x][y]],c);
				heat[x][y]--;
				if ((heat[x][y]&0xFFFF)<=0) // Finally disappears
				{
					heat[x][y]=grid[x][y]=0;
					plotbgsq(x,y,background,bgsub,bplot);
				}
			}
// Dust loop
		if ((int)(framesdone)%100==0) // Check for stationary dust
			for dust_particles(n) if (fxs[n]==0) if (fys[n]==0)
			{
				fxs[n]=frnd(2)-1;
				fys[n]=frnd(2)-1;
			}
		efy=pow(1.03,dt); i=Min(floor(dt+0.5),3);
		if (i==1) for dust_particles(n)
		{
			if (fx[n]<0 || fy[n]<0 || fx[n]>=xres || fy[n]>=yres)
			{
				fm--;
				fx[n]=fx[fm]; fy[n]=fy[fm];
				fxs[n]=fxs[fm]; fys[n]=fys[fm];
				fc[n]=fc[fm];
			}
			fx[n]+=fxs[n]*dt; fy[n]+=fys[n]*dt;
			fxs[n]*=efy; fys[n]*=efy;
			putpix(fx[n],fy[n],fc[n]);
		}
		else for dust_particles(n)
		{
			if (fx[n]<0 || fy[n]<0 || fx[n]>=xres || fy[n]>=yres)
			{
				fm--;
				fx[n]=fx[fm]; fy[n]=fy[fm];
				fxs[n]=fxs[fm]; fys[n]=fys[fm];
				fc[n]=fc[fm];
			}
			fx[n]+=fxs[n]*dt; fy[n]+=fys[n]*dt;
			fxs[n]*=efy; fys[n]*=efy;
			for (efx=0;efx<i;efx++)
				putpix(fx[n]+efx*fxs[n],fy[n]+efx*fys[n],fc[n]);
		}
		for (n=am;n<amax;n++) // Flags to zero for new objects
			afire[n]=aplas[n]=awhit[n]=0;
		for (n=0;n<amax;n++) ahit[n]=0; // No enemies hit
		for (n=0;n<bmax;n++) set(bhit,n,0); // No bullets used up
		cm=0; // No collisions so far
// Enemies plotting/movement loop
		for every_enemy(n)
		{
			float dd[MAXP],erad=radius(at[n],as[n],asz);
			ax[n]+=axs[n]*dt; ay[n]+=ays[n]*dt; ar[n]+=ars[n]*dt;
			ax[n]+=xres; ay[n]+=yres;
			while (ax[n]>=xres) ax[n]-=xres;
			while (ay[n]>=yres) ay[n]-=yres;
			// Enemy interaction with grid
			for (efx=0;efx<M_PI*2;efx+=4.0/erad)
			{
				float cx=ax[n]+erad*cos(efx),cy=ay[n]+erad*sin(efx);
				int gx,gy;
				if (cx<0) cx+=xres;
				else if (cx>=xres) cx-=xres;
				if (cy<0) cy+=yres;
				else if (cy>=yres) cy-=yres;
				gx=(int)cx/20; gy=(int)cy/20;
				if (eattrib[i=grid[gx][gy]]) // Enemy bounces (not straight through)
				{
					int q;
					cx-=gx*20+10; cy-=gy*20+10;
					q=(cx<cy)+(cx<-cy)*2;
					if (q==0 && axs[n]<=0) {axs[n]=fabs(axs[n]); ax[n]=(gx+1)*20+erad;}
					else if (q==3 && axs[n]>=0) {axs[n]=-fabs(axs[n]); ax[n]=gx*20-erad;}
					else if (q==1 && ays[n]<=0) {ays[n]=fabs(ays[n]); ay[n]=(gy+1)*20+erad;}
					else if (q==2 && ays[n]>=0) {ays[n]=-fabs(ays[n]); ay[n]=gy*20-erad;}
					if (i>>4==0xD && at[n]!=21) heat[gx][gy]=1; // Friction generates heat
				}
			}
			{// Emergency relocate of trapped enemy
				int gx=(int)ax[n]/20,gy=(int)ay[n]/20;
				if (eattrib[grid[gx][gy]] && eattrib[grid[(gx+1)%32][gy]] && eattrib[grid[(gx+31)%32][gy]]
					&& eattrib[grid[gx][(gy+1)%24]] && eattrib[grid[gx][(gy+23)%24]])
					while (eattrib[grid[gx][gy]])
					{
						ax[n]=frnd(xres); ay[n]=frnd(yres);
						gx=ax[n]/20; gy=ay[n]/20;
					}
				gx=(int)(ax[n]-10)/20; gy=(int)(ay[n]-10)/20;
				if (eattrib[grid[gx][gy]] && eattrib[grid[(gx+1)%32][gy]]
					&& eattrib[grid[gx][(gy+1)%24]] && eattrib[grid[(gx+1)%32][(gy+1)%24]])
					while (eattrib[grid[gx][gy]])
					{
						ax[n]=frnd(xres); ay[n]=frnd(yres);
						gx=ax[n]/20; gy=ay[n]/20;
					}
			}
			for every_player(p) dd[p]=ddfrom(px[p],py[p],ax[n],ay[n]);
			if (afire[n]) // On fire
			{
				int nffm=Min(ffm+5,ffmax);
				if (!rnd(150)) ahit[n]=afire[n];
				if (!rnd(300)) afire[n]=0;
				if (at[n]==10 && as[n]<2) nffm=Min(ffm+1+2*as[n],ffmax);
				if (at[n]==1 || at[n]==12)
				{
					nffm=Min(ffm+2*as[n]+2,ffmax);
					for new_flame(i)
					{
						float rad=0.5*(asz[as[n]]-1),th=frnd(M_PI*2);
						ffx[i]=ax[n]+rad*sin(th); ffy[i]=ay[n]+rad*cos(th);
						ffx[i]=(ffx[i]+xres)%xres; ffy[i]=(ffy[i]+yres)%yres;
						ffc[i]=firecolour();
					}
				}
				else for new_flame(i)
				{
					ffx[i]=ax[n]; ffy[i]=ay[n];
					ffc[i]=firecolour();
				}
				ffm=nffm;
			}
			if (aplas[n]) // Plasma fire
			{
				int nffm=Min(ffm+5,ffmax);
				if (!rnd(150)) ahit[n]=aplas[n];
				if (at[n]==10 && as[n]<2) nffm=Min(ffm+1+2*as[n],ffmax);
				if (at[n]==1 || at[n]==12)
				{
					nffm=Min(ffm+2*as[n]+2,ffmax);
					for new_flame(i)
					{
						float rad=0.5*(asz[as[n]]-1),th=frnd(M_PI*2);
						ffx[i]=ax[n]+rad*sin(th); ffy[i]=ay[n]+rad*cos(th);
						ffx[i]=(ffx[i]+xres)%xres; ffy[i]=(ffy[i]+yres)%yres;
						ffc[i]=plasmacolour();
					}
				}
				else for new_flame(i)
				{
					ffx[i]=ax[n]; ffy[i]=ay[n];
					ffc[i]=plasmacolour();
				}
				ffm=nffm;
			}
			if (awhit[n]) // White fire
			{
				int nffm=Min(ffm+4,ffmax);
				if (!rnd(300)) awhit[n]=0;
				if (!rnd(15)) ahit[n]=awhit[n];
				if (at[n]==10 && as[n]<2) nffm=Min(ffm+1+2*as[n],ffmax);
				if (at[n]==1 || at[n]==12)
				{
					nffm=Min(ffm+2*as[n]+2,ffmax);
					for new_flame(i)
					{
						float rad=0.5*(asz[as[n]]-1),th=frnd(M_PI*2);
						ffx[i]=ax[n]+rad*sin(th); ffy[i]=ay[n]+rad*cos(th);
						ffx[i]=(ffx[i]+xres)%xres; ffy[i]=(ffy[i]+yres)%yres;
						ffc[i]=15;
					}
				}
				else for new_flame(i)
				{
					ffx[i]=ax[n]; ffy[i]=ay[n];
					ffc[i]=15;
				}
				ffm=nffm;
			}

			for every_player(p)
				if (dd[p]<sq(erad+vr[p])) shit[p]=n+1;


			if (at[n]==1) // Asteroid type
			{
				float hd,vv;
				if (as[n]>20 || as[n]<1) as[n]=rnd(20)+1;
				hd=0.5*(asz[as[n]]-1);
				vv=axs[n]*axs[n]+ays[n]*ays[n];
				if (vv>1) {axs[n]*=0.99; ays[n]*=0.99;}
				wrapbit(ax[n]-hd,ay[n]-hd,ast[as[n]-1]);
				for every_bullet(i)
					if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(asz[as[n]]/2) && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==2) // Spinner type
			{
				float q;
				if (as[n]<1 || as[n]>666) as[n]=rnd(666)+1;
				for (q=0;q<M_PI*2;q+=M_PI*2/as[n])
					line(ax[n],ay[n],ax[n]+sin(q+ar[n])*10,ay[n]+cos(q+ar[n])*10,12);
				for (q=0;q<M_PI*2;q+=0.1)
					putpix(ax[n]+sin(q)*5,ay[n]+cos(q)*5,12);
				for every_bullet(i)
					if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(10) && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==3) // Flier type
			{
				if (as[n]==0) // Normal
				{
					float x1,y1,x2,y2;
					axs[n]=sin(ar[n]); ays[n]=cos(ar[n]);
					if (!rnd(100)) ars[n]=frnd(0.08)-0.04;
					if (fm<fmax && !rnd(2))
					{
						fx[fm]=ax[n]; fy[fm]=ay[n];
						fxs[fm]=frnd(0.2)-0.1-axs[n]*2; fys[fm]=frnd(0.2)-0.1-ays[n]*2;
						fc[fm]=14;
						fm++;
					}
					x1=ax[n]+sin(ar[n])*5; y1=ay[n]+cos(ar[n])*5;
					x2=ax[n]+sin(ar[n]+2.5)*5; y2=ay[n]+cos(ar[n]+2.5)*5;
					line(x1,y1,x2,y2,42);
					x2=ax[n]+sin(ar[n]-2.5)*5; y2=ay[n]+cos(ar[n]-2.5)*5;
					line(x1,y1,x2,y2,42);
				}
				else if (as[n]==1) // Pointy
				{
					float x1,y1,x2,y2;
					axs[n]=1.2*sin(ar[n]); ays[n]=1.2*cos(ar[n]);
					if (!rnd(50)) ars[n]=frnd(0.2)-0.1;
					if (fm<fmax && !rnd(2))
					{
						fx[fm]=ax[n]; fy[fm]=ay[n];
						fxs[fm]=frnd(0.2)-0.1-axs[n]*2; fys[fm]=frnd(0.2)-0.1-ays[n]*2;
						fc[fm]=11;
						fm++;
					}
					x1=ax[n]+sin(ar[n])*5; y1=ay[n]+cos(ar[n])*5;
					x2=ax[n]+sin(ar[n]+2.8)*5; y2=ay[n]+cos(ar[n]+2.8)*5;
					line(x1,y1,x2,y2,238);
					x2=ax[n]+sin(ar[n]-2.8)*5; y2=ay[n]+cos(ar[n]-2.8)*5;
					line(x1,y1,x2,y2,238);
				}
				else if (as[n]==2) // Splits into 3 normal ones
				{
					float x1,y1,x2,y2;
					axs[n]=sin(ar[n]); ays[n]=cos(ar[n]);
					if (!rnd(100)) ars[n]=frnd(0.08)-0.04;
					if (fm<fmax && !rnd(2))
					{
						fx[fm]=ax[n]; fy[fm]=ay[n];
						fxs[fm]=frnd(1)-0.5-axs[n]*2; fys[fm]=frnd(1)-0.5-ays[n]*2;
						fc[fm]=42;
						fm++;
					}
					x1=ax[n]+sin(ar[n])*5; y1=ay[n]+cos(ar[n])*5;
					x2=ax[n]+sin(ar[n]+2.5)*5; y2=ay[n]+cos(ar[n]+2.5)*5;
					line(x1,y1,x2,y2,14);
					x2=ax[n]+sin(ar[n]-2.5)*5; y2=ay[n]+cos(ar[n]-2.5)*5;
					line(x1,y1,x2,y2,14);
					x2=ax[n]-sin(ar[n])*5; y2=ay[n]-cos(ar[n])*5;
					line(x1,y1,x2,y2,14);
				}
				else if (as[n]==3) // Homes in on you
				{
					int victim; float bdd=1e9;
					float x1,y1,x2,y2;
					for every_player(p)
					{
						float dd=ddfrom(px[p],py[p],ax[n],ay[n]);
						if (dd<bdd) {bdd=dd; victim=p;}
					}
					axs[n]=(0.8+n%10*0.02)*sin(ar[n]);
					ays[n]=(0.8+n%10*0.02)*cos(ar[n]);
					if (axs[n]*yfrom(py[victim]-ay[n])-ays[n]*xfrom(px[victim]-ax[n])>0)
						ars[n]=-0.03; else ars[n]=0.03;
					if (fm<fmax && !rnd(2))
					{
						fx[fm]=ax[n]; fy[fm]=ay[n];
						fxs[fm]=frnd(0.2)-0.1-axs[n]; fys[fm]=frnd(0.2)-0.1-ays[n];
						fc[fm]=2;
						fm++;
					}
					x1=ax[n]; y1=ay[n];
					x2=ax[n]+sin(ar[n]+2.4)*5; y2=ay[n]+cos(ar[n]+2.4)*5;
					line(x1,y1,x2,y2,10);
					x2=ax[n]+sin(ar[n]-2.4)*5; y2=ay[n]+cos(ar[n]-2.4)*5;
					line(x1,y1,x2,y2,10);
				}
				else if (as[n]==4) // This rocket shoots at you
				{
					float x1,y1,x2,y2;
					axs[n]=sin(ar[n]); ays[n]=cos(ar[n]);
					if (!rnd(100)) ars[n]=frnd(0.08)-0.04;
					if (!rnd(75) && am<amax)
					{
						ax[am]=ax[n]; ay[am]=ay[n];
						axs[am]=sin(ar[n])*2; ays[am]=cos(ar[n])*2;
						ar[am]=ars[am]=0;
						as[am]=100;
						at[am]=5;
						am++;
					}
					if (fm<fmax && !rnd(2))
					{
						fx[fm]=ax[n]; fy[fm]=ay[n];
						fxs[fm]=frnd(0.2)-0.1-axs[n]*2; fys[fm]=frnd(0.2)-0.1-ays[n]*2;
						fc[fm]=12;
						fm++;
					}
					x1=ax[n]+sin(ar[n])*5; y1=ay[n]+cos(ar[n])*5;
					x2=ax[n]+sin(ar[n]+2.5)*5; y2=ay[n]+cos(ar[n]+2.5)*5;
					line(x1,y1,x2,y2,13);
					line(ax[n],ay[n],x2,y2,13);
					x2=ax[n]+sin(ar[n]-2.5)*5; y2=ay[n]+cos(ar[n]-2.5)*5;
					line(x1,y1,x2,y2,13);
					line(ax[n],ay[n],x2,y2,13);
				}
				else if (as[n]==5) // Splits into 4 splitters
				{
					float x1,y1,x2,y2;
					axs[n]=sin(ar[n]); ays[n]=cos(ar[n]);
					if (!rnd(100)) ars[n]=frnd(0.08)-0.04;
					if (fm<fmax && !rnd(2))
					{
						fx[fm]=ax[n]; fy[fm]=ay[n];
						fxs[fm]=frnd(1)-0.5-axs[n]*2; fys[fm]=frnd(1)-0.5-ays[n]*2;
						fc[fm]=7;
						fm++;
					}
					x1=ax[n]+sin(ar[n])*5; y1=ay[n]+cos(ar[n])*5;
					x2=ax[n]+sin(ar[n]+2.5)*5; y2=ay[n]+cos(ar[n]+2.5)*5;
					line(x1,y1,x2,y2,227);
					x2=ax[n]+sin(ar[n]-2.5)*5; y2=ay[n]+cos(ar[n]-2.5)*5;
					line(x1,y1,x2,y2,227);
					x2=ax[n]+sin(ar[n]+2.9)*5; y2=ay[n]+cos(ar[n]+2.9)*5;
					line(x1,y1,x2,y2,227);
					x2=ax[n]+sin(ar[n]-2.9)*5; y2=ay[n]+cos(ar[n]-2.9)*5;
					line(x1,y1,x2,y2,227);
				}
				else if (as[n]==6) // This rocket is evil
				{
					float x1,y1,x2,y2;
					axs[n]=sin(ar[n]); ays[n]=cos(ar[n]);
					if (!rnd(100)) ars[n]=frnd(0.08)-0.04;
					if ((int)(framesdone)%60==0 && am<amax)
					{
						ax[am]=ax[n]; ay[am]=ay[n];
						axs[am]=ays[am]=0;
						ar[am]=ar[n]; ars[am]=0;
						at[am]=3; as[am]=3;
						afire[am]=afire[n]; aplas[am]=aplas[n]; awhit[am]=awhit[n];
						am++;
					}
					if (fm<fmax && !rnd(2))
					{
						fx[fm]=ax[n]; fy[fm]=ay[n];
						fxs[fm]=frnd(0.2)-0.1-axs[n]*2; fys[fm]=frnd(0.2)-0.1-ays[n]*2;
						fc[fm]=28;
						fm++;
					}
					x1=ax[n]+sin(ar[n])*7; y1=ay[n]+cos(ar[n])*7;
					x2=ax[n]+sin(ar[n]+2.5)*7; y2=ay[n]+cos(ar[n]+2.5)*7;
					line(x1,y1,x2,y2,15);
					x2=ax[n]+sin(ar[n]-2.5)*7; y2=ay[n]+cos(ar[n]-2.5)*7;
					line(x1,y1,x2,y2,15);
				}
				else as[n]=rnd(7); // End of rocket types
				for every_bullet(i)
					if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(5) && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==4) // Saucer type
			{
				if (!rnd(100)) {axs[n]=frnd(2)-1; ays[n]=frnd(2)-1;}
				if (!rnd(75) && am<amax)
				{
					float xd,yd,dd;
					int victim=rnd(pm);
					ax[am]=ax[n]; ay[am]=ay[n];
					xd=xfrom(px[victim]-ax[n]),yd=yfrom(py[victim]-ay[n]),dd=xd*xd+yd*yd;
					dd=sqrt(dd); xd*=1.0/dd; yd*=1.0/dd;
					axs[am]=xd+axs[n]+frnd(1)-0.5; ays[am]=yd+ays[n]+frnd(1)-0.5;
					ar[am]=ars[am]=0;
					as[am]=100;
					at[am]=5;
					am++;
				}
				wrapbit(ax[n]-7,ay[n]-3,saucer);
				for every_bullet(i)
					if (fabs(xfrom(bx[i]-ax[n]))<8 && fabs(yfrom(by[i]-ay[n]))<4 && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==5) // Enemy weapons
			{
				if (ar[n]==0) wrapbit(ax[n]-1,ay[n]-1,redbull); // Red bullets
				else if (ar[n]==1) // White cross thingy
				{
					putpix(ax[n],ay[n],15);
					putpix(ax[n]-1,ay[n]-1,15);
					putpix(ax[n]-1,ay[n]+1,15);
					putpix(ax[n]+1,ay[n]-1,15);
					putpix(ax[n]+1,ay[n]+1,15);
				}
				else if (ar[n]==2) // MG enemy bullet
					wrapbit(ax[n]-0.5,ay[n]-0.5,mbullet);
/*cj*/			else if (ar[n]==3) // Flame turret flame bullet
				{
					int nffm=Min(ffm+10,ffmax);
					for new_flame(i)
					{
						ffx[i]=ax[n]; ffy[i]=ay[n];
						ffc[i]=firecolour();
					}
					ffm=nffm;
				}
/*/cj*/
				else if (ar[n]==5); // Invisible thing from nowhere
				else ar[n]=rnd(3);
				as[n]=noneg(as[n]-idt);
				if (as[n]<=0) ahit[n]=1+MAXP;
			}
			else if (at[n]==6) // Homer type (DOH!)
			{
				float rad;
				unsigned char col;
				float oth,th,nth=1e6,x1,y1,x2,y2;
				if (as[n]<1 || as[n]>50) as[n]=rnd(50)+1;
				if (as[n]==1) // Homes in
				{
					int victim=-1;
					for every_player(p) if (!dead[p] && !subgame[p])
					{
						rad=ddfrom(px[p],py[p],ax[n],ay[n]);
						if (rad<nth) {nth=rad; victim=p;}
					}
					if (victim>=0)
					{
						float dx=xfrom(px[victim]-ax[n]),dy=yfrom(py[victim]-ay[n]);
						rad=sqrt(dx*dx+dy*dy);
						axs[n]+=0.01*dx/rad; ays[n]+=0.01*dy/rad;
					}
				}
				rad=axs[n]*axs[n]+ays[n]*ays[n];
				if (rad>1) {axs[n]*=0.99; ays[n]*=0.99;}
				col=56-as[n];
				for (oth=ar[n];oth<ar[n]+M_PI*0.66;oth+=M_PI*0.67/as[n])
					for (th=0;th<M_PI*2;th=nth)
					{
						nth=th+M_PI*0.67;
						x1=ax[n]+sin(oth+th)*6; y1=ay[n]+cos(oth+th)*6;
						x2=ax[n]+sin(oth+nth)*6; y2=ay[n]+cos(oth+nth)*6;
						line(x1,y1,x2,y2,col);
					}
				for every_bullet(i)
					if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(6) && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==7) // Pentagram type
			{
				float p,q;
				if (as[n]<1 || as[n]>42) as[n]=rnd(42)+1;
				for (p=0;p<M_PI*2;p+=M_PI*2/as[n])
					for (q=0;q<M_PI*2;q+=M_PI*2/as[n])
						if (q>p)
						{
							float pp=p+ar[n],qq=q+ar[n];
							line(ax[n]+sin(pp)*3*as[n],ay[n]+cos(pp)*3*as[n],
									ax[n]+sin(qq)*3*as[n],ay[n]+cos(qq)*3*as[n],(rnd(4))?15:10);
						}
				for every_bullet(i)
					if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(3*as[n]) && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==8) // Resons [lit: 'thing - ies']
			{
				float tc[4]={1,0.707,0,-0.707},ts[4]={0,0.707,1,0.707};
				ax[n]+=ars[n]*as[n]*cos(ar[n])+xres; ay[n]+=ars[n]*as[n]*sin(ar[n])+yres;
				while (ax[n]>=xres) ax[n]-=xres; while (ay[n]>=yres) ay[n]-=yres;
				for (i=0;i<4;i++)
					line(ax[n]-frnd(5)*tc[i],ay[n]-frnd(5)*ts[i],
							ax[n]+frnd(5)*tc[i],ay[n]+frnd(5)*ts[i],(rnd(5))?15:11);
				for every_bullet(i)
					if (ddfrom(bx[i],by[i],ax[n],ay[n])<sq(5) && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==9) // Time-reversing ship
			{
				char rev=0;
				for every_bullet(i)
					if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(50)) rev=1;
				for every_player(p)
					if (ddfrom(px[p],py[p],ax[n],ay[n])<sq(50)) rev=1;
				if (rev)
				{
					int nffm=Min(ffm+80,ffmax);
					for every_enemy(i) if (ddfrom(ax[n],ay[n],ax[i],ay[i])<sq(50) && i!=n)
						{axs[i]=-axs[i]; ays[i]=-ays[i]; ars[i]=-ars[i];}
					for every_bullet(i) if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(50))
						{bxs[i]=-bxs[i]; bys[i]=-bys[i];}
					for dust_particles(i) if (ddfrom(ax[n],ay[n],fx[i],fy[i])<sq(50))
						{fxs[i]=-fxs[i]; fys[i]=-fys[i];}
					for every_item(i) if (ddfrom(ax[n],ay[n],ix[i],iy[i])<sq(50))
						{ixs[i]=-ixs[i]; iys[i]=-iys[i];}
					for every_player(p)
						if (ddfrom(ax[n],ay[n],px[p],py[p])<sq(50))
							{xs[p]*=-1; ys[p]*=-1; rs[p]*=-1;}
					for new_flame(i)
					{
						ffx[i]=ax[n]+50*cos(M_PI*2*i/80);
						ffy[i]=ay[n]+50*sin(M_PI*2*i/80);
						ffc[i]=7;
					}
					ffm=nffm;
				}
				wrapbit(ax[n]-10,ay[n]-7,timership);
				for every_bullet(i)
					if (fabs(xfrom(bx[i]-ax[n]))<=10 && fabs(yfrom(by[i]-ay[n]))<=8 && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==10) // Macropixels
			{
				int step;
				if (as[n]>=5 || as[n]<0) as[n]=rnd(5);
				step=4*pot[as[n]];
				if (!rnd(10)) // Move
				{
					int ox=ax[n],oy=ay[n];
					i=rnd(4);
					if (i==0) ax[n]+=step;
					else if (i==1) ax[n]-=step;
					else if (i==2) ay[n]+=step;
					else if (i==3) ay[n]-=step;
					ax[n]=(int)(ax[n]+xres)%xres;
					ay[n]=(int)(ay[n]+yres)%yres;
					k=0;
					for (i=-1;i<2;i+=2) for (j=-1;j<2;j+=2)
					{
						int gx=(int)((ax[n]+i*step/2+xres)/20.0)%32,
							gy=(int)((ay[n]+j*step/2+yres)/20.0)%24;
						if (eattrib[grid[gx][gy]]) k=1;
					}
					if (k) {ax[n]=ox; ay[n]=oy;}
				}
				axs[n]=ays[n]=0;
				wrapbit(ax[n]-step/2,ay[n]-step/2,macpix[as[n]]);
				for every_bullet(i)
					if (fabs(xfrom(bx[i]-ax[n]))<=(step+3)/2 && fabs(yfrom(by[i]-ay[n]))<=(step+3)/2
						&& cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==11) // Self-replicating mines
			{
				as[n]-=idt;
				if (as[n]<0)
				{
					float p=0;
					for every_enemy(i) if (at[i]==11 && i!=n)
						p+=1e6/(ddfrom(ax[n],ay[n],ax[i],ay[i])+1);
					if (p>1000 || am>=amax) as[n]=0;
				}
				if (as[n]<-150 && am<amax)
				{
					float fx=frnd(1)-0.5,fy=frnd(1)-0.5;
					ax[am]=ax[n]; ay[am]=ay[n];
					axs[am]=axs[n]+fx; ays[am]=ays[n]+fy;
					axs[n]-=fx; ays[n]-=fy;
					ar[am]=ars[am]=as[am]=as[n]=0;
					at[am]=11;
					afire[am]=afire[n];
					aplas[am]=aplas[n];
					awhit[am]=awhit[n];
					am++;
				}
				if (sq(axs[n])+sq(ays[n])>0.01) {axs[n]*=0.99; ays[n]*=0.99;}
				wrapbit(ax[n]-6,ay[n]-6,(abs(as[n])%2)?glowmine:mine);
				for every_bullet(i)
					if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(7) && cm<cmax)
					{
						if (as[n])
						{
							cola[cm]=n; colb[cm]=i;
							cm++;
						}
						else // Bullet gets absorbed
						{
							as[n]=50;
							axs[n]+=bxs[i]/5; ays[n]+=bys[i]/5;
							cola[cm]=-1; colb[cm]=i;
							cm++;
						}
					}
			}
			else if (at[n]==12) // Recoalescing asteroids
			{
				float hd;
				int mi,mn;
				if (as[n]>20 || as[n]<1) as[n]=rnd(20)+1;
				hd=0.5*(asz[as[n]]-1);
				mn=legalsize[as[n]];
				wrapbit(ax[n]-hd,ay[n]-hd,ast[as[n]-1]);
				for every_enemy(i) if (at[i]==12 && i!=n &&
											islegalsize((mi=legalsize[as[i]])+mn))
				{
					float dx=xfrom(ax[i]-ax[n]),dy=yfrom(ay[i]-ay[n]),rad=dx*dx+dy*dy;
					if (rad<sq(0.5*(asz[as[n]]+asz[as[i]]))) // Asteroids merge
					{
						ax[n]-=ax[i]; ay[n]-=ay[i];
						if (ax[n]>xres/2) ax[n]-=xres;
						if (ax[n]<-xres/2) ax[n]+=xres;
						if (ay[n]>yres/2) ay[n]-=yres;
						if (ay[n]<-yres/2) ay[n]+=yres;
						ax[n]*=mn/(mn+mi);
						ay[n]*=mn/(mn+mi);
						ax[n]+=ax[i]; ay[n]+=ay[i];
						for every_item(j) // Items reabsorbed
							if (ddfrom(ax[n],ay[n],ix[j],iy[j])<mn+mi)
							{
								im--;
								ix[n]=ix[im]; iy[n]=iy[im];
								ixs[n]=ixs[im]; iys[n]=iys[im];
								it[n]=it[im];
							}
						axs[n]=(axs[n]*mn+axs[i]*mi)/(mn+mi);
						ays[n]=(ays[n]*mn+ays[i]*mi)/(mn+mi);
						as[n]=islegalsize(mn+mi);
						afire[n]|=afire[i];
						aplas[n]|=aplas[i];
						awhit[n]|=awhit[i];
						am--;
						ax[i]=ax[am]; ay[i]=ay[am];
						axs[i]=axs[am]; ays[i]=ays[am];
						ar[i]=ar[am]; ars[i]=ars[am];
						as[i]=as[am]; at[i]=at[am];
						afire[i]=afire[am];
						aplas[i]=aplas[am];
						awhit[i]=awhit[am];
					}
					else
					{
						rad*=sqrt(rad);
						axs[n]+=mi*dx/rad;
						ays[n]+=mi*dy/rad;
						axs[i]-=mn*dx/rad;
						ays[i]-=mn*dy/rad;
					}
				}
				hd=axs[n]*axs[n]+ays[n]*ays[n];
				if (hd>1) {axs[n]*=0.99; ays[n]*=0.99;}
				for every_bullet(i)
					if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(asz[as[n]]/2) && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==13) // Pulsar
			{
				int nfm=Min(fm+5,fmax);
				float dx,dy,rad;
				if (axs[n]*axs[n]+ays[n]*ays[n]>1) {axs[n]*=0.9; ays[n]*=0.9;}
				if (ars[n]>0.05) ars[n]*=0.997;
				wrapbit(ax[n]-5,ay[n]-5,pulsar);
				for new_dust(i)
				{
					fx[i]=ax[n]; fy[i]=ay[n];
					fxs[i]=axs[n]+(frnd(1)+3)*sin(ar[n]+frnd(0.1));
					fys[i]=ays[n]+(frnd(1)+3)*cos(ar[n]+frnd(0.1));
					fc[i]=(rnd(2))?15:11;
				}
				fm=nfm;
				for every_enemy(i) if (i!=n && at[i]!=10)
				{
					dx=xfrom(ax[n]-ax[i]),dy=yfrom(ay[n]-ay[i]),rad=dx*dx+dy*dy;
					if (rad<100) rad=100;
					axs[i]+=dx/rad;
					ays[i]+=dy/rad;
				}
				for every_bullet(i)
				{
					dx=xfrom(ax[n]-bx[i]),dy=yfrom(ay[n]-by[i]),rad=dx*dx+dy*dy;
					bxs[i]+=10.0*dx/rad;
					bys[i]+=10.0*dy/rad;
				}
				for every_player(p)
				{
					dx=xfrom(ax[n]-px[p]),dy=yfrom(ay[n]-py[p]),rad=dx*dx+dy*dy;
					if (rad<100) rad=100;
					xs[p]+=1.0*dx/rad; ys[p]+=1.0*dy/rad;
				}
				for every_bullet(i)
					if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(6) && cm<cmax)
					{
						ars[n]+=0.1;
						cola[cm]=-1; colb[cm]=i;
						if (ars[n]>0.8) cola[cm]=n;
						cm++;
					}
			}
			else if (at[n]==14) // Rubber ducks
			{
				if (sq(axs[n])+sq(ays[n])>2) {axs[n]*=0.95; ays[n]*=0.95;}
				wrapbit(ax[n]-10,ay[n]-7.5,rubberduck);
				if (!rnd(500))
				{
					int nam=Min(am+1,amax);
					for (i=am;i<nam;i++)
					{
						ax[i]=ax[n]; ay[i]=ay[n];
						axs[i]=axs[n]+frnd(2)-1; ays[i]=ays[n]+frnd(2)-1;
						ar[i]=ars[i]=0; at[i]=15;
						as[i]=0;
					}
					am=nam;
				}
				for every_bullet(i)
					if (fabs(xfrom(bx[i]-ax[n]))<10 && fabs(yfrom(by[i]-ay[n]))<7.5 && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==15) // Duck eggs
			{
				if (sq(axs[n])+sq(ays[n])>2) {axs[n]*=0.95; ays[n]*=0.95;}
				wrapbit(ax[n]-2,ay[n]-3,duckegg);
				if (!rnd(150) && cm<cmax)
				{
					int nfm=Min(fm+200,fmax);
					for every_enemy(i) if (i!=n)
					{
						float dx=xfrom(ax[i]-ax[n]),dy=yfrom(ay[i]-ay[n]),dd=dx*dx+dy*dy;
						if (dd<5000) dd=5000;
						axs[i]+=1e4*dx/mass(at[i],as[i])/sqrt(dd)/dd;
						ays[i]+=1e4*dy/mass(at[i],as[i])/sqrt(dd)/dd;
					}
					for every_bullet(i)
					{
						float dx=xfrom(bx[i]-ax[n]),dy=yfrom(by[i]-ay[n]),dd=dx*dx+dy*dy;
						if (dd<5000) dd=5000;
						bxs[i]+=1e4*dx/sqrt(dd)/dd; bys[i]+=1e4*dy/sqrt(dd)/dd;
					}
					for every_item(i)
					{
						float dx=xfrom(ix[i]-ax[n]),dy=yfrom(iy[i]-ay[n]),dd=dx*dx+dy*dy;
						if (dd<5000) dd=5000;
						ixs[i]+=1e4*dx/sqrt(dd)/dd; iys[i]+=1e4*dy/sqrt(dd)/dd;
					}
					for dust_particles(i)
					{
						float dx=xfrom(fx[i]-ax[n]),dy=yfrom(fy[i]-ay[n]),dd=dx*dx+dy*dy;
						if (dd<5000) dd=5000;
						fxs[i]+=3e4*dx/sqrt(dd)/dd; fys[i]+=3e4*dy/sqrt(dd)/dd;
					}
					for every_player(p)
					{
						float dx=xfrom(px[p]-ax[n]),dy=yfrom(py[p]-ay[n]),dd=dx*dx+dy*dy;
						if (dd<5000) dd=5000;
						xs[p]+=1e4*dx/sqrt(dd)/dd; ys[p]+=1e4*dy/sqrt(dd)/dd;
					}
					for new_dust(i)
					{
						float rad=frnd(2)+1e-6,th=frnd(M_PI*2);
						fx[i]=ax[n]; fy[i]=ay[n];
						fxs[i]=rad*cos(th); fys[i]=rad*sin(th)*1.5;
						fc[i]=227;
					}
					fm=nfm;
					cola[cm]=n; colb[cm]=-1;
					cm++;
				}
				for every_bullet(i)
					if (fabs(xfrom(bx[i]-ax[n]))<3 && fabs(yfrom(by[i]-ay[n]))<4 && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==16) // Defender (TM) Bomber
			{
				float dd;
				axs[n]+=frnd(0.2)-0.1; ays[n]+=frnd(0.2)-0.1;
				dd=sqrt(sq(axs[n])+sq(ays[n]));
				if (dd>2) {axs[n]/=dd; ays[n]/=dd;}
				if (!rnd(20) && am<amax)
				{
					ax[am]=ax[n]; ay[am]=ay[n];
					axs[am]=ays[am]=0;
					ar[am]=1; ars[am]=0;
					as[am]=200;
					at[am]=5;
					am++;
				}
				wrapbit(ax[n]-5,ay[n]-5,bomber);
				for every_bullet(i)
					if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(7) && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==17) // Defender (TM) Baiter
			{
				float bdd; int victim;
				if (as[n]<0 || as[n]>100) as[n]=rnd(101);
				for every_player(p)
				{
					float dd=ddfrom(ax[n],ay[n],px[p],py[p]);
					if (dd<bdd || p==0) {bdd=dd; victim=p;}
				}
				if (xfrom(ax[n]-px[victim])<0) axs[n]+=0.005; else axs[n]-=0.005;
				if (yfrom(ay[n]-py[victim])<0) ays[n]+=0.001; else ays[n]-=0.001;
				axs[n]*=pow(0.999,dt); ays[n]*=pow(0.999,dt);
				wrapbit(ax[n]-10,ay[n]-2,baiter);
				for (i=0;i<5*as[n];i++)
				{
					float th=M_PI*2*i/(5*as[n])+framestart*0.01;
					x=ax[n]+cos(th)*10; y=ay[n]+sin(th)*10;
					x=(x+xres)%xres; y=(y+yres)%yres;
					putpix(x,y,11);
				}
				for every_bullet(i)
				{
					if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(10) && as[n]>0 && cm<cmax)
					{// Its refraction shields
						cola[cm]=-1; colb[cm]=i;
						cm++;
						if (am<amax) // Return fire!
						{
							ax[am]=bx[i]; ay[am]=by[i];
							axs[am]=-bxs[i]; ays[am]=-bys[i];
							ar[am]=ars[am]=0;
							at[am]=5; as[am]=100;
							am++;
						}
						as[n]--;
					}
					else if (fabs(xfrom(ax[n]-bx[i]))<10 && fabs(yfrom(ay[n]-by[i]))<2.5 && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
				}
			}
			else if (at[n]==18) // Homing spinner
			{
				float rad,oth,th,nth=1e6,x1,y1,x2,y2;
				int victim=-1;
				for every_player(p) if (!dead[p] && !subgame[p])
				{
					rad=ddfrom(px[p],py[p],ax[n],ay[n]);
					if (rad<nth) {nth=rad; victim=p;}
				}
				if (victim>=0)
				{
					float dx,dy;
					dx=xfrom(px[victim]-ax[n]);
					dy=yfrom(py[victim]-ay[n]);
					rad=sqrt(dx*dx+dy*dy);
					axs[n]=0.9*axs[n]+0.1*dx/rad;
					ays[n]=0.9*ays[n]+0.1*dy/rad;
				}
				for every_enemy(i) if (at[i]==18 && i!=n) // Pauli-Johnson exclusion principle
				{
					float dx=xfrom(ax[n]-ax[i]),dy=yfrom(ay[n]-ay[i]),dd=dx*dx+dy*dy;
					axs[n]+=dx/dd; ays[n]+=dy/dd;
				}
				ars[n]-=ar[n]/as[n]; // Wobbling motion
				for (oth=ar[n];oth<ar[n]+M_PI*0.67;oth+=M_PI*0.23)
					for (th=0;th<M_PI*2;th=nth)
					{
						nth=th+M_PI*0.67;
						x1=ax[n]+sin(oth+th)*7; y1=ay[n]+cos(oth+th)*7;
						x2=ax[n]+sin(oth+nth)*14; y2=ay[n]+cos(oth+nth)*14;
						line(x1,y1,x2,y2,68+sin(oth+th)*Min(rad/14,12));
					}
				for every_bullet(i)
					if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(20) && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==19) // Flamethrower boss
			{
				float oth,th,nth,x1,y1,x2,y2;
				oth=ar[n];
				for (th=0;th<M_PI*2;th=nth)
				{
					nth=th+M_PI*0.67;
					x1=ax[n]+sin(oth+th)*20; y1=ay[n]+cos(oth+th)*20; // Big dark triangle
					x2=ax[n]+sin(oth+nth)*20; y2=ay[n]+cos(oth+nth)*20;
					line(x1,y1,x2,y2,130); line(ax[n],ay[n],x2,y2,130);
					x1=ax[n]+sin(-oth-th-M_PI)*14; y1=ay[n]+cos(-oth-th-M_PI)*14; // Smaller lighter triangle
					x2=ax[n]+sin(-oth-nth-M_PI)*14; y2=ay[n]+cos(-oth-nth-M_PI)*14;
					line(x1,y1,x2,y2,131); line(ax[n],ay[n],x2,y2,131);
					x1=ax[n]+sin(2*(oth+th+M_PI))*7; y1=ay[n]+cos(2*(oth+th+M_PI))*7; // Central triangles
					x2=ax[n]+sin(2*(oth+nth+M_PI))*7; y2=ay[n]+cos(2*(oth+nth+M_PI))*7;
					line(x1,y1,x2,y2,161);
					x1=ax[n]+sin(2*(oth+th+M_PI))*4; y1=ay[n]+cos(2*(oth+th+M_PI))*4;
					x2=ax[n]+sin(2*(oth+nth+M_PI))*4; y2=ay[n]+cos(2*(oth+nth+M_PI))*4;
					line(x1,y1,x2,y2,162);
				}
				for (i=1;i<=3;i++) // Guns
					if (as[n]>i*5)
						for (th=0;th<M_PI*2;th=nth)
						{
							nth=th+M_PI*0.67;
							x1=ax[n]+sin(-5*(oth+th))*6+sin(oth+i*M_TWOPI/3)*20;
							y1=ay[n]+cos(-5*(oth+th))*6+cos(oth+i*M_TWOPI/3)*20;
							x2=ax[n]+sin(-5*(oth+nth))*6+sin(oth+i*M_TWOPI/3)*20;
							y2=ay[n]+cos(-5*(oth+nth))*6+cos(oth+i*M_TWOPI/3)*20;
							line(x1,y1,x2,y2,132);
						}
					else
					{
						int nffm=Min(ffm+3,ffmax);
						for new_flame(j)
						{
							ffx[j]=ax[n]+sin(oth+i*M_TWOPI/3)*20; ffy[j]=ay[n]+cos(oth+i*M_TWOPI/3)*20;
							ffc[j]=155+rnd(2);
						}
						ffm=nffm;
					}
				if (!rnd(200) && am<amax) // Fires red triangles at you
				{
					ax[am]=ax[n]; ay[am]=ay[n];
					axs[am]=frnd(8)-4; ays[am]=frnd(8)-4;
					ar[am]=frnd(M_TWOPI); ars[am]=frnd(0.2)-0.1;
					as[am]=30;
					afire[am]=aplas[am]=awhit[am]=ahit[am]=0;
					at[am]=20;
					am++;
				}
				for every_bullet(i) if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(20) && cm<cmax)
				{
					if (as[n]>0) // Survives hit
					{
						cola[cm]=-1; colb[cm]=i;
						cm++;
						as[n]--;
						for (j=1;j<=3;j++) // Emits flame
						{
							unsigned char col[6]={129,155,186,217,247,252};
							int nffm=Min(ffm+80,ffmax);
							for new_flame(k)
							{
								ffx[k]=ax[n]+sin(oth+j*M_TWOPI/3)*20+frnd(10)-5;
								ffy[k]=ay[n]+cos(oth+j*M_TWOPI/3)*20+frnd(10)-5;
								ffc[k]=col[rnd(6)];
							}
							ffm=nffm;
						}
						for (j=0;j<3;j++) if (am<amax) // Produces triangle things
						{
							ax[am]=ax[n]+frnd(20)-10; ay[am]=ay[n]+frnd(20)-10;
							axs[am]=frnd(8)-4; ays[am]=frnd(8)-4;
							ar[am]=ars[am]=0.1;
							as[am]=300;
							afire[am]=aplas[am]=awhit[am]=ahit[am]=0;
							at[am]=20;
							am++;
						}
					}
					else // Does not survive
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
				}
			}
			else if (at[n]==20) // Red triangle
			{
				float rad,oth,th,nth=1e6,x1,y1,x2,y2;
				int nffm,victim=-1;
				for every_player(p) if (!dead[p] && !subgame[p])
				{
					rad=ddfrom(px[p],py[p],ax[n],ay[n]);
					if (rad<nth) {nth=rad; victim=p;}
				}
				if (victim>=0)
				{
					float dx,dy;
					dx=xfrom(px[victim]-ax[n]);
					dy=yfrom(py[victim]-ay[n]);
					rad=sqrt(dx*dx+dy*dy);
					axs[n]=0.5*axs[n]+0.5*dx/rad;
					ays[n]=0.5*ays[n]+0.5*dy/rad;
				}
				else if (sq(axs[n])+sq(ays[n])>1) {axs[n]/=1.1; ays[n]/=1.1;}
				axs[n]+=frnd(1)-0.5; ays[n]+=frnd(1)-0.5;
				for every_enemy(i) if (at[i]==20 && i!=n) // Pauli-Johnson exclusion principle
				{
					float dx=xfrom(ax[n]-ax[i]),dy=yfrom(ay[n]-ay[i]),dd=dx*dx+dy*dy;
					axs[n]+=dx/dd; ays[n]+=dy/dd;
				}
				nffm=Min(ffm+5,ffmax);
				for new_flame(i) // Mauve flame
				{
					ffx[i]=ax[n]; ffy[i]=ay[n];
					ffc[i]=155+rnd(2);
				}
				ffm=nffm;
				oth=ar[n];
				for (th=0;th<M_PI*2;th=nth)
				{
					nth=th+M_PI*0.67;
					x1=ax[n]+sin(oth+th)*6; y1=ay[n]+cos(oth+th)*6;
					x2=ax[n]+sin(oth+nth)*6; y2=ay[n]+cos(oth+nth)*6;
					line(x1,y1,x2,y2,132);
				}
				for every_bullet(i) if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(6) && cm<cmax)
				{
					cola[cm]=n; colb[cm]=i;
					cm++;
				}
			}
			else if (at[n]==21) // Ghost in pacman
			{
				float ghostspeed=log(level*10)/log(1000);
				if (as[n] && (int)framesdone%2==0) wrapbitc(ax[n]-10,ay[n]-10,ghost,11);
					else wrapbit(ax[n]-10,ay[n]-10,ghost);
				if (ays[n]>axs[n])
				{
					if (ays[n]>-axs[n]) {axs[n]=0; ays[n]=ghostspeed;}
					else {axs[n]=-ghostspeed; ays[n]=0;}
				}
				else
				{
					if (ays[n]>-axs[n]) {axs[n]=ghostspeed; ays[n]=0;}
					else {axs[n]=0; ays[n]=-ghostspeed;}
				}
				if (axs[n]==0) ax[n]=floor(ax[n]/20)*20+10;
				if (ays[n]==0) ay[n]=floor(ay[n]/20)*20+10;
				if (sq(fmod(ax[n],20)-10)+sq(fmod(ay[n],20)-10)<sq(ghostspeed))
				{
					float oxs=axs[n],oys=ays[n];
					int nx,ny,gx=floor(ax[n]/20),gy=floor(ay[n]/20),
						gxi=(gx+1)%32,gxd=(gx+31)%32,gyi=(gy+1)%24,gyd=(gy+23)%24;
					if (grid[gxi][gy]==0xDE || grid[gxd][gy]==0xDE ||
						grid[gx][gyi]==0xDE || grid[gx][gyd]==0xDE) as[n]=0;
					eattrib[0]=1; for (i=0x20;i<0x28;i++) eattrib[i]=1;
					j=eattrib[grid[gxi][gy]]+eattrib[grid[gxd][gy]]+
						eattrib[grid[gx][gyi]]+eattrib[grid[gx][gyd]];
					k=0;
					if (j==4 || eattrib[grid[gx][gy]]) // Stranded ghost removal
					{
						if (!eattrib[grid[gxi][gy]]) {axs[n]=ghostspeed; ays[n]=0;}
						else if (!eattrib[grid[gxd][gy]]) {axs[n]=-ghostspeed; ays[n]=0;}
						else if (!eattrib[grid[gx][gyi]]) {axs[n]=0; ays[n]=ghostspeed;}
						else if (!eattrib[grid[gx][gyd]]) {axs[n]=0; ays[n]=-ghostspeed;}
						else if (cm<cmax) {cola[cm]=n; colb[cm++]=-1;}
						ax[n]=fmod(ax[n]+xres,xres); ay[n]=fmod(ay[n]+yres,yres);
					}
					else do // Change direction
					{
						i=rnd(4); k++;
						if (i==0) {axs[n]=ghostspeed; ays[n]=0;}
						else if (i==1) {axs[n]=-ghostspeed; ays[n]=0;}
						else if (i==2) {axs[n]=0; ays[n]=ghostspeed;}
						else if (i==3) {axs[n]=0; ays[n]=-ghostspeed;}
						nx=((int)floor(ax[n]/20+axs[n]/ghostspeed)+32)%32;
						ny=((int)floor(ay[n]/20+ays[n]/ghostspeed)+24)%24;
					}
					while (k<100 && (eattrib[grid[nx][ny]] || axs[n]==-oxs && ays[n]==-oys && j<3));
					eattrib[0]=0; for (i=0x20;i<0x28;i++) eattrib[i]=0;
				}
				if (as[n]) for every_player(p) // Edible ghost
					if (ddfrom(px[p],py[p],ax[n],ay[n])<sq(20) && cm<cmax)
					{
						score[p]+=100; upgrade[p][9]++;
						cola[cm]=n; colb[cm]=-1;
						cm++;
					}
				for every_bullet(i)
					if (fabs(xfrom(ax[n]-bx[i]))<10 && fabs(yfrom(ay[n]-by[i]))<10 && cm<cmax)
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			else if (at[n]==22) // Subgame N balls
			{
				wrapbit(ax[n]-5,ay[n]-5,nball[as[n]]);
				for every_bullet(i)
					if (ddfrom(ax[n],ay[n],bx[i],by[i])<sq(5+1))
					{
						cola[cm]=n; colb[cm]=i;
						cm++;
					}
			}
			for every_player(p) if (shit[p] && !upgrade[p][9] && !dead[p] && !subgame[p] || autodestruct[p]>=350)
			{ // Collision with ship
				int nfm=Min(fm+fmax/pm,fmax);
				char ok=0;
				if (upgrade[p][17]>0 && autodestruct[p]<350) // Dealt with by shields
				{
					float dx=ax[k=shit[p]-1]-px[p],dy=ay[k]-py[p],dd=dx*dx+dy*dy;
					upgrade[p][17]--;
					axs[k]+=1.5*dx/sqrt(dd+0.1);
					ays[k]+=1.5*dy/sqrt(dd+0.1);
					if (upgrade[p][15]==12 && cm<cmax) // DESTRUCTOR (tm) shields
					{
						cola[cm]=k; colb[cm]=-1;
						cm++;
						if (at[k]!=5) score[p]+=5;
					}
					else if (at[k]==5 && cm<cmax) // (red bullet removed)
					{
						cola[cm]=k; colb[cm]=-1;
						cm++;
						if (upgrade[p][15]==11 && bm<bmax) // Your refraction shields
						{
							bx[bm]=ax[k]; by[bm]=ay[k];
							bxs[bm]=-axs[k]*2.5; bys[bm]=-ays[k]*2.5;
							bt[bm]=1; bs[bm]=100; bo[bm]=p;
							bm++;
						}
					}
					else if (at[k]==10) // (macropixel repelled)
					{
						ax[shit[p]-1]+=15.0*dx/sqrt(dd+0.1);
						ay[shit[p]-1]+=15.0*dy/sqrt(dd+0.1);
					}
					shit[p]=0;
					goto saved_by_shields;
				}
				for new_dust(i)
				{
					float col=frnd(1),rad=pow(3,cos(col*M_PI*2+M_PI/3)+1),th=frnd(M_PI*2);
					fx[i]=px[p]; fy[i]=py[p];
					fxs[i]=rad*cos(th); fys[i]=rad*sin(th);
					fc[i]=32+col*24;
				}
				fm=nfm;
				ncgisound(px[p],n_dead,20000,5000,1);
				dead[p]=1; lives[p]--; // Ship destroyed!

				for (x=0;x<32;x++) for (y=0;y<24;y++)
					if (grid[x][y]>>4==0xC)
					{
						grid[x][y]=0;
						plotbgsq(x,y,background,bgsub,bplot);
						ok=1;
					}
				if (ok) // Reset block puzzle
				{
					in=fopen(".\\levs.dat","rb"); fseek(in,(level-1)*1024,SEEK_SET);
					for (x=0;x<32;x++)
						for (y=0;y<24;y++) if ((i=fgetc(in))>>4==0xC)
						{
							grid[x][y]=i;
							wrapbitb(x*20,y*20,blox[i]);
						}
					fclose(in);
				}
				autodestruct[p]=0;
				strcpy(gamefile,openfile); signal_savegame=*openfile;
				saved_by_shields:;
			}
			if (ahit[n] && at[n]!=5) ncgisound(ax[n],n_explode,1000,2000+rnd(1500),0); // Flame attrition
		}
// Collision processing loop
		for (n=0;n<cm;n++)
		{
			int ca=cola[n],cb=colb[n];
			if (bt[cb]==10) {ahit[ca]=bhit[cb]=0; bs[cb]=1; ca=cb=-1;}
			if (ca>=0) ahit[ca]=1+MAXP;
			if (cb>=0)
			{
				set(bhit,cb,1);
				if (bt[i=cb]==7)
					if (bs[i]/30>0 && bm+2<=bmax) // Lasers bifurcate
					{
						for (j=bm;j<=bm+1;j++)
						{
							bx[j]=bx[i]; by[j]=by[i];
							efx=0.1*(bs[i]/30)*(j-bm-0.5);
							bxs[j]=bxs[i]+efx*bys[i];
							bys[j]=bys[i]-efx*bxs[i];
							bt[j]=7; bs[j]=(bs[i]/30-1)*30+29;
							bo[j]=bo[i];
						}
						bm+=2;
					}
			}
			if (ca>=0 && cb>=0)
			{
				ahit[ca]=1+bo[cb];
				if (bt[cb]==3) {aplas[ca]=1+bo[cb]; ahit[ca]=0;}
				else if (bt[cb]==8) {axs[ca]+=bxs[cb]; ays[ca]+=bys[cb]; ahit[ca]=0;}
				else if (bt[cb]==9) {axs[ca]-=bxs[cb]; ays[ca]-=bys[cb]; ahit[ca]=0;}
				ncgisound(ax[ca],n_explode,1000,(bt[n]>=2 && bt[n]<=4)?2000+rnd(1500):11025,0); // Enemy hit
			}
		}
// Attrition loop
		for every_enemy(n) if (ahit[n])
		{
			p=ahit[n]-1;
			if (at[n]==1) // ASTEROID
			{
				int nfm=Min(fm+100,fmax);
				if (as[n]>=2 && am+1<amax) // The asteroid splits
				{
					int mass=legalsize[as[n]];
					j=k=0;
					while (!islegalsize(j) || !islegalsize(k))
					{
						j=rnd(mass-1)+1; k=mass-j;
					}
					j=islegalsize(j); k=islegalsize(k);
					ax[am]=ax[am+1]=ax[n];
					ay[am]=ay[am+1]=ay[n];
					at[am]=at[am+1]=1;
					efx=frnd(1)-0.5; efy=frnd(1)-0.5;
					axs[am]=axs[n]+efx;
					ays[am]=ays[n]+efy;
					axs[am+1]=axs[n]-efx;
					ays[am+1]=ays[n]-efy;
					ar[am]=ar[am+1]=ars[am]=ars[am+1]=0;
					as[am]=j; as[am+1]=k;
					afire[am]=afire[am+1]=afire[n];
					aplas[am]=aplas[am+1]=aplas[n];
					awhit[am]=awhit[am+1]=awhit[n];
					am+=2;
					score[p]+=5;
				}
				else score[p]+=10;
				if (fm<fmax) for new_dust(j) // Asteroid hit generates dust
				{
					float dd=0;
					fx[j]=ax[n]; fy[j]=ay[n];
					while (dd<10 || dd>25)
					{
						fxs[j]=frnd(10)-5; fys[j]=frnd(10)-5;
						dd=fxs[j]*fxs[j]+fys[j]*fys[j];
					}
					fc[j]=48+rnd(5);
				}
				fm=nfm;
			}
			else if (at[n]==2) // SPINNER
			{
				int nfm=Min(fm+1000,fmax);
				score[p]+=50;
				for (j=am;j<am+as[n] && j<amax;j++)
				{
					ax[j]=ax[n]; ay[j]=ay[n];
					ar[j]=M_PI*2*j/as[n];
					axs[j]=ays[j]=ars[j]=0;
					at[j]=3; as[j]=0;
					afire[j]=afire[n]; aplas[j]=aplas[n]; awhit[j]=awhit[n];
				}
				i=(int)(0.03*level*as[n]+frnd(1));
				for (j=0;j<i;j++)	as[am+rnd(as[n])]++; // Flyer upgrades
				for (j=am;j<am+as[n];j++) as[j]=Min(as[j],6);
				am=Min(am+as[n],amax);
				if (fm<fmax) for new_dust(j) // Spinner hit generates red dust
				{
					float dd=0;
					fx[j]=ax[n]; fy[j]=ay[n];
					while (dd<1e-6 || dd>25)
					{
						fxs[j]=frnd(10)-5; fys[j]=frnd(10)-5;
						dd=fxs[j]*fxs[j]+fys[j]*fys[j];
					}
					fc[j]=12;
				}
				fm=nfm;
			}
			else if (at[n]==3) // FLIER (aka rocket)
			{
				int nfm=Min(fm+100,fmax); // Flier hit generates dust
				score[p]+=15+10*as[n];
				if (as[n]==2 && am+3<=amax) // Actually splits into three normal ones
				{
					for (i=am;i<am+3;i++)
					{
						ax[i]=ax[n]; ay[i]=ay[n];
						axs[i]=ays[i]=ars[i]=0;
						ar[i]=ar[n]+0.5*(i-am-1);
						at[i]=3; as[i]=0;
						afire[i]=afire[n]; aplas[i]=aplas[n]; awhit[i]=awhit[n];
					}
					am+=3;
				}
				else if (as[n]==5 && am+4<=amax) // Actually splits into four splitters
				{
					for (i=am;i<am+4;i++)
					{
						ax[i]=ax[n]; ay[i]=ay[n];
						axs[i]=ays[i]=ars[i]=0;
						ar[i]=ar[n]+0.5*(i-am-1.5);
						at[i]=3; as[i]=2;
						afire[i]=afire[n]; aplas[i]=aplas[n]; awhit[i]=awhit[n];
					}
					am+=4;
				}
				if (fm<fmax) for new_dust(j)
				{
					float dd=0;
					fx[j]=ax[n]; fy[j]=ay[n];
					while (dd<1e-6 || dd>25)
					{
						fxs[j]=frnd(10)-5; fys[j]=frnd(10)-5;
						dd=fxs[j]*fxs[j]+fys[j]*fys[j];
					}
					fc[j]=41+rnd(4);
				}
				fm=nfm;
			}
			else if (at[n]==4) // SAUCER
			{
				int nfm=Min(fm+500,fmax); // Saucer hit generates dust
				score[p]+=75;
				if (fm<fmax) for new_dust(j)
				{
					float dd=0;
					fx[j]=ax[n]; fy[j]=ay[n];
					while (dd<20 || dd>25)
					{
						fxs[j]=frnd(10)-5; fys[j]=frnd(10)-5;
						dd=fxs[j]*fxs[j]+fys[j]*fys[j];
						if (j%3>=1) {fxs[j]*=0.3; fys[j]*=0.3;}
						if (j%3==2) {fxs[j]*=0.3; fys[j]*=0.3;}
					}
					fc[j]=15;
					if (j%3==1) fc[j]=13;
					if (j%3==2) fc[j]=10;
				}
				fm=nfm;
			}
			else if (at[n]==5) // RED BULLET
			{
				if (ar[n]==0)
				{
					int nfm=Min(fm+10,fmax);
					if (fm<fmax) for new_dust(i)
					{
						fx[i]=ax[n]; fy[i]=ay[n];
						fxs[i]=frnd(10)-5; fys[i]=frnd(10)-5;
						if (i%3==0) fc[i]=4;
						else if (i%3==1) fc[i]=12;
						else fc[i]=13;
					}
					fm=nfm;
				}
			}
			else if (at[n]==6) // HOMER
			{
				int nfm=Min(fm+50,fmax);
				score[p]+=5*as[n];
				if (as[n]>1 && am+3<=amax) // The triangle triangulates
				{
					for (i=0;i<3;i++)
					{
						ax[am+i]=ax[n]; ay[am+i]=ay[n];
						axs[am+i]=axs[n]+0.5*cos(M_PI*i*2/3+ar[n]);
						ays[am+i]=ays[n]+0.5*sin(M_PI*i*2/3+ar[n]);
						ar[am+i]=ar[n]; ars[am+i]=frnd(0.1)-0.05;
						at[am+i]=6; as[am+i]=as[n]-1;
						afire[am+i]=afire[n];
						aplas[am+i]=aplas[n];
						awhit[am+i]=awhit[n];
					}
					am+=3;
				}
				for new_dust(i) // Homer hit generates dust
				{
					fx[i]=ax[n]; fy[i]=ay[n];
					fxs[i]=sin(M_PI*2*i/50)*5;
					fys[i]=cos(M_PI*2*i/50)*5;
					fc[i]=32.0+(i-fm)*24.0/50.0;
				}
				fm=nfm;
				if (as[n]==1 && em+3<=emax) // Real corners!
				{
					for (i=0;i<3;i++)
					{
						float rad=0.6+frnd(0.3),th=n+i*M_PI*2/3;
						ex[em+i]=ax[n]-3; ey[em+i]=ay[n]-3;
						exs[em+i]=axs[n]+rad*cos(th); eys[em+i]=ays[n]+rad*sin(th);
						enf[em+i]=10; ef[em+i]=frnd(10); efs[em+i]=0.3+frnd(0.1);
						ea[em+i]=e_corner; er[em+i]=rnd(3)+1;
					}
					em+=3;
				}
			}
			else if (at[n]==7) // PENTAGRAM
			{
				float pf,vx,vy;
				int nfm=Min(fm+200*as[n],fmax);
				score[p]+=20*as[n];
				for (i=0;i<sq(as[n]) && am+i<amax;i++)
				{
					if (i%as[n]==0)
					{
						pf=frnd(0.1)-0.05;
						vx=frnd(1)-0.5;
						vy=frnd(1)-0.5;
					}
					axs[am+i]=axs[n]+vx;
					ays[am+i]=ays[n]+vy;
					ar[am+i]=M_PI*2/as[n]*i; ars[am+i]=pf;
					at[am+i]=8; as[am+i]=(i/as[n]+1)*3;
					ax[am+i]=ax[n]+as[am+i]*sin(ar[am+i]);
					ay[am+i]=ay[n]-as[am+i]*cos(ar[am+i]);
					afire[am+i]=afire[n];
					aplas[am+i]=aplas[n];
					awhit[am+i]=awhit[n];
				}
				am=Min(am+sq(as[n]),amax);
				if (fm<fmax) for new_dust(i) // Pentagram hit generates dust formation
				{
					float th=frnd(M_PI*2);
					fx[i]=ax[n]; fy[i]=ay[n];
					fxs[i]=sin(th)+(0.8+frnd(0.1))*sin(th*(as[n]+1));
					fys[i]=cos(th)+(0.8+frnd(0.1))*cos(th*(as[n]+1));
					if (!rnd(2)) {fxs[i]/=10; fys[i]/=10;}
					fc[i]=36+rnd(8);
				}
				fm=nfm;
			}
			else if (at[n]==8) // RESON
			{
				int nfm=Min(fm+50,fmax);
				score[p]+=5;
				if (fm<fmax) for new_dust(i)
				{
					float th=frnd(M_PI*2),rad=frnd(1)+1;
					fx[i]=ax[n]; fy[i]=ay[n];
					fxs[i]=rad*sin(th);
					fys[i]=rad*cos(th);
					fc[i]=(rnd(3))?9:(rnd(2))?11:15;
				}
				fm=nfm;
			}
			else if (at[n]==9) // TIME SHIP
			{
				int nfm=Min(fm+500,fmax);
				score[p]+=100;
				if (fm<fmax) for new_dust(i)
				{
					float th=frnd(M_PI*16),rad=exp(-th/5);
					fx[i]=ax[n]; fy[i]=ay[n];
					fxs[i]=rad*sin(th);
					fys[i]=rad*cos(th);
					fc[i]=15;
				}
				fm=nfm;
			}
			else if (at[n]==10) // MACROPIXEL
			{
				int step=4*pot[as[n]];
				score[p]+=5*sq(as[n]+1);
				if (as[n]>0 && am+4<=amax)
				{
					i=-step/4; j=i+step/2;
					ax[am]=ax[n]+i; ay[am]=ay[n]+i;
					ax[am+1]=ax[n]+i; ay[am+1]=ay[n]+j;
					ax[am+2]=ax[n]+j; ay[am+2]=ay[n]+i;
					ax[am+3]=ax[n]+j; ay[am+3]=ay[n]+j;
					for (i=am;i<am+4;i++)
					{
						axs[i]=ays[i]=ar[i]=ars[i]=0;
						as[i]=as[n]-1;
						at[i]=at[n];
						afire[i]=afire[n]; aplas[i]=aplas[n]; awhit[i]=awhit[n];
					}
					am+=4;
				}
				if (fm+4<=fmax) // Corners fly out
				{
					fx[fm]=ax[n]+.5; fy[fm]=ay[n]+.5;
					fxs[fm]=-0.001; fys[fm]=-0.001;
					fx[fm+1]=ax[n]+.5; fy[fm+1]=ay[n]+step-1.5;
					fxs[fm+1]=-0.001; fys[fm+1]=0.001;
					fx[fm+2]=ax[n]+step-1.5; fy[fm+2]=ay[n]+.5;
					fxs[fm+2]=0.001; fys[fm+2]=-0.001;
					fx[fm+3]=ax[n]+step-1.5; fy[fm+3]=ay[n]+step-1.5;
					fxs[fm+3]=0.001; fys[fm+3]=0.001;
					fc[fm]=fc[fm+1]=fc[fm+2]=fc[fm+3]=macpixcol[as[n]];
					fm+=4;
				}
			}
			else if (at[n]==11) // SELF-REPLICATING MINE
			{
				int nfm=Min(fm+200,fmax);
				score[p]+=10;
				if (fm<fmax) for new_dust(i)
				{
					float th=frnd(M_PI*2),rad=3+sin(th*10);
					fx[i]=ax[n]; fy[i]=ay[n];
					fxs[i]=rad*sin(th);
					fys[i]=rad*cos(th);
					fc[i]=7;
				}
				fm=nfm;
				if (em<emax)
				{
					ex[em]=ax[n]-10; ey[em]=ay[n]-10;
					exs[em]=axs[n]; eys[em]=ays[n];
					enf[em]=10; ef[em]=0; efs[em]=0.3;
					ea[em]=e_fireball; er[em]=1;
					em++;
				}
			}
			if (at[n]==12) // RECOALESCING ASTEROID
			{
				int nfm=Min(fm+100,fmax);
				if (as[n]>=2 && am+2<=amax) // The asteroid splits
				{
					float rc;
					int mass=legalsize[as[n]];
					j=k=0;
					while (!islegalsize(j) || !islegalsize(k))
					{
						j=rnd(mass-1)+1; k=mass-j;
					}
					j=islegalsize(j); k=islegalsize(k);
					at[am]=at[am+1]=12;
					efx=frnd(1)-0.5; efy=frnd(1)-0.5;
					axs[am]=axs[n]+efx;
					ays[am]=ays[n]+efy;
					axs[am+1]=axs[n]-efx;
					ays[am+1]=ays[n]-efy;
					ar[am]=ar[am+1]=ars[am]=ars[am+1]=0;
					as[am]=j; as[am+1]=k;
					rc=0.25*(asz[as[am]]+asz[as[am+1]]);
					rc/=sqrt(efx*efx+efy*efy);
					efx*=rc; efy*=rc;
					ax[am]=ax[n]+efx*1.1;
					ay[am]=ay[n]+efy*1.1;
					ax[am+1]=ax[n]-efx;
					ay[am+1]=ay[n]-efy;
					afire[am]=afire[am+1]=afire[n];
					aplas[am]=aplas[am+1]=aplas[n];
					awhit[am]=awhit[am+1]=awhit[n];
					am+=2;
					score[p]+=5;
				}
				else score[p]+=10;
				if (fm<fmax) for new_dust(j) // Asteroid hit generates dust
				{
					float dd=0;
					fx[j]=ax[n]; fy[j]=ay[n];
					while (dd<10 || dd>25)
					{
						fxs[j]=frnd(10)-5; fys[j]=frnd(10)-5;
						dd=fxs[j]*fxs[j]+fys[j]*fys[j];
					}
					fc[j]=32+rnd(5);
				}
				fm=nfm;
			}
			else if (at[n]==13) // PULSAR
			{
				score[p]+=250;
			}
			else if (at[n]==14) // RUBBER DUCK
			{
				int nam=Min(am+3,amax);
				int nfm=Min(fm+1000,fmax);
				score[p]+=40;
				for (i=am;i<nam;i++)
				{
					ax[i]=ax[n]; ay[i]=ay[n];
					axs[i]=axs[n]+frnd(2)-1; ays[i]=ays[n]+frnd(2)-1;
					ar[i]=ars[i]=0; at[i]=15;
					as[i]=0;
				}
				am=nam;
				for new_dust(i)
				{
					float rad=frnd(5)+1e-6,th=frnd(M_PI*2);
					fx[i]=ax[n]; fy[i]=ay[n];
					fxs[i]=rad*cos(th); fys[i]=rad*sin(th);
					fc[i]=14;
				}
				fm=nfm;
				if (em<emax)
				{
					int nem=Min(emax,em+500);
					for (i=em;i<nem;i++)
					{
						float rad=frnd(8),th=frnd(M_PI*2);
						ex[i]=ax[n]-10; ey[i]=ay[n]-10;
						exs[i]=axs[n]+rad*cos(th); eys[i]=ays[n]+rad*sin(th);
						enf[i]=10; ef[i]=frnd(10); efs[i]=0.2+frnd(0.2);
						ea[i]=e_fireball; er[i]=1;
					}
					em=nem;
				}
			}
			else if (at[n]==15) // DUCK EGG
			{
				score[p]+=20;
			}
			else if (at[n]==16) // DEFENDER (tm) BOMBER
			{
				int nfm=Min(fm+250,fmax); // Bomber hit generates dust
				score[p]+=80;
				if (fm<fmax) for new_dust(i)
				{
					float dd=0;
					fx[i]=ax[n]; fy[i]=ay[n];
					while (dd<1e-6 || dd>25)
					{
						fxs[i]=frnd(10)-5; fys[i]=frnd(10)-5;
						dd=fxs[i]*fxs[i]+fys[i]*fys[i];
					}
					fc[i]=13;
				}
				fm=nfm;
			}
			else if (at[n]==17) // DEFENDER (tm) BAITER
			{
				int nfm=Min(fm+250,fmax); // Baiter hit generates dust
				score[p]+=150;
				if (fm<fmax) for new_dust(i)
				{
					float dd=0;
					fx[i]=ax[n]; fy[i]=ay[n];
					while (dd<1e-6 || dd>25)
					{
						fxs[i]=frnd(10)-5; fys[i]=frnd(10)-5;
						dd=fxs[i]*fxs[i]+fys[i]*fys[i];
					}
					fys[i]/=4;
					fc[i]=10+4*rnd(2);
				}
				fm=nfm;
			}
			else if (at[n]==18) // HOMING SPINNER
			{
				int nfm=Min(fm+500,fmax);
				score[p]+=75;
				for new_dust(i) // Homing Spinner hit generates dust - centre circle
				{
					float r=frnd(1),g=frnd(1),b=frnd(1),rad=sqrt(frnd(200)),th=frnd(M_TWOPI);
					fx[i]=ax[n]+rad*cos(th); fy[i]=ay[n]+rad*sin(th);
					fxs[i]=r*2-(g+b); fys[i]=(g-b)*sqrt(3);
					fc[i]=floor(r*5)+floor(g*5)*5+floor(b*5)*25+128;
				}
				fm=nfm;
			}
			else if (at[n]==19) // FLAMETHROWER BOSS
			{
				int nfm=Min(fm+600,fmax);
				score[p]+=750;
				for new_dust(i) // Boss hit generates dust
				{
					float spd=frnd(4);
					fx[i]=ax[n]; fy[i]=ay[n];
					fxs[i]=sin(M_PI*i/150)*(0.5+spd/4);
					fys[i]=cos(M_PI*i/150)*(0.5+spd/4);
					fc[i]=128+(int)spd;
				}
				fm=nfm;
			}
			else if (at[n]==20) // RED TRIANGLE
			{
				int nfm=Min(fm+60,fmax);
				for new_dust(i) // triangle hit generates dust
				{
					float spd=frnd(4);
					fx[i]=ax[n]; fy[i]=ay[n];
					fxs[i]=sin(M_PI*i/15)*(0.3+spd/6);
					fys[i]=cos(M_PI*i/15)*(0.3+spd/6);
					fc[i]=128+(int)spd;
				}
				fm=nfm;
			}
			else if (at[n]==22) // N BALL
			{
				score[p]+=5;
				int nfm=Min(fm+60,fmax);
				for new_dust(i)
				{
					float spd=frnd(4);
					fx[i]=ax[n]; fy[i]=ay[n];
					fxs[i]=sin(M_PI*i/15)*(0.3+spd/6);
					fys[i]=cos(M_PI*i/15)*(0.3+spd/6);
					fc[i]=40+as[n]*2;
				}
				fm=nfm;
				float new_xs = frnd(1)-.5, new_ys = frnd(1)-.5;
				if (as[n]>0) for (j=0;j<2;j++) if (am<amax)
				{
					ax[am]=ax[n]; ay[am]=ay[n];
/*cj*/					if (j==0) {axs[am] = axs[n] + new_xs; ays[am] = ays[n] + new_ys;}  // New code to make balls
					else      {axs[am] = axs[n] - new_xs; ays[am] = ays[n] - new_ys;}  //    fly apart on hit
/*/cj*/				ar[am]=ars[am]=0;
					as[am]=as[n]-1;
					afire[am]=afire[n]; aplas[am]=aplas[n]; awhit[am]=awhit[n]; ahit[am]=0;
					at[am]=22;
					am++;
				}
			}
			if (!rnd(10+5*im) && im<imax && at[n]!=5) // Bonus creation
			{
				float dd=10;
				ix[im]=ax[n]; iy[im]=ay[n];
				while (dd>8)
				{
					ixs[im]=frnd(2)-1; iys[im]=frnd(2)-1;
					dd=sq(ixs[im])+sq(iys[im]);
				}
				iage[im]=0;
				if (frnd(1)<0.04+0.001*level)
				{
					i=rnd(4);
					if (i==0) it[im]=1000+'C';
					else if (i==1) it[im]=1000+'G';
					else if (i==2) it[im]=1000+'I';
					else if (i==3) it[im]=1000+'N';
				}
				else if (!rnd(2)) it[im]=1000+(int)floor(log(frnd(0.998)+0.002)/log(0.5)); // Money creation
				else it[im]=rnd(itypes)+1;
				im++;
			}
			am--; // The only place where enemy objects are removed
			ax[n]=ax[am]; ay[n]=ay[am];
			axs[n]=axs[am]; ays[n]=ays[am];
			ar[n]=ar[am]; ars[n]=ars[am];
			as[n]=as[am]; at[n]=at[am];
			afire[n]=afire[am]; aplas[n]=aplas[am]; awhit[n]=awhit[am];
			ahit[n]=ahit[am];
			ahit[am]=0;
		}
// Flame loop
		for flame_particles(n)
		{
			ffx[n]+=rnd(5)-2; ffy[n]+=rnd(5)-2;
			if (ffx[n]<0 || ffy[n]<0 || ffx[n]>xres || ffy[n]>yres || !rnd(10))
			{
				ffm--;
				ffx[n]=ffx[ffm]; ffy[n]=ffy[ffm];
				ffc[n]=ffc[ffm];
			}
			plussign(ffx[n],ffy[n],ffc[n]);
		}
// Bitmapped explosion loop
		for (n=0;n<em;n++)
		{
			wrapbit(ex[n],ey[n],ea[n][(int)floor(ef[n])]);
			ex[n]=fmod(ex[n]+exs[n]*dt+xres,xres);
			ey[n]=fmod(ey[n]+eys[n]*dt+yres,yres);
			ef[n]+=efs[n]*dt;
			if (ef[n]>=enf[n])
			{
				er[n]--; ef[n]=0;
				if (er[n]==0)
				{
					em--;
					ex[n]=ex[em]; ey[n]=ey[em];
					exs[n]=exs[em]; eys[n]=eys[em];
					ef[n]=ef[em]; efs[n]=efs[em];
					er[n]=er[em]; enf[n]=enf[em];
					ea[n]=ea[em];
					n--;
				}
			}
		}
// Items loop
		for every_item(n)
		{
			float vv,irad=6;
			if ((n-1000)/3==0) irad=2.5;
			else if ((n-1003)/3==0) irad=5;
			else if ((n-1006)/3==0) irad=7.5;
			for every_player(p)
			{
				ixs[n]+=2.5e-4*xfrom(px[p]-ix[n])*upgrade[p][5]*dt;
				iys[n]+=2.5e-4*yfrom(py[p]-iy[n])*upgrade[p][5]*dt;
			}
			vv=sq(ixs[n])+sq(iys[n]);
			if (vv>4) {ixs[n]/=sqrt(vv)/2; iys[n]/=sqrt(vv)/2;}
			ix[n]+=ixs[n]*dt; iy[n]+=iys[n]*dt;
			if (iage[n]<32767) iage[n]++;
			if (ix[n]<0) ix[n]+=xres;
			if (ix[n]>=xres) ix[n]-=xres;
			if (iy[n]<0) iy[n]+=yres;
			if (iy[n]>=yres) iy[n]-=yres;
			// Item interaction with grid
			for (efx=0;efx<M_PI*2;efx+=4.0/irad)
			{
				float cx=ix[n]+irad*cos(efx),cy=iy[n]+irad*sin(efx);
				int gx,gy;
				if (cx<0) cx+=xres;
				else if (cx>=xres) cx-=xres;
				if (cy<0) cy+=yres;
				else if (cy>=yres) cy-=yres;
				gx=(int)cx/20; gy=(int)cy/20;
				if (iattrib[grid[gx][gy]]) // Item bounces (not straight through)
				{
					int q;
					cx-=gx*20+10; cy-=gy*20+10;
					q=(cx<cy)+(cx<-cy)*2;
					if (q==0 && ixs[n]<0) {ixs[n]*=-1; ix[n]++;}
					else if (q==3 && ixs[n]>0) {ixs[n]*=-1; ix[n]--;}
					else if (q==1 && iys[n]<0) {iys[n]*=-1; iy[n]++;}
					else if (q==2 && iys[n]>0) {iys[n]*=-1; iy[n]--;}
				}
			}
			for every_player(p) if (ddfrom(px[p],py[p],ix[n],iy[n])<sq(vr[p]+irad) &&
				!subgame[p] && !dead[p] && (upgrade[p][9]==0 || it[n]>=1000)) // Bonus collected
			{
				int nffm,nfm;
				if (it[n]==1) {upgrade[p][0]++; meterdisp[p]=150;}
				else if (it[n]==2) {upgrade[p][1]+=3; meterdisp[p]=150;}
				else if (it[n]==3) {upgrade[p][2]=400; meterdisp[p]=150;}
				else if (it[n]==4) upgrade[p][3]++;
				else if (it[n]==5) upgrade[p][5]++;
				else if (it[n]==6) score[p]+=150;
				else if (it[n]==7)
				{
					upgrade[p][12]+=0.4*upgrade[p][12]+1;
					if (upgrade[p][12]>20000) upgrade[p][12]=20000;
				}
				else if (it[n]==8)
				{
					int nfm=Min(fm+1000,fmax); // Extra life special effect
					if (lives[p]<255) lives[p]++;
					if (fm<fmax) for new_dust(i)
					{
						float rad=0,th=0;
						fx[i]=px[p]; fy[i]=py[p];
						while (!rad || (int)(th/(float)(M_PI*2)*(int)(rad*4))%2)
						{
							rad=frnd(4.5)+0.5; th=frnd(M_PI*2);
						}
						fxs[i]=rad*sin(th); fys[i]=rad*cos(th);
						fc[i]=rad*5+30;
					}
					fm=nfm;
				}
				else if (it[n]==9) upgrade[p][4]++;
				else if (it[n]==10) upgrade[p][6]++;
				else if (it[n]==11) {upgrade[p][7]+=300; meterdisp[p]=150; if (upgrade[p][7]>2500) upgrade[p][7]=2500;}
				else if (it[n]==12) upgrade[p][8]++;
				else if (it[n]==13) upgrade[p][9]=1000;
				else if (it[n]==14)
				{
					upgrade[p][14]+=0.4*upgrade[p][14]+1;
					if (upgrade[p][14]>10000000) upgrade[p][14]=10000000;
				}
				else if (it[n]>=1000 && it[n]<=1008) // Money collected
				{
					if (it[n]==1000) money[p]+=1;
					else if (it[n]==1001) money[p]+=2;
					else if (it[n]==1002) money[p]+=5;
					else if (it[n]==1003) money[p]+=10;
					else if (it[n]==1004) money[p]+=20;
					else if (it[n]==1005) money[p]+=50;
					else if (it[n]==1006) money[p]+=100;
					else if (it[n]==1007) money[p]+=200;
					else if (it[n]==1008) money[p]+=500;
					if (em<emax) // Sparkle
					{
						ex[em]=ix[n]-5; ey[em]=iy[n]-5;
						exs[em]=ixs[n]; eys[em]=iys[n];
						enf[em]=2; ef[em]=0; efs[em]=0.5;
						ea[em]=e_sparkle; er[em]=12;
						em++;
					}
					moneydisp[p]=100;
				}
				else if (it[n]>=1000+'A' && it[n]<=1000+'Z')
				{
					if (wm<wmax)
					{
						wx[wm]=ix[n]; wy[wm]=iy[n];
						ws[wm]=1000; wt[wm]=it[n]-1000;
						wm++;
					}
				}
				if (it[n]<1000)
				{
					nffm=Min(ffm+100,ffmax);
					if (ffm<ffmax) for new_flame(i) // Item flare
					{
						ffx[i]=px[p]; ffy[i]=py[p];
						ffc[i]=10;
						if (!rnd(2)) ffc[i]=11;
						if (!rnd(3)) ffc[i]=15;
					}
					ffm=nffm;
					nfm=Min(fm+100,fmax);
					if (fm<fmax) for new_dust(i) // Item dust
					{
						fx[i]=px[p]; fy[i]=py[p];
						j=i%100;
						if (j<25) {fxs[i]=1; fys[i]=0.08*j-1;}
						else if (j<50) {fxs[i]=-1; fys[i]=0.08*(j-25)-1;}
						else if (j<75) {fxs[i]=0.08*(j-50)-1; fys[i]=1;}
						else {fxs[i]=0.08*(j-75)-1; fys[i]=-1;}
						fc[i]=5;
					}
					fm=nfm;
				}
				else; // do something cool
				im--;
				ix[n]=ix[im]; iy[n]=iy[im];
				ixs[n]=ixs[im]; iys[n]=iys[im];
				it[n]=it[im]; iage[n]=iage[im];
			}
			if (iage[n]>5000 && it[n]<1000)
			{
				im--;
				ix[n]=ix[im]; iy[n]=iy[im];
				ixs[n]=ixs[im]; iys[n]=iys[im];
				it[n]=it[im]; iage[n]=iage[im];
			}
			if (rnd(1000)<5000-iage[n] && it[n]<1000)
				wrapbit(ix[n]-5,iy[n]-5,items[it[n]-1]);
			else if (it[n]>=1000 && it[n]<=1002)
				wrapbit(ix[n]-2,iy[n]-2,coin[it[n]-1000][(int)(0.01*GetTickCount()+n)%4]);
			else if (it[n]>=1003 && it[n]<=1005)
				wrapbit(ix[n]-5,iy[n]-5,coin[it[n]-1000][(int)(0.01*GetTickCount()+n)%4]);
			else if (it[n]>=1006 && it[n]<=1008)
				wrapbit(ix[n]-7,iy[n]-7,coin[it[n]-1000][(int)(0.01*GetTickCount()+n)%6]);
			else if (it[n]>=1000+'A' && it[n]<=1000+'Z')
			{
				str[0]=it[n]-1000; str[1]=0;
				bwrite(ix[n]-4,iy[n]-4,xres,str,(int)(framesdone)%2*15);
			}
		}
// Bullets loop
		for every_bullet(n)
		{
			int ox=(int)bx[n]/20,oy=(int)by[n]/20,gx,gy,divisor;
			for (divisor=1,j=-2;divisor<=1000000000 && j==-2;divisor<<=1)
			{
				bx[n]+=bxs[n]*dt/divisor; by[n]+=bys[n]*dt/divisor;
				if (bx[n]<0) bx[n]+=xres;
				if (bx[n]>xres) bx[n]-=xres;
				if (by[n]<0) by[n]+=yres;
				if (by[n]>yres) by[n]-=yres;
				gx=(int)bx[n]/20;
				gy=(int)by[n]/20;
				i=sq((gx-ox+48)%32-16)+sq((gy-oy+36)%24-12);
				if (i==0) j=-1; // Just leave
				else if (i>1) j=-2; // Loop once more
				else j=battrib[grid[gx][gy]];
				if (bs[n]<=0 || bit(bhit,n) || j==2) // Remove bullet
				{
					bm--;
					bx[n]=bx[bm]; by[n]=by[bm];
					bxs[n]=bxs[bm]; bys[n]=bys[bm];
					bs[n]=bs[bm]; bt[n]=bt[bm]; bo[n]=bo[bm];
					set(bhit,n,bit(bhit,bm));
					j=3;
				}
				if (grid[gx][gy]>>4==0xC && bt[n]>=8 && bt[n]<=9) j=0;
				if (j==1) // Bullet bounces
				{
					if ((gx-ox+32)%32==31 && bxs[n]<0) {bxs[n]*=-1; bx[n]=floor(bx[n]/20+1)*20;}
					else if ((gx-ox+32)%32==1 && bxs[n]>0) {bxs[n]*=-1; bx[n]=floor(bx[n]/20)*20;}
					else if ((gy-oy+24)%24==23 && bys[n]<0) {bys[n]*=-1; by[n]=floor(by[n]/20+1)*20;}
					else if ((gy-oy+24)%24==1 && bys[n]>0) {bys[n]*=-1; by[n]=floor(by[n]/20)*20;}
				}
				if (j==-2) {bx[n]-=bxs[n]*dt/divisor; by[n]-=bys[n]*dt/divisor;}
			}
			if (j==3) n--;
			else if (bt[n]==1) // Normal bullet
			{
				wrapbit(bx[n]-1,by[n]-1,bullet[bo[n]]);
				bs[n]-=idt;
			}
			else if (bt[n]==2) // Flamethrower
			{
				int nffm=Min(ffm+10,ffmax);
				for new_flame(i)
				{
					ffx[i]=bx[n]; ffy[i]=by[n];
					ffc[i]=firecolour();
				}
				ffm=nffm;
				for every_enemy(i)
					if (ddfrom(bx[n],by[n],ax[i],ay[i])<1000) afire[i]=1+bo[n];
				bs[n]-=(float)(1+rnd(2))*dt;
			}
			else if (bt[n]==3) // Plasma thrower
			{
				bxs[n]*=0.98; bys[n]*=0.98;
				for (i=0;i<20;i++)
				{
					float dx=10,dy=10;
					while (dx*dx+dy*dy>sq(4.5))
					{
						dx=rnd(9)-4; dy=rnd(9)-4;
					}
					plussign(bx[n]+dx,by[n]+dy,plasmacolour());
				}
				for every_enemy(i)
					if (ddfrom(bx[n],by[n],ax[i],ay[i])<300) aplas[i]=1+bo[n];
				if (sq(bxs[n])+sq(bys[n])<1) bs[n]-=2.0*dt;
			}
			else if (bt[n]==4) // White fire
			{
				int nffm=Min(ffm+10,ffmax),nfm=Min(fm+4,fmax);
				for new_flame(i)
				{
					ffx[i]=bx[n]; ffy[i]=by[n];
					ffc[i]=15;
				}
				ffm=nffm;
				for new_dust(i)
				{
					float rad=1+frnd(1),th=frnd(M_PI*2);
					fx[i]=bx[n]; fy[i]=by[n];
					fxs[i]=rad*sin(th); fys[i]=rad*cos(th);
					fc[i]=24+rnd(8);
				}
				fm=nfm;
				for every_enemy(i)
					if (ddfrom(bx[n],by[n],ax[i],ay[i])<1000) awhit[i]=1+bo[n];
				bs[n]-=2.0*dt;
			}
			else if (bt[n]==5) // Machine-gun bullets
			{
				wrapbit(bx[n]-1,by[n]-1,mbullet);
				bs[n]-=idt;
			}
			else if (bt[n]==6) bs[n]-=idt; // Null bullet
			else if (bt[n]==7) // Laser
			{
				unsigned char lightc[8]={12,142,14,10,11,9,13,15},
					darkc[8]={4,135,6,2,3,1,5,7};
				line(bx[n]-bxs[n]/2+1,by[n]-bys[n]/2,bx[n]+bxs[n]/2+1,by[n]+bys[n]/2,darkc[bs[n]/30]);
				line(bx[n]-bxs[n]/2-1,by[n]-bys[n]/2,bx[n]+bxs[n]/2-1,by[n]+bys[n]/2,darkc[bs[n]/30]);
				line(bx[n]-bxs[n]/2,by[n]-bys[n]/2+1,bx[n]+bxs[n]/2,by[n]+bys[n]/2+1,darkc[bs[n]/30]);
				line(bx[n]-bxs[n]/2,by[n]-bys[n]/2-1,bx[n]+bxs[n]/2,by[n]+bys[n]/2-1,darkc[bs[n]/30]);
				line(bx[n]-bxs[n]/2,by[n]-bys[n]/2,bx[n]+bxs[n]/2,by[n]+bys[n]/2,lightc[bs[n]/30]);
				if (bs[n]%30) {if (frnd(8)>(int)bs[n]/30) bs[n]--;}
					else bs[n]=0;
			}
			else if (bt[n]==8 || bt[n]==9) // Prodder of some sort
			{
				int qx=(int)bx[n]/20,qy=(int)by[n]/20;
				float v=sqrt(sq(bxs[n])+sq(bys[n]))+0.001,
							vx=bxs[n]*3/v,vy=bys[n]*3/v,ux=bys[n]*3/v,uy=-bxs[n]*3/v;
				if (bt[n]==9) {ux*=-1; uy*=-1; vx*=-1; vy*=-1;}
				line(bx[n]+vx,by[n]+vy,bx[n]+ux,by[n]+uy,11+2*(bt[n]==9));
				line(bx[n]+vx,by[n]+vy,bx[n]-ux,by[n]-uy,11+2*(bt[n]==9));
				line(bx[n],by[n],bx[n]-vx+ux,by[n]-vy+uy,3+2*(bt[n]==9));
				line(bx[n],by[n],bx[n]-vx-ux,by[n]-vy-uy,3+2*(bt[n]==9));
				for every_player(p) if (ddfrom(bx[n],by[n],px[p],py[p])<sq(vr[p]) && bo[n]!=p)
				{
					if (bt[n]==8) {xs[p]+=bxs[n]; ys[p]+=bys[n];}
						else {xs[p]-=bxs[n]; ys[p]-=bys[n];}
					bm--;
					bx[n]=bx[bm]; by[n]=by[bm];
					bxs[n]=bxs[bm]; bys[n]=bys[bm];
					bs[n]=bs[bm]; bt[n]=bt[bm]; bo[n]=bo[bm];
					set(bhit,n,bit(bhit,bm));
					n--;
				}
			}
			else if (bt[n]==10) // Flame mine
			{
				wrapbit(bx[n]-8,by[n]-8,flamemine);
				if (bs[n]>15) {if ((int)sqrt(bs[n]-15)%2) rect(bx[n]-1,by[n]-1,2,2,0);}
				else if (bs[n]%2) rect(bx[n]-1,by[n]-1,2,2,15);
				bs[n]--;
				if (bs[n]<=0) // Flame mine explodes
				{
					int nfm=Min(fm+600,fmax),nbm=Min(bm+40,bmax);
					for new_dust(i)
					{
						float spd=frnd(4);
						fx[i]=bx[n]; fy[i]=by[n];
						fxs[i]=sin(M_PI*i/150)*(0.3+spd/6);
						fys[i]=cos(M_PI*i/150)*(0.3+spd/6);
						fc[i]=40+(int)spd;
					}
					fm=nfm;
					for (i=bm;i<nbm;i++) // equivalent to for new_bullet(i) but that doesn't exist
					{
						float th=frnd(M_TWOPI),rad=frnd(30);
						bx[i]=bx[n]+rad*sin(th);
						by[i]=by[n]+rad*cos(th);
						bxs[i]=sin(th); bys[i]=cos(th);
						bt[i]=2; bs[i]=300+rnd(50);
						bo[i]=bo[n];
					}
					bm=nbm;
					ncgisound(bx[n],n_fmine,25000,5000,1);
					ncgisound(bx[n],n_fmine,25000,15000,1);
				}
			}
		}
// Wormhole loop
		for (n=0;n<wm;n++)
		{
			char wormhole_collapse=0;
			circlefill(wx[n],wy[n],0.01*ws[n],0);
			ws[n]-=idt;
			if (ws[n]<0) wormhole_collapse=1;
			for every_player(p) if (ddfrom(px[p],py[p],wx[n],wy[n])<sq(vr[p]+0.01*ws[n]) &&
				!subgame[p] && !dead[p])
			{
				int q; char ok=1;
				subgame[p]=wt[n];
				for every_player(q) ok*=(subgame[q]==wt[n] || dead[q]);
				if (ok) wormhole_collapse=1;
			}
			if (wormhole_collapse)
			{
				i=strlen(gamestack);
				gamestack[i]=wt[n]; gamestack[i+1]=0;
				wm--;
				wx[n]=wx[wm]; wy[n]=wy[wm];
				ws[n]=ws[wm]; wt[n]=wt[wm];
			}
			for dust_particles(i) // Wormhole attracts dust
			{
				float dx=xfrom(wx[n]-fx[i]),dy=yfrom(wy[n]-fy[i]),rad=dx*dx+dy*dy+1;
				if (rad<sq(200)) {fxs[i]*=0.965; fys[i]*=0.965;}
				fxs[i]+=10.0*dx/rad;
				fys[i]+=10.0*dy/rad;
			}
			{ // New wormhole dust
				int nfm=Min(fm+2,fmax);
				for new_dust(i)
				{
					do {fx[i]=frnd(xres); fy[i]=rnd(yres);}
					while (ddfrom(fx[i],fy[i],wx[n],wy[n])>sq(200));
					fxs[i]=0.01*(fy[i]-wy[n]);
					fys[i]=0.01*(wx[n]-fx[i]);
					fc[i]=248+rnd(5);
				}
				fm=nfm;
			}
		}
		// Advanced pacman algorithms
		for (x=0;x<32;x++) for (y=0;y<24;y++)
		{
			char dots[0x100];
			int xi=(x+1)%32,xd=(x+31)%32,yi=(y+1)%24,yd=(y+23)%24,
				i1=grid[x][yd],i2=grid[xi][y],i3=grid[x][yi],i4=grid[xd][y],si=0,sd;
			if (i1>>4==0xD) si+=sattrib[i1]?10:1;
			if (i2>>4==0xD) si+=sattrib[i2]?10:1;
			if (i3>>4==0xD) si+=sattrib[i3]?10:1;
			if (i4>>4==0xD) si+=sattrib[i4]?10:1;
			sd=si/10+si%10;
			for (n=0;n<0x100;n++) dots[n]=(n>=0xD0 && n<=0xD6);
			i=grid[x][y];
			if (i>>4==0xD)
			{
				if (!sattrib[grid[xd][yd]] && !sattrib[grid[xd][y]] && !sattrib[grid[xd][yi]] &&
					!sattrib[grid[x][yd]] && dots[i] && !sattrib[grid[x][yi]] &&
					!sattrib[grid[xi][yd]] && !sattrib[grid[xi][y]] && !sattrib[grid[xi][yi]]) heat[x][y]=1;
				if (grid[xd][y]>>4==0xC || grid[xi][y]>>4==0xC ||
					grid[x][yd]>>4==0xC || grid[x][yi]>>4==0xC) heat[x][y]=(grid[x][y]==0xDF);
				if (grid[x][y]==0xDE) heat[x][y]=0;
				if (heat[x][y]) // Vapourisation
				{
					if (sattrib[grid[x][y]]) for (n=0;n<10;n++) if (em<emax) // Sparkles
					{
						ex[em]=x*20+rnd(20); ey[em]=y*20+rnd(20);
						exs[em]=frnd(3)-1.5; eys[em]=frnd(3)-1.5;
						enf[em]=2; ef[em]=0; efs[em]=0.5;
						ea[em]=e_sparkle; er[em]=rnd(5)+4;
						em++;
					}
					for (n=0;n<em;n++) if (ea[n]==e_square) // Remove square flash (if any)
						if ((int)ex[n]/20==x && (int)ey[n]/20==y)
						{
							em--;
							ex[n]=ex[em]; ey[n]=ey[em];
							exs[n]=exs[em]; eys[n]=eys[em];
							ef[n]=ef[em]; efs[n]=efs[em]; enf[n]=enf[em];
							er[n]=er[em]; ea[n]=ea[em];
							n--;
						}
					grid[x][y]=0; heat[x][y]=500;
					plotbgsq(x,y,background,bgsub,bplot);
				}
				if (!sattrib[grid[x][y]] && (grid[xi][y]==0xDE || grid[xd][y]==0xDE ||
					grid[x][yi]==0xDE || grid[x][yd]==0xDE) && !rnd(1000) && am<amax)
				{
					j=0;
					for every_enemy(n) j+=(at[n]==21);
					ax[am]=x*20+10; ay[am]=y*20+10;
					axs[am]=ays[am]=ar[am]=ars[am]=as[am]=0;
					at[am]=21; // New ghost
					am++;
					if (rnd(12)<j) am--;
				}
			}
			else if (i==0)
			{
				if (heat[x][y]>0) heat[x][y]--;
				for every_player(p) if (x==(int)px[p]/20 && y==(int)py[p]/20) si=0;
				if (si>0 && !rnd(4000/sd/sd/sd) && !heat[x][y]) // Grid growth
				{
					int xx,yy;
					char blockage=0,going,ab[32][24],aa[32][24];
					for (xx=0;xx<32;xx++) for (yy=0;yy<24;yy++) ab[xx][yy]=2*sattrib[grid[xx][yy]];
					ab[(int)px[0]/20][(int)py[0]/20]=1; going=1;
					while (going) // Access-check before
					{
						going=0;
						for (xx=0;xx<32;xx++) for (yy=0;yy<24;yy++) if (ab[xx][yy]==0)
						{
							int xxi=(xx+1)%32,xxd=(xx+31)%32,yyi=(yy+1)%24,yyd=(yy+23)%24;
							if (ab[xxi][yy]==1 || ab[xxd][yy]==1 || ab[xx][yyi]==1 || ab[xx][yyd]==1)
								{ab[xx][yy]=1; going=1;}
						}
					}
					grid[x][y]=0xDF;
					for (xx=0;xx<32;xx++) for (yy=0;yy<24;yy++) aa[xx][yy]=2*sattrib[grid[xx][yy]];
					aa[(int)px[0]/20][(int)py[0]/20]=1; going=1;
					while (going) // Access-check after
					{
						going=0;
						for (xx=0;xx<32;xx++) for (yy=0;yy<24;yy++) if (aa[xx][yy]==0)
						{
							int xxi=(xx+1)%32,xxd=(xx+31)%32,yyi=(yy+1)%24,yyd=(yy+23)%24;
							if (aa[xxi][yy]==1 || aa[xxd][yy]==1 || aa[xx][yyi]==1 || aa[xx][yyd]==1)
								{aa[xx][yy]=1; going=1;}
						}
					}
					for (xx=0;xx<32;xx++) for (yy=0;yy<24;yy++)
						if (ab[xx][yy] && !aa[xx][yy]) blockage=1;
					if (sattrib[grid[xd][y]] && sattrib[grid[xd][yd]] && sattrib[grid[x][yd]] ||
						sattrib[grid[xd][y]] && sattrib[grid[xd][yi]] && sattrib[grid[x][yi]] ||
						sattrib[grid[xi][y]] && sattrib[grid[xi][yd]] && sattrib[grid[x][yd]] ||
						sattrib[grid[xi][y]] && sattrib[grid[xi][yi]] && sattrib[grid[x][yi]]) blockage=1;
					if (!sattrib[grid[xd][y]] && sattrib[grid[xd][yd]] && !sattrib[grid[x][yd]] ||
						!sattrib[grid[xd][y]] && sattrib[grid[xd][yi]] && !sattrib[grid[x][yi]] ||
						!sattrib[grid[xi][y]] && sattrib[grid[xi][yd]] && !sattrib[grid[x][yd]] ||
						!sattrib[grid[xi][y]] && sattrib[grid[xi][yi]] && !sattrib[grid[x][yi]]) blockage=1;
					if (grid[xd][y]>>4==0xC || grid[xi][y]>>4==0xC ||
						grid[x][yd]>>4==0xC || grid[x][yi]>>4==0xC) blockage=1;
					if (blockage)
					{
						grid[x][y]=0xD1;
						if (!rnd(20)) grid[x][y]=0xD2+rnd(5);
						if (!rnd(40)) grid[x][y]=0xD7;
					}
					if (dots[grid[x][y]] && si<10 && !sattrib[grid[xd][yd]] && !sattrib[grid[xd][yi]] &&
						!sattrib[grid[xi][yd]] && !sattrib[grid[xi][yi]]) grid[x][y]=0;
					if (grid[x][y]!=0)
					{
						wrapbitb(x*20,y*20,blox[grid[x][y]]);
						if (em<emax)
						{
							ex[em]=x*20; ey[em]=y*20;
							exs[em]=eys[em]=0;
							enf[em]=30; ef[em]=0; efs[em]=0.5;
							ea[em]=e_square; er[em]=1;
							em++;
						}
					}
				}
			}
		}
		// Gameplay logic
		for every_player(p) if (!subgame[p])
		{
			vr[p]=5;
			if (!dead[p])
			{
				if (pacman[p]) // Plot the pacman!
				{
					unsigned char col[MAXP]={10,231,11,14};
					if (upgrade[p][9]>0 && (int)framesdone%4==0) col[p]=0;
					else if (upgrade[p][9]>80 && (int)framesdone%4==2) col[p]=15;
					if (upgrade[p][25] && (int)framesdone%4<2) col[p]=12;
					for (x=-8;x<=8;x++) for (y=-8;y<=8;y++)
						if (x*x+y*y<=70)
						{
							char on=0;
							efx=sq(sin(0.01*framestart));
							if (fabs(xs[p])>=Max(ys[p],0.01))
							{
								if (xs[p]>0) on|=(fabs(y)>=x*efx);
								else on|=(fabs(y)>=-x*efx);
							}
							else if (fabs(ys[p])>=Max(xs[p],0.01))
							{
								if (ys[p]>0) on|=(fabs(x)>=y*efx);
								else on|=(fabs(x)>=-y*efx);
							}
							else on=1;
							if (on) putpix((int)(x+px[p]+xres)%xres,(int)(y+py[p]+yres)%yres,col[p]);
						}
				}
				else // Plot the ship!
				{
					if (upgrade[p][9]>0 && (int)framesdone%4==0)
						wrapbitc(px[p]-7.5,py[p]-7.5,ship[p][(int)(rot[p]*10)],0);
					else if (upgrade[p][9]>80 && (int)framesdone%4==2)
						wrapbitc(px[p]-7.5,py[p]-7.5,ship[p][(int)(rot[p]*10)],15);
					else wrapbit(px[p]-7.5,py[p]-7.5,ship[p][(int)(rot[p]*10)]);
					if (meterdisp[p]>0) // Show available weapons
					{
						unsigned char col[11]={10,42,9,15,24,11,4,13,3,5,66};
						y=0;
						for (n=0;n<11;n++)
						{
							if (n==0 || n>=1 && n<=3 && upgrade[p][n-1] || n==4 && upgrade[p][7] ||
								n>=5 && n<=10 && upgrade[p][n+14])
							{
								rect(px[p]+10,py[p]+y,1+(btype[p]==n+1)*2,1,col[n]);
								y+=2;
							}
						}
					}
				}
				for (j=0;j<=Min(upgrade[p][17]/50,20);j++) // Plot the shields!
				{
					k=Min(upgrade[p][17]-j*50,50);
					for (i=0;i<k;i++)
					{
						float th=M_PI*2*i/k+framestart*0.01;
						x=px[p]+cos(th)*(10+j); y=py[p]+sin(th)*(10+j);
						x=(x+xres)%xres; y=(y+yres)%yres;
						putpix(x,y,upgrade[p][15]);
					}
				}
				if (upgrade[p][17]>=1000) // Excessive shields
				{
					sprintf(str,"#%d",upgrade[p][17]);
					write(px[p]+25,py[p]+25,xres,str,upgrade[p][15],0);
				}
				if (moneydisp[p]>0) // Money display
				{
					sprintf(str,"%d.%d%d",money[p]/100,money[p]/10%10,money[p]%10);
					bwrite(px[p]-10,py[p]-20,xres,str,128+6*(moneydisp[p]/21));
					moneydisp[p]=noneg(moneydisp[p]-idt);
				}
				if (autodestruct[p]>100) // Autodestruct sequence
				{
					sprintf(str,"%d",1+(350-autodestruct[p])/50);
					bwrite(px[p]-4,py[p]-4,xres,str,12*(autodestruct[p]/10%2));
				}
			}
			else if (dead[p]>0) dead[p]+=idt;
			upgrade[p][2]=noneg(upgrade[p][2]-idt);
			upgrade[p][9]=noneg(upgrade[p][9]-idt);
			upgrade[p][25]=noneg(upgrade[p][25]-idt);
			if (frnd(500)<upgrade[p][18] && upgrade[p][17]<upgrade[p][16]) upgrade[p][17]++;
			if (shopped[p]>0) shopped[p]--;
			if (meterdisp[p]>0) meterdisp[p]--;
		}
		for (n=x=0;x<32;x++) for (y=0;y<24;y++)
		{ // Count pipes and blocks and dots remaining
			i=grid[x][y];
			n+=(i>=0x50 && i<=0x53 || i>>4==0xC || i>=0xD1 && i<=0xD6);
		}
		j=0;
		for every_enemy(i) if (at[i]==21) j++;
		if (!(am-j) && !n)
		{
			if (!blank) // Just finished current level
			{
				FILE *lev;
				am=0;
				for (x=0;x<32;x++)
					for (y=0;y<24;y++)
						grid[x][y]=heat[x][y]=0;
				levtrans_start=bplot;
				levtrans_complete=0;
				memset(back,0,xres*yres);
				lev=fopen(".\\levs.dat","rb");
				fseek(lev,level*1024,SEEK_SET);
				for (x=0;x<32;x++)
					for (y=0;y<24;y++)
					{
						unsigned char c=fgetc(lev);
						for every_player(p) if (c==0xFF-p)
						{
							destx[p]=x*20+10;
							desty[p]=y*20+10;
						}
					}
				fseek(lev,level*1024+851,SEEK_SET);
				background=fgetc(lev); bgsub=fgetc(lev);
				fclose(lev);
				calcmagfield(grid,magneticfield);
			}
			efx=(float)(bplot-levtrans_start)/bplotsize/bplotsize; // Seek starting squares
			if (efx<0) efx++;
			if (efx>1) efx=1;
			efx*=efx;
			efy=1.0-efx;
			for every_player(p)
			{
				px[p]=efy*px[p]+efx*destx[p];
				py[p]=efy*py[p]+efx*desty[p];
				xs[p]=ys[p]=0;
			}
			sprintf(str,"Level %d",level+1);
			for (n=0;n<strlen(str);n++)
			{
				char a=lc(str[n]);
				x=xres/2+Max(9.0,4e2/(blank+1))*(n-0.5*strlen(str));
				if (fabs(x-xres/2)<xres/2)
				{
					if (a>='a' && a<='z') wrapbit(x,yres/2-4,bletter[a-'a']);
					else if (a>='0' && a<='9') wrapbit(x,yres/2-4,bletter[26+a-'0']);
				}
			}
			if (blank>20) for (n=0;n<strlen(levname[level]);n++)
			{
				char a=lc(levname[level][n]),str2[100];
				strcpy(str2,levname[level]); str2[n]=0;
				x=xres/2+Max(1.0,1.5e2/(blank-20))*(estlength(str2)-0.5*estlength(levname[level]));
				if (fabs(x-xres/2)<xres/2)
				{
					if (a>='a' && a<='z') wrapbit(x,yres/2+6,letter[a-'a']);
					else if (a>='0' && a<='9') wrapbit(x,yres/2+6,letter[26+a-'0']);
				}
			}
			blank+=idt;
		}
		for every_player(p) if (lives[p]>0 && dead[p]>=100 && vsc[p]==score[p]) // New life
		{
			int nfm=Min(fm+1000,fmax);
			dead[p]=0;
			in=fopen(".\\levs.dat","rb");
			fseek(in,(level-1)*1024,SEEK_SET);
			for (x=0;x<32;x++)
				for (y=0;y<24;y++)
					if (fgetc(in)==0xFF-p)
					{
						px[p]=x*20+10; py[p]=y*20+10;
						xs[p]=ys[p]=0;
					}
			fclose(in);
			for new_dust(n)
			{
				float dd=0;
				fx[n]=px[p]; fy[n]=py[p];
				while (dd>25 || dd<20)
				{
					fxs[n]=frnd(10)-5; fys[n]=frnd(10)-5;
					dd=fxs[n]*fxs[n]+fys[n]*fys[n];
				}
				fc[n]=rnd(8)+24;
			}
			fm=nfm;
			for (n=0;n<upgrades;n++) upgrade[p][n]=basemodel[n];
			upgrade[p][12]=upgrade[p][11];
			upgrade[p][14]=upgrade[p][13];
		}
		// "GAME OVER" sign
		if (gameover) bwrite(xres/2-40,yres/2-4,xres,"GAME OVER",(int)(GetTickCount()-TSTART)/750%2*15);
		for every_player(p) if (upgrade[p][9]==1 && !dead[p]) // Warning
		{
			int nfm=Min(fm+1000,fmax);
			for new_dust(n)
			{
				float rad=frnd(sqrt(25)-sqrt(20))+sqrt(20),th=frnd(M_PI*2);
				fx[n]=px[p]; fy[n]=py[p];
				fxs[n]=rad*sin(th); fys[n]=rad*cos(th);
				fc[n]=1;
			}
			fm=nfm;
		}
		for every_player(p) if (vsc[p]<score[p] && blank>240) blank=240;
		if (blank>240 && levtrans_complete) // Next level *#*#*#*#*#*#*#*#*#*#*
		{
			FILE *lev;
			blank=0; level++; for every_player(p) upgrade[p][9]=250;
			lev=fopen(".\\levs.dat","rb");
			fseek(lev,(level-1)*1024,SEEK_SET);
			am=bm=0;
			for (x=0;x<32;x++) // Load grid
				for (y=0;y<24;y++)
				{
					i=fgetc(lev);
					grid[x][y]=i;
					if (i>=0x80 && i<=0xBF) // Enemy
					{
						grid[x][y]=0;
						ax[am]=x*20+10; ay[am]=y*20+10;
						axs[am]=frnd(2)-1; ays[am]=frnd(2)-1;
						ar[am]=ars[am]=0.1;
						if (i==0x80) at[am]=4; // saucer
						else if (i==0x81) at[am]=9; // time ship
						else if (i==0x82) {at[am]=11; as[am]=0;} // self-rep mine
						else if (i==0x83) at[am]=13; // pulsar
						else if (i==0x84) at[am]=14; // rubber duck
						else if (i>=0x85 && i<=0x8D) // asteroid
						{
							at[am]=1;
							as[am]=islegalsize(pot[i-0x85]);
						}
						else if (i>=0x8E && i<=0x99) // spinner
						{
							at[am]=2;
							if (i>=0x8E && i<=0x93) as[am]=1+i-0x8E;
							else if (i==0x94) as[am]=8;
							else if (i==0x95) as[am]=12;
							else if (i>=0x96 && i<=0x99) as[am]=pot[4+i-0x96];
						}
						else if (i>=0x9A && i<=0xA1) // trianguloid
						{
							at[am]=6;
							as[am]=1+i-0x9A;
						}
						else if (i>=0xA2 && i<=0xAA) // pentagram
						{
							at[am]=7;
							if (i>=0xA2 && i<=0xA7) as[am]=3+i-0xA2;
							else if (i==0xA8) as[am]=10;
							else if (i==0xA9) as[am]=12;
							else if (i==0xAA) as[am]=16;
						}
						else if (i>=0xAB && i<=0xAF) // macropixel
						{
							at[am]=10;
							as[am]=i-0xAB;
						}
						else if (i>=0xB0 && i<=0xB8) // recoalescing asteroid
						{
							at[am]=12;
							as[am]=islegalsize(pot[i-0xB0]);
						}
						else if (i==0xB9) at[am]=16; // bomber
						else if (i==0xBA) // baiter
						{
							at[am]=17;
							as[am]=5;
						}
						else if (i==0xBB) // homing spinner
						{
							at[am]=18;
							ar[am]=frnd(5)-2.5; ars[am]=0;
							as[am]=100+frnd(800);
						}
						else if (i==0xBC) // flamethrower boss
						{
							at[am]=19;
							ar[am]=0; ars[am]=0.02;
							as[am]=20; // health
							axs[am]=frnd(1)-0.5; ays[am]=frnd(1)-0.5;
						}
						else if (i==0xBD) // subgame N ball
						{
							at[am]=22;
							as[am]=6;
						}
						am++;
					}
					if (i>=0x100-MAXP) // Initial position of ship
					{
						px[p=0xFF-i]=x*20+10; py[p]=y*20+10;
						xs[p]=ys[p]=0;
						grid[x][y]=0;
					}
					heat[x][y]=0;
				}
			fseek(lev,(level-1)*1024+853,SEEK_SET);
			makeasteroidset(ast,asz,fgetc(lev)); // Asteroid set
			fclose(lev);
			loadblockset(blox,level);
			replotgrid(grid,blox); // Plot the grid
			calcmagfield(grid,magneticfield);
			for every_player(p)
			{
				if (upgrade[p][0]) upgrade[p][0]--;
				upgrade[p][2]=0;
				if (upgrade[p][5]) upgrade[p][5]--;
				upgrade[p][12]-=0.3*upgrade[p][12]+1;
				if (upgrade[p][12]<upgrade[p][11]) upgrade[p][12]=upgrade[p][11];
				upgrade[p][14]-=0.3*upgrade[p][14]+1;
				if (upgrade[p][14]<upgrade[p][13]) upgrade[p][14]=upgrade[p][13];
				if (upgrade[p][6]) upgrade[p][6]--;
			}
		}
		if (KEY(VK_P)) if (pause()) goto THE_END; // Pause button
		// Level display
		if (!blank)
		{
			if (level>=10000) wrapbitc(32,0,num[0][3][level/10000%10],15);
			if (level>=1000) wrapbitc(40,0,num[0][3][level/1000%10],15);
			if (level>=100) wrapbitc(48,0,num[0][3][level/100%10],15);
			if (level>=10) wrapbitc(56,0,num[0][3][level/10%10],15);
			wrapbitc(64,0,num[0][3][level%10],15);
		}
		for every_player(p) for (n=0;n<lives[p];n++)
		{
			int col=(n==lives[p]-1)?12:14;
			putpix(n*3+90,2+p*8,col);
			putpix(n*3+91,3+p*8,col);
			putpix(n*3+92,4+p*8,col);
			putpix(n*3+91,5+p*8,col);
			putpix(n*3+90,6+p*8,col);
		}
		// Score display
		for every_player(p) for (n=0;n<dt-frnd(1);n++)
		{
			if (vsc[p]<score[p]-2500) vsc[p]+=(score[p]-vsc[p])/23;
			else if (vsc[p]<score[p]-1000) vsc[p]=25*(int)(vsc[p]/25+1);
			else if (vsc[p]<score[p]-500) vsc[p]=10*(int)(vsc[p]/10+1);
			else if (vsc[p]<score[p]-300) vsc[p]=5*(int)(vsc[p]/5+1);
			else if (vsc[p]<score[p]-100) vsc[p]+=3;
			else if (vsc[p]<score[p]) vsc[p]++;
		}
		dispscores((1<<pm)-1,vsc);
		gtdisp(gametime,openfile); // Game-time display
		if (technical)
		{
			sprintf(str,"#%d:%02d:%02d",
				(int)(nextsave-gamestart)/3600,
				(int)((nextsave-gamestart)/60)%60,
				(int)(nextsave-gamestart)%60);
			rect(xres/2-1,5,41,7,0);
			write(xres/2,6,xres,str,12,8);
			sprintf(str,"%d",am);
			bwrite(0,yres-10,xres,str,2);
		}
		if (!signal_loadgame) cclb(); else cl(0);
		// Background plotting
		for (n=0;n<(blank?10:3) || (dt=GetTickCount()-framestart)<mswait;n++) // Synchronize end of frame
		{
			for (x=bplotx[bplot];x<xres;x+=bplotsize)
				for (y=bploty[bplot];y<yres;y+=bplotsize)
				{
					i=grid[x/20][y/20];
					if (i==0)
						putpixb(x,y,bcalc(x,y,background,bgsub,0.001*GetTickCount()));
					else if (blox[i][2+x%20+y%20*20]==255)
						putpixb(x,y,bcalc(x,y,background,bgsub,0.001*GetTickCount()));
					else if (i>=0x50 && i<=0x53) // Notches cut into WKobo blocks
					{
						int xi=(x/20+1)%32,xd=(x/20+31)%32,yi=(y/20+1)%24,yd=(y/20+23)%24;
						if (!grid[xi][y/20] && x%20>=14 && y%20>=7 && y%20<13)
							putpixb(x,y,bcalc(x,y,background,bgsub,0.001*GetTickCount()));
						else if (!grid[xd][y/20] && x%20<6 && y%20>=7 && y%20<13)
							putpixb(x,y,bcalc(x,y,background,bgsub,0.001*GetTickCount()));
						else if (!grid[x/20][yi] && y%20>=14 && x%20>=7 && x%20<13)
							putpixb(x,y,bcalc(x,y,background,bgsub,0.001*GetTickCount()));
						else if (!grid[x/20][yd] && y%20<6 && x%20>=7 && x%20<13)
							putpixb(x,y,bcalc(x,y,background,bgsub,0.001*GetTickCount()));
					}
				}
			bplot=(bplot+1)%(bplotsize*bplotsize);
			if (bplot==0) {bglag=0.001*GetTickCount()-bgstart; bgstart+=bglag;}
			if (bplot==levtrans_start) levtrans_complete=1;
		}
		dt/=mswait; if (dt>5) dt=5; else if (dt<1) dt=1;
		f12ed=KEY(VK_F12);
		if (KEY(VK_ESCAPE) || f12ed || (gameover>100 && kbhit)) // Quit/Suspend current game
		{
			leave=1;
			if (!*openfile && !gameover) // Save game screen
			{
				char fin=0;
				while (KEY(VK_ESCAPE));
				titlepal(gamma);
				str[0]=0;
				while (!fin && !dealwithmessages())
				{
					char a=lc(getch); n=strlen(str);
					if (KEY(VK_F12) || KEY(VK_ESCAPE)) {str[0]=0; fin=1;}
					else while (kbhit) if (dealwithmessages()) gracefulexit();
					memset(shad,0,xres*40);
					memset(shad+xres*40,160,xres*400);
					memset(shad+xres*440,0,xres*40);
					bwrite(60,60,xres,"Quit:",15);
					bwrite(60,80,xres,"Enter filename below to save\nor press enter to discard game",0);
					bwrite(60,100,xres,str,10);
					if (a>='a' && a<='z' || a>='0' && a<='9' || a=='-')
					{
						if (n<32) {str[n]=a; str[n+1]=0;}
					}
					else if (a==8 && n) str[n-1]=0;
					else if (a==13) fin=1;
					gtdisp(gametime,openfile);
					ccl(0);
				}
				if (*str)
				{
					gamefile=".\\";
					strcat(gamefile,strlwr(str));
					if (gamefile[2]>='a' && gamefile[2]<='z') gamefile[2]-=32;
					strcat(gamefile,".ncgi");
				}
				else gamefile="";
			}
			else strcpy(gamefile,openfile);
			signal_savegame=*gamefile;
		}
		if (signal_savegame && !gameover || gamedepth<strlen(gamestack)) //// Save Gamestate ////
		{
			begin_save:;
			if (gamedepth<strlen(gamestack))
			{
				char str2[10];
				strcpy(str,".\\");
				strcat(str,itoa(gamedepth,str2,10));
				strcat(str,".tmp");
			}
			else strcpy(str,gamefile);
			out=fopen(str,"wb");
			if (out==NULL) goto give_up_save;
			fwrite(&gametime,8,1,out);
			fwrite(&level,4,1,out);
			fwrite(&blank,4,1,out);
			fwrite(&pm,4,1,out);
			fwrite(&am,4,1,out);
			fwrite(&bm,4,1,out);
			fwrite(&im,4,1,out);
			fwrite(&ffm,4,1,out);
			fwrite(&fm,4,1,out);
			fwrite(&wm,4,1,out);
			for (x=0;x<32;x++)
				for (y=0;y<24;y++)
				{
					fputc(grid[x][y],out);
					fwrite(&heat[x][y],4,1,out);
				}
			for every_player(p)
			{
				fwrite(&lives[p],4,1,out);
				fwrite(&vr[p],4,1,out);
				fwrite(&dead[p],4,1,out);
				fwrite(&btype[p],4,1,out);
				fwrite(&mgrech[p],4,1,out);
				for (n=0;n<upgrades;n++) fwrite(&upgrade[p][n],4,1,out);
				fwrite(&score[p],4,1,out);
				fwrite(&vsc[p],4,1,out);
				fwrite(&money[p],4,1,out);
				fwrite(&px[p],4,1,out);
				fwrite(&py[p],4,1,out);
				fwrite(&xs[p],4,1,out);
				fwrite(&ys[p],4,1,out);
				fwrite(&rot[p],4,1,out);
				fwrite(&rs[p],4,1,out);
				fputc(subgame[p],out);
			}
			for every_enemy(n)
			{
				fwrite(&ax[n],4,1,out);
				fwrite(&ay[n],4,1,out);
				fwrite(&axs[n],4,1,out);
				fwrite(&ays[n],4,1,out);
				fwrite(&ar[n],4,1,out);
				fwrite(&ars[n],4,1,out);
				fwrite(&at[n],4,1,out);
				fwrite(&as[n],4,1,out);
				fputc(afire[n],out);
				fputc(aplas[n],out);
				fputc(awhit[n],out);
			}
			for every_bullet(n)
			{
				fwrite(&bx[n],4,1,out);
				fwrite(&by[n],4,1,out);
				fwrite(&bxs[n],4,1,out);
				fwrite(&bys[n],4,1,out);
				fwrite(&bs[n],4,1,out);
				fwrite(&bt[n],4,1,out);
				fputc(bo[n],out);
			}
			for every_item(n)
			{
				fwrite(&ix[n],4,1,out);
				fwrite(&iy[n],4,1,out);
				fwrite(&ixs[n],4,1,out);
				fwrite(&iys[n],4,1,out);
				fwrite(&iage[n],4,1,out);
				fwrite(&it[n],4,1,out);
			}
			for flame_particles(n)
			{
				fwrite(&ffx[n],4,1,out);
				fwrite(&ffy[n],4,1,out);
				fputc(ffc[n],out);
			}
			for dust_particles(n)
			{
				fwrite(&fx[n],4,1,out);
				fwrite(&fy[n],4,1,out);
				fwrite(&fxs[n],4,1,out);
				fwrite(&fys[n],4,1,out);
				fputc(fc[n],out);
			}
			for (n=0;n<wm;n++)
			{
				fwrite(&wx[n],4,1,out);
				fwrite(&wy[n],4,1,out);
				fwrite(&ws[n],4,1,out);
				fputc(wt[n],out);
			}
			fclose(out);
			signal_savegame=0;
			if (gamedepth<strlen(gamestack)) // Enter the subgame
			{
				char into=gamestack[gamedepth]; int them=0;
				for every_player(p) if (!dead[p])
				{
					them+=(1<<p)*(subgame[p]==into);
					if (sound_working && thstatus[p]>=0) // Stop that annoying noise
					{
						i=thstatus[p];
						stopsound(soundchannel[i]); killbuffer(soundchannel[i]);
						soundchannel[i]=makebuffer(100,10000);
						thstatus[p]=-1;
					}
				}
				if (them)
				{
					enum sg_retcode result;
					gamedepth++;
					result=startsubgame(into,them);
					gamedepth--;
					if (result==Error) goto THE_END;
					if (result==Quit) leave=1;
					if (result==Fail) for every_player(p) if (them&(1<<p)) // Kill off failures
					{
						int nfm=Min(fm+fmax/pm,fmax);
						for new_dust(i)
						{
							float rad=frnd(10)+1e-6,th=frnd(M_PI*2);
							fx[i]=px[p]; fy[i]=py[p];
							fxs[i]=rad*cos(th); fys[i]=rad*sin(th);
							fc[i]=32+rnd(24);
						}
						fm=nfm;
						ncgisound(px[p],n_dead,20000,5000,1);
						dead[p]=1; lives[p]--;
						strcpy(gamefile,openfile); signal_savegame=*openfile;
					}
					for every_player(p) if (subgame[p]==into)
					{
						subgame[p]=0;
						score[p]=(score[p]+4)/5*5;
						vsc[p]=score[p];
						upgrade[p][9]=250;
					}
					replotgrid(grid,blox);
					calcmagfield(grid,magneticfield);
				}
				deltmp(gamedepth);
				gamestack[gamedepth]=0;
			}
			if (signal_savegame) goto begin_save;
			give_up_save:;
		}
		if (signal_loadgame) //// Load Gamestate ////
		{
			in=fopen(gamefile,"rb");
			if (in==NULL) goto give_up_load;
			fread(&gametime,8,1,in);
			gamestart=0.001*GetTickCount()-gametime;
			fread(&level,4,1,in);
			fread(&blank,4,1,in);
			fread(&pm,4,1,in);
			fread(&am,4,1,in);
			fread(&bm,4,1,in);
			fread(&im,4,1,in);
			fread(&ffm,4,1,in);
			fread(&fm,4,1,in);
			fread(&wm,4,1,in);
			for (x=0;x<32;x++)
				for (y=0;y<24;y++)
				{
					grid[x][y]=fgetc(in);
					fread(&heat[x][y],4,1,in);
				}
			for every_player(p)
			{
				fread(&lives[p],4,1,in);
				fread(&vr[p],4,1,in);
				fread(&dead[p],4,1,in);
				fread(&btype[p],4,1,in);
				fread(&mgrech[p],4,1,in);
				for (n=0;n<upgrades;n++) fread(&upgrade[p][n],4,1,in);
				fread(&score[p],4,1,in);
				fread(&vsc[p],4,1,in);
				fread(&money[p],4,1,in);
				fread(&px[p],4,1,in);
				fread(&py[p],4,1,in);
				fread(&xs[p],4,1,in);
				fread(&ys[p],4,1,in);
				fread(&rot[p],4,1,in);
				fread(&rs[p],4,1,in);
				subgame[p]=fgetc(in);
			}
			for every_enemy(n)
			{
				fread(&ax[n],4,1,in);
				fread(&ay[n],4,1,in);
				fread(&axs[n],4,1,in);
				fread(&ays[n],4,1,in);
				fread(&ar[n],4,1,in);
				fread(&ars[n],4,1,in);
				fread(&at[n],4,1,in);
				fread(&as[n],4,1,in);
				afire[n]=fgetc(in);
				aplas[n]=fgetc(in);
				awhit[n]=fgetc(in);
			}
			for every_bullet(n)
			{
				fread(&bx[n],4,1,in);
				fread(&by[n],4,1,in);
				fread(&bxs[n],4,1,in);
				fread(&bys[n],4,1,in);
				fread(&bs[n],4,1,in);
				fread(&bt[n],4,1,in);
				bo[n]=fgetc(in);
			}
			for every_item(n)
			{
				fread(&ix[n],4,1,in);
				fread(&iy[n],4,1,in);
				fread(&ixs[n],4,1,in);
				fread(&iys[n],4,1,in);
				fread(&iage[n],4,1,in);
				fread(&it[n],4,1,in);
			}
			for flame_particles(n)
			{
				fread(&ffx[n],4,1,in);
				fread(&ffy[n],4,1,in);
				ffc[n]=fgetc(in);
			}
			for dust_particles(n)
			{
				fread(&fx[n],4,1,in);
				fread(&fy[n],4,1,in);
				fread(&fxs[n],4,1,in);
				fread(&fys[n],4,1,in);
				fc[n]=fgetc(in);
			}
			for (n=0;n<wm;n++)
			{
				fread(&wx[n],4,1,in);
				fread(&wy[n],4,1,in);
				fread(&ws[n],4,1,in);
				wt[n]=fgetc(in);
			}
			fclose(in);
			levtrans_complete=1;
			signal_loadgame=0;
			memset(back,0,xres*yres);
			{ // Load level
				FILE *lev=fopen(".\\levs.dat","rb");
				fseek(lev,(level-1)*1024+851,SEEK_SET);
				background=fgetc(lev); bgsub=fgetc(lev);
				makeasteroidset(ast,asz,fgetc(lev));
				fclose(lev);
				loadblockset(blox,level);
			}
			replotgrid(grid,blox);
			calcmagfield(grid,magneticfield);
			give_up_load:;
		}
	}
	if (sound_working) for every_player(p) // Stop that annoying noise
		if ((i=thstatus[p])>=0)
		{
			stopsound(soundchannel[i]); killbuffer(soundchannel[i]);
			soundchannel[i]=makebuffer(100,10000);
			thstatus[p]=-1;
		}
	if (f12ed) goto THE_END;
	names=(char **)malloc(HILEN*sizeof(char *)); scores=(long *)malloc(HILEN*sizeof(long));
	for (n=0;n<HILEN;n++) names[n]=(char *)malloc(MAXNAMELEN);
	titlepal(gamma);
//	pal(160,pow(0.0625,1.0/gamma)*255,pow(0.25,1.0/gamma)*255,pow(0.5,1.0/gamma));
	while (KEY(VK_ESCAPE));
	for every_player(p)
	{
		if (pm==1) sprintf(str,"Final score:  %ld  on level  %d",score[p],level);
			else sprintf(str,"Player %d / final score:  %ld  on level  %d",p+1,score[p],level);
		gethiscores(names,scores);
		place=HILEN;
		while (scores[place-1]<=score[p] && place>0) place--;
		if (place<HILEN && lives[p]==0)
		{
			char *str2=(char *)malloc(40),a;
			memset(back,0,xres*40);
			memset(back+xres*40,160,xres*400);
			memset(back+xres*440,0,xres*40);
			cclb();
			sprintf(str2,"New high-score: %d%s place",place+1,ordinal(place+1));
			for (n=HILEN-1;n>place;n--)
			{
				strcpy(names[n],names[n-1]);
				scores[n]=scores[n-1];
			}
			names[place]="";
			for (leave=0;!leave;)
			{
				bwrite(xres/2-40,yres/2-4,xres,"GAME OVER",15);
				bwrite(30,yres/2+40,xres,str,15);
				bwrite(30,yres/2+60,xres,str2,15);
				bwrite(30,yres/2+70,xres,names[place],10);
				cclb();
				while (kbhit) if (dealwithmessages()) goto THE_END;
				while (!kbhit) if (dealwithmessages()) goto THE_END;
				a=lc(getch);
				if (a==13) leave=1;
				else if ((i=strlen(names[place]))>0 && a==8)
				{
					i--;
					names[place][i]=0;
				}
				else if (n<MAXNAMELEN-1)
				{
					names[place][i]=a;
					names[place][i+1]=0;
				}
			}
			free(str2);
			scores[place]=score[p];
			sethiscores(names,scores);
		}
	}
	for (n=0;n<HILEN;n++) free(names[n]); free(names); free(scores);
	*cmdline=0;
	goto title_screen;
THE_END:
	gracefulexit();
	return 0;
}

#define participants(p) (p=0;p<MAXP;p++) if (them>>p&1)

#include "N_Particles.c"
#include "C_Squares.c"
#include "G_Platform.c"
#include "I_Space.c"
